/*
 * BRIEF MODULE DESCRIPTION
 *	Au1000 Power Management routines.
 *
 * Copyright 2001 MontaVista Software Inc.
 * Author: MontaVista Software, Inc.
 *		ppopov@mvista.com or source@mvista.com
 *
 *  Some of the routines are right out of init/main.c, whose
 *  copyrights apply here.
 *
 *  This program is free software; you can redistribute	 it and/or modify it
 *  under  the terms of	 the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the	License, or (at your
 *  option) any later version.
 *
 *  THIS  SOFTWARE  IS PROVIDED	  ``AS	IS'' AND   ANY	EXPRESS OR IMPLIED
 *  WARRANTIES,	  INCLUDING, BUT NOT  LIMITED  TO, THE IMPLIED WARRANTIES OF
 *  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN
 *  NO	EVENT  SHALL   THE AUTHOR  BE	 LIABLE FOR ANY	  DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED	  TO, PROCUREMENT OF  SUBSTITUTE GOODS	OR SERVICES; LOSS OF
 *  USE, DATA,	OR PROFITS; OR	BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 *  ANY THEORY OF LIABILITY, WHETHER IN	 CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  You should have received a copy of the  GNU General Public License along
 *  with this program; if not, write  to the Free Software Foundation, Inc.,
 *  675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <linux/init.h>
#include <linux/slab.h>
#include <linux/sysctl.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>
#include <linux/time.h>
#include <linux/jiffies.h>

#include <asm/string.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/mach-au1x00/au1000.h>
#include <asm/mach-au1x00/au1xxx_pm.h>
#include <asm/mach-au1x00/au1xxx_psc.h>
#include <asm/mach-au1x00/au1xxx_dbdma.h>

#if defined(CONFIG_MIPS_PB1200)
#include <asm/mach-pb1x00/pb1200.h>
#elif defined (CONFIG_MIPS_DB1200)
#include <asm/mach-db1x00/db1200.h>
#endif

#define DEBUG 0
#if DEBUG
#  define DPRINTK(fmt, args...)	printk("%s: " fmt, __FUNCTION__ , ## args)
#else
#  define DPRINTK(fmt, args...)
#endif

#ifdef CONFIG_PM

#define AU1XXX_PM_VERSION "1.0"
#define TRUE 1
#define FALSE 0
#define MAX_POWER_CONF_LEN 64
/* set this to a non-zero value to enable a fake device for testing */
#define ENABLE_TEST_DEV 0

/* function definitions */
int remove_au1xxx_pm_error(au1xxx_pm_error_t *error);
static void au1xxx_pm_service_timer(unsigned long var);

static void au1xxx_pm_debug(void);

/* system settings */
static struct proc_dir_entry *power_dir;
static struct proc_dir_entry *timeout_dir;
static struct proc_dir_entry *pending_proc_entry;
static struct proc_dir_entry *error_proc_entry;

au1xxx_power_dev_t sys_power_array[MAX_SUPPORTED_DEVICES]; 
au1xxx_power_dev_t *sys_power_pointers[MAX_SUPPORTED_DEVICES];
au1xxx_power_dev_t *system_dev = &sys_power_array[0];

u16 sleep_bcsr_int, sleep_bcsr_mask,sleep_bcsr_resets;
u16 sleep_bcsr_pcmcia, sleep_bcsr_boards;


struct au1xxx_pm_settings { 
	int wakeup_mode;
	int sleep_time;  /* in seconds */
	int feedback_mode;
	int valid;
	struct proc_dir_entry *proc_entry;
};
typedef struct au1xxx_pm_settings au1xxx_pm_settings_t;

/* Definition PM Catch All */
struct pm_module
{
	char modname[64];
	char pmname[64];
	spinlock_t	lock;			/* Used to block on state transitions */
	au1xxx_power_dev_t *dev;	/* Power Managers device structure */
};

typedef struct pm_module pm_module;

char modprobe_path_pm[64] = "/sbin/modprobe";

/* Add entries for modules that need power Management support
   To add : .modname should be the name of the module when you do lsmod
   .pmname could be anything.
   For eg: Ethernet:  
   root@(none):~# lsmod
   Module                  Size  Used by
   smc91111               25248  0
  
 */
static pm_module pm_modules[] = {
	{	.modname = "ehci-hcd",	.pmname="ehci" },
	{	.modname = "ohci-hcd",	.pmname="ohci" },
	{	.modname = "smc91111",	.pmname="ethernet" }};

static int num_modules=sizeof(pm_modules)/sizeof(pm_module);

/* End PM Catch All definition*/


au1xxx_pm_settings_t set;
au1xxx_pm_settings_t *au1xxx_pm_setup = &set;

static spinlock_t au1xxx_pm_lock = SPIN_LOCK_UNLOCKED;
unsigned long spin_lock_flags;

/* error related */
au1xxx_pm_error_t sys_error_array[MAX_SUPPORTED_ERRORS];
au1xxx_pm_error_t *sys_error_pointers[MAX_SUPPORTED_ERRORS];
au1xxx_pm_error_t *first_error = &sys_error_array[0];

/* timer related */
static struct timeval zero_timeval = {0, 0};

/* We need to save/restore a bunch of core registers that are
 * either volatile or reset to some state across a processor sleep.
 * If reading a register doesn't provide a proper result for a
 * later restore, we have to provide a function for loading that
 * register and save a copy.
 *
 * We only have to save/restore registers that aren't otherwise
 * done as part of a driver au1xxx_pm_* function.
 */
static uint	sleep_aux_pll_cntrl;
static uint	sleep_cpu_pll_cntrl;
static uint	sleep_pin_function;
static uint	sleep_uart0_inten;
static uint	sleep_uart0_fifoctl;
static uint	sleep_uart0_linectl;
static uint	sleep_uart0_clkdiv;
static uint	sleep_uart0_enable;
static uint	sleep_usbhost_enable;
static uint	sleep_usbdev_enable;
static uint	sleep_static_memctlr[4][3];
static uint sleep_ddma[16][8];
static uint sleep_psc[1][12];
static uint sleep_sys_clksrc, sleep_sys_clksrc_temp;
static uint sleep_sys_freqctl0;
static uint sleep_sys_freqctl1;
static uint sleep_mae[60];

/* Define this to cause the value you write to /proc/sys/pm/sleep to
 * set the TOY timer for the amount of time you want to sleep.
 * The value is number of seconds to sleep.
 */
extern void wakeup_toym2_set(int ticks);
  
/***********************************************************************
 *                     SLEEP FUNCTIONS
 ***********************************************************************/
static void save_core_regs(void) {
	extern void save_au1xxx_intctl(void);
	extern void pm_eth0_shutdown(void);
	unsigned int addr,i;

	/* Do the serial ports.....these really should be a pm_*
	 * registered function by the driver......but of course the
	 * standard serial driver doesn't understand our Au1xxx
	 * unique registers.
	 */
	sleep_uart0_inten = au_readl(UART0_ADDR + UART_IER);
	sleep_uart0_fifoctl = au_readl(UART0_ADDR + UART_FCR);
	sleep_uart0_linectl = au_readl(UART0_ADDR + UART_LCR);
	sleep_uart0_clkdiv = au_readl(UART0_ADDR + UART_CLK);
	sleep_uart0_enable = au_readl(UART0_ADDR + UART_MOD_CNTRL);

#ifndef CONFIG_SOC_AU1200
	/* Shutdown USB host/device.  */
	sleep_usbhost_enable = au_readl(USB_HOST_CONFIG);

	/* There appears to be some undocumented reset register....  */
	au_writel(0, 0xb0100004); au_sync();
	au_writel(0, USB_HOST_CONFIG); au_sync();

	sleep_usbdev_enable = au_readl(USBD_ENABLE);
	au_writel(0, USBD_ENABLE); au_sync();
#endif

	/* Save interrupt controller state.  */
	save_au1xxx_intctl();

	/* Clocks and PLLs.  */
	sleep_aux_pll_cntrl = au_readl(SYS_AUXPLL);

	/* We don't really need to do this one, but unless we
	 * write it again it won't have a valid value if we
	 * happen to read it. */
	sleep_cpu_pll_cntrl = au_readl(SYS_CPUPLL);

	sleep_pin_function = au_readl(SYS_PINFUNC);

	/* Save the static memory controller configuration. */
	sleep_static_memctlr[0][0] = au_readl(MEM_STCFG0);
	sleep_static_memctlr[0][1] = au_readl(MEM_STTIME0);
	sleep_static_memctlr[0][2] = au_readl(MEM_STADDR0);
	sleep_static_memctlr[1][0] = au_readl(MEM_STCFG1);
	sleep_static_memctlr[1][1] = au_readl(MEM_STTIME1);
	sleep_static_memctlr[1][2] = au_readl(MEM_STADDR1);
	sleep_static_memctlr[2][0] = au_readl(MEM_STCFG2);
	sleep_static_memctlr[2][1] = au_readl(MEM_STTIME2);
	sleep_static_memctlr[2][2] = au_readl(MEM_STADDR2);
	sleep_static_memctlr[3][0] = au_readl(MEM_STCFG3);
	sleep_static_memctlr[3][1] = au_readl(MEM_STTIME3);
	sleep_static_memctlr[3][2] = au_readl(MEM_STADDR3);
	
	/* Save the DDMA gloabl regs */
	addr = DDMA_GLOBAL_BASE;
	sleep_ddma[0][0] = au_readl(addr);
	sleep_ddma[0][1] = au_readl(addr+4);
	sleep_ddma[0][2] = au_readl(addr+8);
	sleep_ddma[0][3] = au_readl(addr+0xC);

	for(i=1,addr=DDMA_CHANNEL_BASE; i<NUM_DBDMA_CHANS; i++, addr+=0x100 ) {
		sleep_ddma[i][0] = au_readl(addr);
		sleep_ddma[i][1] = au_readl(addr+4);
		sleep_ddma[i][2] = au_readl(addr+8);
		sleep_ddma[i][3] = au_readl(addr+0xC);
		sleep_ddma[i][4] = au_readl(addr+0x10);
		sleep_ddma[i][5] = au_readl(addr+0x14);
		sleep_ddma[i][6] = au_readl(addr+0x18);	
	}

	/* Save the Programable Serial Controller registers */
	for (i = 0, addr = PSC0_BASE_ADDR; addr < (PSC0_BASE_ADDR + 0x30); i++, addr += 0x04)
		sleep_psc[0][i] = au_readl(addr);
	for (i = 0, addr = PSC1_BASE_ADDR; addr < (PSC1_BASE_ADDR + 0x30); i++, addr += 0x04)
		sleep_psc[1][i] = au_readl(addr);
	
	/* Save clock information */
	sleep_sys_clksrc = au_readl(SYS_CLKSRC);
	sleep_sys_freqctl0 = au_readl(SYS_FREQCTRL0);
	sleep_sys_freqctl1 = au_readl(SYS_FREQCTRL1);

	DPRINTK("Saving MAE FE regs\n");
	/* Save MAE front end registers */
	addr = 0xb4012000;
	sleep_mae[0] = au_readl(addr + 0x0000);
	sleep_mae[1] = au_readl(addr + 0x0004);
	sleep_mae[2] = au_readl(addr + 0x0008);
	sleep_mae[3] = au_readl(addr + 0x000C);
	sleep_mae[4] = au_readl(addr + 0x0010);
	sleep_mae[5] = au_readl(addr + 0x0014);
	sleep_mae[6] = au_readl(addr + 0x0018);
	sleep_mae[7] = au_readl(addr + 0x001C);
	sleep_mae[8] = au_readl(addr + 0x0020);
	sleep_mae[9] = au_readl(addr + 0x0024);
	sleep_mae[10] = au_readl(addr + 0x0028);
	sleep_mae[11] = au_readl(addr + 0x002C);
	sleep_mae[12] = au_readl(addr + 0x0038);
	sleep_mae[13] = au_readl(addr + 0x003C);
	sleep_mae[14] = au_readl(addr + 0x0040);
	sleep_mae[15] = au_readl(addr + 0x0044);
	sleep_mae[16] = au_readl(addr + 0x0048);
	sleep_mae[17] = au_readl(addr + 0x004C);
	sleep_mae[18] = au_readl(addr + 0x0050);
	sleep_mae[19] = au_readl(addr + 0x0054);
	sleep_mae[20] = au_readl(addr + 0x0058);
	sleep_mae[21] = au_readl(addr + 0x005C);
	sleep_mae[22] = au_readl(addr + 0x1004);
	sleep_mae[23] = au_readl(addr + 0x1008);
	DPRINTK("Saving MAE BE regs\n");
	/* Save MAE backend registers */
	addr = 0xb4010000;
	sleep_mae[24] = au_readl(addr + 0x0000);
	sleep_mae[25] = au_readl(addr + 0x0004);
	sleep_mae[26] = au_readl(addr + 0x0008);
	sleep_mae[27] = au_readl(addr + 0x0100);
	sleep_mae[28] = au_readl(addr + 0x0180);
	sleep_mae[29] = au_readl(addr + 0x0200);
	sleep_mae[30] = au_readl(addr + 0x0280);
	sleep_mae[31] = au_readl(addr + 0x0300);
	sleep_mae[32] = au_readl(addr + 0x0380);
	DPRINTK("Saving MAE CSC regs\n");
	/* CSC registers */
	sleep_mae[33] = au_readl(addr + 0x0400);
	sleep_mae[34] = au_readl(addr + 0x0404);
	sleep_mae[35] = au_readl(addr + 0x0408);
	sleep_mae[36] = au_readl(addr + 0x040C);
	sleep_mae[37] = au_readl(addr + 0x0410);
	sleep_mae[38] = au_readl(addr + 0x0414);
	sleep_mae[39] = au_readl(addr + 0x0418);
	sleep_mae[40] = au_readl(addr + 0x041C);
	sleep_mae[41] = au_readl(addr + 0x0420);
	sleep_mae[42] = au_readl(addr + 0x0424);
	sleep_mae[43] = au_readl(addr + 0x0428);
	sleep_mae[44] = au_readl(addr + 0x042C);
	sleep_mae[45] = au_readl(addr + 0x0430);
	DPRINTK("Saving MAE SRC regs\n");
	/* SRC registers */
	sleep_mae[46] = au_readl(addr + 0x0500);
	sleep_mae[47] = au_readl(addr + 0x0504);
	sleep_mae[48] = au_readl(addr + 0x0508);
	sleep_mae[49] = au_readl(addr + 0x050C);
	sleep_mae[50] = au_readl(addr + 0x0510);
	sleep_mae[51] = au_readl(addr + 0x0514);
	sleep_mae[52] = au_readl(addr + 0x0518);
	sleep_mae[53] = au_readl(addr + 0x051C);
	DPRINTK("Saving MAE DST regs\n");
	/* DST registers */
	sleep_mae[54] = au_readl(addr + 0x0600);
	sleep_mae[55] = au_readl(addr + 0x0604);
	sleep_mae[56] = au_readl(addr + 0x0608);
	sleep_mae[57] = au_readl(addr + 0x060C);
	DPRINTK("Saving MAE CTL regs\n");
	/* CTL registers */
	sleep_mae[58] = au_readl(addr + 0x0700);
	sleep_mae[59] = au_readl(addr + 0x070C);
	sleep_mae[60] = au_readl(addr + 0x0710);

#if defined(CONFIG_MIPS_PB1200) || defined(CONFIG_MIPS_DB1200)
	/* save board register */
	sleep_bcsr_int = bcsr->intset;
	sleep_bcsr_mask = bcsr->intset_mask;
	sleep_bcsr_resets = bcsr->resets;
	sleep_bcsr_pcmcia = bcsr->pcmcia;
	sleep_bcsr_boards = bcsr->board;
#endif
}

static void restore_core_regs(void) {
	extern void restore_au1xxx_intctl(void);
	extern void wakeup_kernel_timer(void);
	unsigned int addr,i;

	au_writel(sleep_aux_pll_cntrl, SYS_AUXPLL); au_sync();
	au_writel(sleep_cpu_pll_cntrl, SYS_CPUPLL); au_sync();
	au_writel(sleep_pin_function, SYS_PINFUNC); au_sync();

	/* Restore the static memory controller configuration. */
	au_writel(sleep_static_memctlr[0][0], MEM_STCFG0);
	au_writel(sleep_static_memctlr[0][1], MEM_STTIME0);
	au_writel(sleep_static_memctlr[0][2], MEM_STADDR0);
	au_writel(sleep_static_memctlr[1][0], MEM_STCFG1);
	au_writel(sleep_static_memctlr[1][1], MEM_STTIME1);
	au_writel(sleep_static_memctlr[1][2], MEM_STADDR1);
	au_writel(sleep_static_memctlr[2][0], MEM_STCFG2);
	au_writel(sleep_static_memctlr[2][1], MEM_STTIME2);
	au_writel(sleep_static_memctlr[2][2], MEM_STADDR2);
	au_writel(sleep_static_memctlr[3][0], MEM_STCFG3);
	au_writel(sleep_static_memctlr[3][1], MEM_STTIME3);
	au_writel(sleep_static_memctlr[3][2], MEM_STADDR3);

	/* Enable the UART if it was enabled before sleep.
	 * I guess I should define module control bits........ */
	if (sleep_uart0_enable & 0x02) {
		au_writel(0, UART0_ADDR + UART_MOD_CNTRL); au_sync();
		au_writel(1, UART0_ADDR + UART_MOD_CNTRL); au_sync();
		au_writel(3, UART0_ADDR + UART_MOD_CNTRL); au_sync();
		au_writel(sleep_uart0_inten, UART0_ADDR + UART_IER); au_sync();
		au_writel(sleep_uart0_fifoctl, UART0_ADDR + UART_FCR); au_sync();
		au_writel(sleep_uart0_linectl, UART0_ADDR + UART_LCR); au_sync();
		au_writel(sleep_uart0_clkdiv, UART0_ADDR + UART_CLK); au_sync();
	}

	/* Save the DDMA gloabl regs */
	addr = DDMA_GLOBAL_BASE;
	au_writel(sleep_ddma[0][0],addr+0x0);
	au_writel(sleep_ddma[0][1],addr+0x4);
	au_writel(sleep_ddma[0][2],addr+0x8);
	au_writel(sleep_ddma[0][3],addr+0xC);

	for(i=1,addr=DDMA_CHANNEL_BASE; i<NUM_DBDMA_CHANS; i++, addr+=0x100 )	{
		au_writel( sleep_ddma[i][0], addr+0x0);
		au_writel( sleep_ddma[i][1], addr+0x4);
		au_writel( sleep_ddma[i][2], addr+0x8);
		au_writel( sleep_ddma[i][3], addr+0xC);
		au_writel( sleep_ddma[i][4], addr+0x10);
		au_writel( sleep_ddma[i][5], addr+0x14);
		au_writel( sleep_ddma[i][6], addr+0x18);
	}

	/* Restore Programmable Serial Controller registers */
	for (i = 0, addr = PSC0_BASE_ADDR; addr < (PSC0_BASE_ADDR + 0x30); i++, addr += 0x04)
		au_writel(sleep_psc[0][i], addr);
	for (i = 0, addr = PSC1_BASE_ADDR; addr < (PSC1_BASE_ADDR + 0x30); i++, addr += 0x04)
		au_writel(sleep_psc[1][i], addr);
	
	au_writel(sleep_sys_clksrc , SYS_CLKSRC);
	au_writel(sleep_sys_freqctl0, SYS_FREQCTRL0);
	au_writel(sleep_sys_freqctl1, SYS_FREQCTRL1);

	if (au_readl(PSC0_BASE_ADDR+0x04) == 0x03) {
		u32 stat0;
		do {
			stat0 = au_readl(PSC0_BASE_ADDR+0x14);
		} while ((stat0 & PSC_SMBSTAT_SR) == 0);
	} 
 
	DPRINTK("Restoring MAE FE regs\n");
	/* Save MAE front end registers */
	addr = 0xb4012000;
	au_writel(sleep_mae[0], addr + 0x0000);
	au_writel(sleep_mae[1], addr + 0x0004);
	au_writel(sleep_mae[2], addr + 0x0008);
	au_writel(sleep_mae[3], addr + 0x000C);
	au_writel(sleep_mae[4], addr + 0x0010);
	au_writel(sleep_mae[5], addr + 0x0014);
	au_writel(sleep_mae[6], addr + 0x0018);
	au_writel(sleep_mae[7], addr + 0x001C);
	au_writel(sleep_mae[8], addr + 0x0020);
	au_writel(sleep_mae[9], addr + 0x0024);
	au_writel(sleep_mae[10], addr + 0x0028);
	au_writel(sleep_mae[11], addr + 0x002C);
	au_writel(sleep_mae[12], addr + 0x0038);
	au_writel(sleep_mae[13], addr + 0x003C);
	au_writel(sleep_mae[14], addr + 0x0040);
	au_writel(sleep_mae[15], addr + 0x0044);
	au_writel(sleep_mae[16], addr + 0x0048);
	au_writel(sleep_mae[17], addr + 0x004C);
	au_writel(sleep_mae[18], addr + 0x0050);
	au_writel(sleep_mae[19], addr + 0x0054);
	au_writel(sleep_mae[20], addr + 0x0058);
	au_writel(sleep_mae[21], addr + 0x005C);
	au_writel(sleep_mae[22], addr + 0x1004);
	au_writel(sleep_mae[23], addr + 0x1008);
	DPRINTK("Restoring MAE BE regs\n");
	/* Save MAE backend registers */
	addr = 0xb4010000;
	au_writel(sleep_mae[24], addr + 0x0000);
	au_writel(sleep_mae[25], addr + 0x0004);
	au_writel(sleep_mae[26], addr + 0x0008);
	au_writel(sleep_mae[27], addr + 0x0100);
	au_writel(sleep_mae[28], addr + 0x0180);
	au_writel(sleep_mae[29], addr + 0x0200);
	au_writel(sleep_mae[30], addr + 0x0280);
	au_writel(sleep_mae[31], addr + 0x0300);
	au_writel(sleep_mae[32], addr + 0x0380);
	DPRINTK("Restoring MAE CSC regs\n");
	/* CSC registers */
	au_writel(sleep_mae[33], addr + 0x0400);
	au_writel(sleep_mae[34], addr + 0x0404);
	au_writel(sleep_mae[35], addr + 0x0408);
	au_writel(sleep_mae[36], addr + 0x040C);
	au_writel(sleep_mae[37], addr + 0x0410);
	au_writel(sleep_mae[38], addr + 0x0414);
	au_writel(sleep_mae[39], addr + 0x0418);
	au_writel(sleep_mae[40], addr + 0x041C);
	au_writel(sleep_mae[41], addr + 0x0420);
	au_writel(sleep_mae[42], addr + 0x0424);
	au_writel(sleep_mae[43], addr + 0x0428);
	au_writel(sleep_mae[44], addr + 0x042C);
	au_writel(sleep_mae[45], addr + 0x0430);
	DPRINTK("Restoring MAE SRC regs\n");
	/* SRC registers */
	au_writel(sleep_mae[46], addr + 0x0500);
	au_writel(sleep_mae[47], addr + 0x0504);
	au_writel(sleep_mae[48], addr + 0x0508);
	au_writel(sleep_mae[49], addr + 0x050C);
	au_writel(sleep_mae[50], addr + 0x0510);
	au_writel(sleep_mae[51], addr + 0x0514);
	au_writel(sleep_mae[52], addr + 0x0518);
	au_writel(sleep_mae[53], addr + 0x051C);
	DPRINTK("Restoring MAE DST regs\n");
	/* DST registers */
	au_writel(sleep_mae[54], addr + 0x0600);
	au_writel(sleep_mae[55], addr + 0x0604);
	au_writel(sleep_mae[56], addr + 0x0608);
	au_writel(sleep_mae[57], addr + 0x060C);
	DPRINTK("Restoring MAE CTL regs\n");
	/* CTL registers */
	au_writel(sleep_mae[58], addr + 0x0700);
	au_writel(sleep_mae[59], addr + 0x070C);
	au_writel(sleep_mae[60], addr + 0x0710);

#if defined(CONFIG_MIPS_PB1200) || defined(CONFIG_MIPS_DB1200)
	/* restore board register */
	bcsr->intclr = 0xffff; au_sync();
	bcsr->intset = sleep_bcsr_int; au_sync();
	bcsr->intclr_mask = 0xffff; au_sync();
	bcsr->intset_mask = sleep_bcsr_mask; au_sync();
	bcsr->resets = 0xffff; au_sync();
	bcsr->resets = sleep_bcsr_resets;
	bcsr->pcmcia = 0xffff; au_sync();
	bcsr->pcmcia = sleep_bcsr_pcmcia;
	bcsr->board = 0xffff; au_sync();
	bcsr->board = sleep_bcsr_boards;
#endif
	
	restore_au1xxx_intctl();
	wakeup_kernel_timer();
}
								 
int au_sleep(void) {
	unsigned long wakeup, flags;
	extern unsigned int save_and_sleep(void);

	spin_lock_irqsave(&au1xxx_pm_lock,flags);
  
 	save_core_regs();
  
	/** The code in the if-else statements below is all system dependent 
	 ** and we should probably have a function call out of here to set 
	 ** this up.  You need to configure the GPIO or timer interrupts that 
	 ** will bring you out of sleep.
	 ** For testing, the TOY counter wakeup is useful.
	 **/
  
	if (au1xxx_pm_setup->wakeup_mode == GPIO_WAKEUP) {
		DPRINTK("Attempting GPIO based sleep\n");
		DPRINTK("sleep_ticks: %d\n", au1xxx_pm_setup->sleep_time);
		au_writel(au_readl(SYS_PINSTATERD) & ~(1 << 11), SYS_PINSTATERD);

		/* gpio 5 can cause a wake up event */
		wakeup = au_readl(SYS_WAKEMSK);
		wakeup &= ~(1 << 8);	/* turn off match20 wakeup */
		wakeup = 1 << 5;	/* turn on gpio 5 wakeup   */
	} else if (au1xxx_pm_setup->wakeup_mode == TIMER_WAKEUP) {
		DPRINTK("Attempting TOY based sleep\n");
		/* For testing, allow match20 to wake us up.  */
		wakeup_toym2_set(au1xxx_pm_setup->sleep_time);
		wakeup = 1 << 8;	/* turn on match20 wakeup   */
		wakeup |= 1 << 5;  /* turn on gpio 5 wakeup */
	} else {
		printk(KERN_ERR "Unable to sleep: power management settings are invalid.\n");
		return -EFAULT;
	}
	DPRINTK("wakeup: \t0x%08X\n", (unsigned int) wakeup);
	au_writel(0, SYS_WAKESRC);	/* clear cause */
	au_writel(wakeup, SYS_WAKEMSK);
	DPRINTK("Entering sleep!\n");
	save_and_sleep();

	/* after a wakeup, the cpu vectors back to 0x1fc00000 so
	 * it's up to the boot code to get us back here.
	 */
	restore_core_regs();
	spin_unlock_irqrestore(&au1xxx_pm_lock, flags);
	DPRINTK("Leaving sleep!\n");
	return 0;
}


/***********************************************************************
 *                     CALLBACK FUNCTIONS
 ***********************************************************************/
static int system_power_callback(au1xxx_power_dev_t *device, 
	au1xxx_request_t request, void *data)
{
	int retval = 0;
	int i = 0;
	int power_level;

	DPRINTK("Inside the system_power_callback function: %lu\n", jiffies);
	if (request == AU1XXX_PM_GETSTATUS) {
		/* system device is always awake when processing this function */
		retval = AWAKE_STATE;
	} else if (request == AU1XXX_PM_SLEEP) {
		power_level = *((int *)data);

		/* loop to put all the other devices that are awake to sleep */
		for (i = 1; i < MAX_SUPPORTED_DEVICES; i++) {
			if ((sys_power_pointers[i]->valid) && 
					(sys_power_pointers[i]->cur_state != SLEEP_STATE))
			{
				DPRINTK("Suspending Device: %s\n", device->name);
				sys_power_pointers[i]->prev_state = sys_power_pointers[i]->cur_state;
				sys_power_pointers[i]->prev_state_timestamp = sys_power_pointers[i]->cur_state_timestamp;
				sys_power_pointers[i]->cur_state = SLEEP_STATE;
				do_gettimeofday(&(sys_power_pointers[i]->cur_state_timestamp));
				sys_power_pointers[i]->callback(sys_power_pointers[i],
					AU1XXX_PM_SLEEP, &power_level);
			}
		}
		/* put the system device to sleep */
		device->prev_state = device->cur_state;
		device->prev_state_timestamp = device->cur_state_timestamp;
		do_gettimeofday(&(device->cur_state_timestamp));
		device->cur_state = SLEEP_STATE;

		if (!au1xxx_pm_setup->valid) {
			printk(KERN_ERR "Unable to sleep: power management settings are invalid.\n");
			retval = -EFAULT;
		}
		DPRINTK("retval: %d (must be zero to continue into sleep)\n", retval);
		if (retval)
			return retval;
		au_sleep();

		DPRINTK("resume pc0: \t0x%08X\n", au_readl(SYS_TOYREAD));
		DPRINTK("resume toymatch2: \t0x%08X\n", au_readl(SYS_TOYMATCH2));
		DPRINTK("resume wakesrc: \t0x%08X\n", au_readl(SYS_WAKESRC));
		DPRINTK("resume wakemask: \t0x%08X\n", au_readl(SYS_WAKEMSK));
		/* set the system device as awake now that we're back from sleep */
		power_level = AWAKE_STATE;
		device->callback(device, AU1XXX_PM_WAKEUP, &power_level);
	} else if (request == AU1XXX_PM_WAKEUP) {
		power_level = *((int*)data);
		/* send all valid devices a wakeup call */
		for (i = 1; i < MAX_SUPPORTED_DEVICES; i++) {
			if (sys_power_pointers[i]->cur_state == AWAKE_STATE) {
				DPRINTK("Error: Device: %s is already awake!\n", device->name);
			} else if (sys_power_pointers[i]->valid) { 
				DPRINTK("Awakening Device: %s\n", sys_power_pointers[i]->name);
				sys_power_pointers[i]->prev_state = sys_power_pointers[i]->cur_state;
				sys_power_pointers[i]->prev_state_timestamp = sys_power_pointers[i]->cur_state_timestamp;
				sys_power_pointers[i]->cur_state = AWAKE_STATE;
				do_gettimeofday(&(sys_power_pointers[i]->cur_state_timestamp));
				sys_power_pointers[i]->callback(sys_power_pointers[i],
					AU1XXX_PM_WAKEUP, &power_level);
			}
		}
		device->prev_state = device->cur_state;
		device->prev_state_timestamp = device->cur_state_timestamp;
		do_gettimeofday(&(device->cur_state_timestamp));
		device->cur_state = AWAKE_STATE;
	} else {
		DPRINTK("Error: Invalid request - %d\n", (int) request);
	}
	DPRINTK("Leaving system_power_callback\n");
	return retval;
}

int test_callback(au1xxx_power_dev_t *device, au1xxx_request_t request,
	void *data)
{
	static int my_power_level;

	DPRINTK("Inside the test_callback function.\n");
	if (request == AU1XXX_PM_WAKEUP) {
		DPRINTK("%s has been told to wake up.\n", device->name);
		my_power_level = *((int *)data);
	} else if (request == AU1XXX_PM_GETSTATUS) {
		return my_power_level;
	} else if (request == AU1XXX_PM_SLEEP) {
		my_power_level = SLEEP_STATE;
	}

	return 0;
}


/* 
	This function was heavily leveraged from kmod.c.
*/
int	pm_catchall_callback( au1xxx_power_dev_t *dev, au1xxx_request_t request, void *power_level )
{
	pm_module *pm = dev->data; 
	int ret=0;

	char *argv[] = { modprobe_path_pm, "-q", "add_remove", "empty_name", NULL };
	static char *remove="-r";
	static char *add="-a";
	static char *envp[] = { "HOME=/",
				"TERM=linux",
				"PATH=/sbin:/usr/sbin:/bin:/usr/bin",
				NULL };

	/* Make sure our argv module name is our current module's name */
	argv[3] = pm->modname;
	switch ( request )
	{
		case AU1XXX_PM_SLEEP:
			argv[2] = remove;
			ret = call_usermodehelper(modprobe_path_pm, argv, envp, 1);
			break;

		case AU1XXX_PM_WAKEUP:
			argv[2] = add;
			ret = call_usermodehelper(modprobe_path_pm, argv, envp, 1);
			break;

		case AU1XXX_PM_ACCESS:
			if (dev->cur_state != AWAKE_STATE) {
				/* do nothing */
			}
			else
				ret = -1;
			break;

		case AU1XXX_PM_GETSTATUS:
			ret = dev->cur_state;
			break;

		default:
		case AU1XXX_PM_IDLE:
		case AU1XXX_PM_CLEANUP:
			break;
	}

	return ret;
}


/***********************************************************************
 *                     PROC INTERFACE FUNCTIONS
 ***********************************************************************/
static int proc_read(char *page, char **start, off_t off, 
	int count, int *eof, void *data)
{
	int size = 0;
	int i;
	au1xxx_power_dev_t *power_dev = (au1xxx_power_dev_t *)data;
	struct au1xxx_pm_bin_header bin_hdr;
	au1xxx_pm_bin_header_t *bin_header = &bin_hdr;

	if (!power_dev->valid) {
		printk(KERN_ERR "Kernel Error: Attempted to read from an invalid power device.\n");
		return -EFAULT;
	}

	if (au1xxx_pm_setup->feedback_mode == ASCII_FEEDBACK) {
		/* print out the info on the current device */
		size = sprintf((page + size), "Device Name: %s\nPower State: %d\nDevice Status: %d\nPending Transitions: %d\nError Code: %d\n", 
			power_dev->name, 
			power_dev->cur_state, 
			power_dev->callback(power_dev, AU1XXX_PM_GETSTATUS, NULL),
			power_dev->state_change_pending,
			power_dev->last_error);
	
		/* print all the devices' info if reading the system entry */
		if (power_dev == system_dev) {
			for (i = 1; i < MAX_SUPPORTED_DEVICES; i++) {
				if (sys_power_pointers[i]->valid) {
					size += sprintf((page + size), "-----------------------------\n");
					size += sprintf((page + size), "Device Name: %s\nPower State: %d\nDevice Status: %d\nPending Transitions: %d\nError Code: %d\n", 
						sys_power_pointers[i]->name, 
						sys_power_pointers[i]->cur_state, 
						sys_power_pointers[i]->callback(sys_power_pointers[i], 
							AU1XXX_PM_GETSTATUS, NULL),
						sys_power_pointers[i]->state_change_pending,
						sys_power_pointers[i]->last_error);
				}
			}
		}

		/* handle the buffer */
  	if (size <= off + count) 
			*eof = 1;
  	*start = page + off;
  	size -= off;
  	if (size > count) 
			size = count;
  	if (size < 0) 
			size = 0;

	} else if (au1xxx_pm_setup->feedback_mode == BINARY_FEEDBACK) {
		DPRINTK("Starting BINARY_FEEDBACK output method\n");

		if (power_dev != system_dev) {
			/* copy a header for a single binary report */
			bin_header->report_type = AU1XXX_PM_DEVICE;
			bin_header->num_reports = 1;
			size = sizeof(au1xxx_pm_bin_header_t);
			memcpy(page, bin_header, size);

			/* copy the actual report */
			memcpy((page + size), power_dev, sizeof(au1xxx_pm_public_dev_t));
			size += sizeof(au1xxx_pm_public_dev_t);

			DPRINTK("report_type Size: %d\t\tValue: %d\n", sizeof(bin_header->report_type), (int) bin_header->report_type);
			DPRINTK("num_reports Size: %d\t\tValue: %d\n", sizeof(bin_header->num_reports), (int) bin_header->num_reports);
			DPRINTK("Total size returned: %d\n", size);
		} else {
			/* copy a header for a full system report */
			bin_header->report_type = AU1XXX_PM_DEVICE;
			bin_header->num_reports = MAX_SUPPORTED_DEVICES;
			size = sizeof(au1xxx_pm_bin_header_t);
			memcpy(page, bin_header, size);

			/* copy the full system report */
			for (i = 0; i < MAX_SUPPORTED_DEVICES; i++) {
				memcpy((page + size), sys_power_pointers[i], sizeof(au1xxx_pm_public_dev_t));
				size += sizeof(au1xxx_pm_public_dev_t);
			}
		}
	} else {
		/* reset feedback mode in the error case */
		printk(KERN_WARNING "Invalid power management feedback setting; resetting to default.\n");
		spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
		au1xxx_pm_setup->feedback_mode = DEFAULT_FEEDBACK;
		spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);
		return -EFAULT;
	}

	return size;
}

static int proc_write(struct file *file, const char *buffer, 
	unsigned long count, void *data)
{
	int len;
	int callback_ret_val;
	char data_in[MAX_POWER_CONF_LEN];
	au1xxx_power_dev_t *power_dev = (au1xxx_power_dev_t *)data;
	int power_level;

	DPRINTK("Entering proc_write\n");

	if (!power_dev->valid) {
		printk(KERN_ERR "Kernel Error: Attempted write to an invalid power device.\n");
		return -EFAULT;
	}

	if (count > MAX_POWER_CONF_LEN)
		len = MAX_POWER_CONF_LEN;
	else
		len = count;

	if (copy_from_user(data_in, buffer, len))
		return -EFAULT;

	if (au1xxx_pm_setup->feedback_mode == ASCII_FEEDBACK) {
		power_level = simple_strtoul(data_in, NULL, 0);
		DPRINTK("Used ASCII_FEEDBACK to get power_level of %d\n", power_level);
	} else if (au1xxx_pm_setup->feedback_mode == BINARY_FEEDBACK) {
		memcpy(&power_level, data_in, sizeof(int));
		DPRINTK("Used BINARY_FEEDBACK to get power_level of %d\n", power_level);
	} else {
		/* reset feedback mode in the error case */
		printk(KERN_WARNING "Invalid power management feedback setting; resetting to default.\n");
		spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
		au1xxx_pm_setup->feedback_mode = DEFAULT_FEEDBACK;
		spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);
		return -EFAULT;
	}

	spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
	if (power_level != SLEEP_STATE) {
		power_dev->prev_state = power_dev->cur_state;
		power_dev->prev_state_timestamp = power_dev->cur_state_timestamp;
		do_gettimeofday(&(power_dev->cur_state_timestamp));
		power_dev->cur_state = power_level;  
		power_dev->state_change_pending = TRUE;
		DPRINTK("Waking up %s device.\n", power_dev->name);
		callback_ret_val = power_dev->callback(power_dev, AU1XXX_PM_WAKEUP, &power_level);
		power_dev->state_change_pending = FALSE;
		power_dev->last_error = callback_ret_val;
	} else {
		power_dev->prev_state = power_dev->cur_state;
		power_dev->prev_state_timestamp = power_dev->cur_state_timestamp;
		do_gettimeofday(&(power_dev->cur_state_timestamp));
		power_dev->cur_state = power_level;
		power_dev->state_change_pending = TRUE;
		DPRINTK("Putting %s device to sleep.\n", power_dev->name);
		callback_ret_val = power_dev->callback(power_dev, AU1XXX_PM_SLEEP, &power_level);
		power_dev->state_change_pending = FALSE;
		power_dev->last_error = callback_ret_val;
	}
	spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);

	return len;
}

static int proc_read_settings(char *page, char **start, off_t off, 
	int count, int *eof, void *data)
{
	int size = 0;

	if ((!au1xxx_pm_setup->valid) || 
		(au1xxx_pm_setup != (au1xxx_pm_settings_t *)data))
	{
		printk(KERN_ERR "Kernel Error: Power Management settings are invalid.\n");
		return -EFAULT;
	}

	if (au1xxx_pm_setup->feedback_mode == ASCII_FEEDBACK) {
		size = sprintf((page + size), "Au1xxx Power Management Settings:\n");
		size += sprintf((page + size), "=================================\n");

		size += sprintf((page + size), "Current wakeup method is: ");
		if (au1xxx_pm_setup->wakeup_mode == GPIO_WAKEUP)  
			size += sprintf((page + size), "GPIO 5 controlled.\n");
		else if (au1xxx_pm_setup->wakeup_mode == TIMER_WAKEUP) {
			size += sprintf((page + size), "TOY timer controlled.\n");
			size += sprintf((page + size), "  Current sleep time: %d seconds.\n", 
					au1xxx_pm_setup->sleep_time);
		}
		else
			size += sprintf((page + size), "invalid.\n");
	
		size += sprintf((page + size), "Current feedback method is: ");
		if (au1xxx_pm_setup->feedback_mode == ASCII_FEEDBACK)
			size += sprintf((page + size), "ASCII text file.\n");
		else if (au1xxx_pm_setup->feedback_mode == BINARY_FEEDBACK)
			size += sprintf((page + size), "Binary file.\n");
		else
			size += sprintf((page + size), "invalid.\n");

		/* handle the buffer */
  	if (size <= off + count) 
			*eof = 1;
  	*start = page + off;
  	size -= off;
  	if (size > count) 
			size = count;
  	if (size < 0) 
			size = 0;
	} else if (au1xxx_pm_setup->feedback_mode == BINARY_FEEDBACK) {
		DPRINTK("Attempting to read settings via BINARY_FEEDBACK method.\n");
		memcpy((page + size), &au1xxx_pm_setup->wakeup_mode, sizeof(int));
		size += sizeof(int);
		memcpy((page + size), &au1xxx_pm_setup->sleep_time, size);
		size += sizeof(int);
		memcpy((page + size), &au1xxx_pm_setup->feedback_mode, size);
		size += sizeof(int);
	} else {
		/* reset feedback mode in the error case */
		printk(KERN_WARNING "Invalid power management feedback setting; resetting to default.\n");
		spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
		au1xxx_pm_setup->feedback_mode = DEFAULT_FEEDBACK;
		spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);
		return -EFAULT;
	}

	return size;
}

static int proc_write_settings(struct file *file, const char *buffer, 
	unsigned long count, void *data)
{
	int len, wakeup_int, feedback_int;
	unsigned int sleep_int;
	int size = 0;
	char data_in[MAX_POWER_CONF_LEN];
	struct { 
		unsigned int wakeup_uns;
		unsigned int sleep_uns;
		unsigned int feedback_uns;
	} binary_reader;

	DPRINTK("Entering proc_write_settings\n");

	if ((!au1xxx_pm_setup->valid) ||
		(au1xxx_pm_setup != (au1xxx_pm_settings_t *)data))
	{
		printk(KERN_ERR "Kernel Error: Power Management settings are invalid.\n");
		return -EFAULT;
	}

	if (count > MAX_POWER_CONF_LEN)
		len = MAX_POWER_CONF_LEN;
	else
		len = count;

	DPRINTK("Doing copy_from_user with len: %d\n", len);
	if (copy_from_user(data_in, buffer, len))
		return -EFAULT;

	/* read in the thre values according to the feedback mode */
	if (au1xxx_pm_setup->feedback_mode == ASCII_FEEDBACK) {
		if (sscanf(data_in, "%d:%d, %u\n", &wakeup_int, &sleep_int, 
					&feedback_int) != 3) {
			printk(KERN_ERR "Kernel Error: Cannot read new power management settings.\n");
			return -EFAULT;
		}
	} else if (au1xxx_pm_setup->feedback_mode == BINARY_FEEDBACK) {
		DPRINTK("Attempting to write settings via BINARY_FEEDBACK method.\n");
		/* For some reason it doesn't work when you try and do this piece by piece
		 * so I've simply created a structure to do it all at once. */
		memcpy(&binary_reader, &data_in, sizeof(binary_reader));
		size = sizeof(binary_reader);
		wakeup_int = (int) binary_reader.wakeup_uns;
		sleep_int = (int) binary_reader.sleep_uns;
		feedback_int = (int) binary_reader.feedback_uns;
	} else {
		/* reset feedback mode in the error case */
		printk(KERN_WARNING "Invalid power management feedback setting; resetting to default.\n");
		spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
		au1xxx_pm_setup->feedback_mode = DEFAULT_FEEDBACK;
		spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);
		return -EFAULT;
	}
	
	/* actually record the changes */
	spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
	if (wakeup_int == GPIO_WAKEUP)
		au1xxx_pm_setup->wakeup_mode = GPIO_WAKEUP;
	else if (wakeup_int == TIMER_WAKEUP) {
		au1xxx_pm_setup->wakeup_mode = TIMER_WAKEUP;
		au1xxx_pm_setup->sleep_time = (int) sleep_int;
	}
	else {
		au1xxx_pm_setup->wakeup_mode = DEFAULT_WAKEUP;
		printk(KERN_WARNING "Invalid power management wakeup mode; resetting to default.\n");
	}

	if (feedback_int == ASCII_FEEDBACK)
		au1xxx_pm_setup->feedback_mode = ASCII_FEEDBACK;
	else if (feedback_int == BINARY_FEEDBACK)
		au1xxx_pm_setup->feedback_mode = BINARY_FEEDBACK;
	else {
		au1xxx_pm_setup->feedback_mode = DEFAULT_FEEDBACK;
		printk(KERN_WARNING "Invalid power management feedback mode; resetting to default.\n");
	}
	spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);

	return len;
}

static int proc_read_timer(char *page, char **start, off_t off, 
	int count, int *eof, void *data)
{
	int size = 0;
	au1xxx_power_dev_t *dev = (au1xxx_power_dev_t *)data;
	au1xxx_pm_bin_header_t my_bin_hdr;

	if (au1xxx_pm_setup->feedback_mode == ASCII_FEEDBACK) {
		size += sprintf((page + size), "Timeout: %lu\n", dev->timeout);

		/* handle the buffer */
  	if (size <= off + count) 
			*eof = 1;
  	*start = page + off;
  	size -= off;
  	if (size > count) 
			size = count;
  	if (size < 0) 
			size = 0;
	} else if (au1xxx_pm_setup->feedback_mode == BINARY_FEEDBACK) {
		my_bin_hdr.report_type = AU1XXX_PM_TIMEOUT;
		my_bin_hdr.num_reports = 1;
		memcpy((page + size), &my_bin_hdr, sizeof(au1xxx_pm_bin_header_t));
		size += sizeof(au1xxx_pm_bin_header_t);
		memcpy((page + size), &dev->timeout, sizeof(unsigned long));
		size += sizeof(unsigned long);
	} else {
		/* reset feedback mode in the error case */
		printk(KERN_WARNING "Invalid power management feedback setting; resetting to default.\n");
		spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
		au1xxx_pm_setup->feedback_mode = DEFAULT_FEEDBACK;
		spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);
		return -EFAULT;
	}

	return size;
}

static int proc_write_timer(struct file *file, const char *buffer, 
	unsigned long count, void *data)
{
	int len;
	unsigned long timeout_val = 0;
	int size = 0;
	char data_in[MAX_POWER_CONF_LEN];
	au1xxx_power_dev_t *dev = (au1xxx_power_dev_t *)data;

	DPRINTK("Entering proc_write_timer\n");
	
	if (count > MAX_POWER_CONF_LEN)
		len = MAX_POWER_CONF_LEN;
	else
		len = count;

	DPRINTK("Doing copy_from_user with len: %d\n", len);
	if (copy_from_user(data_in, buffer, len))
		return -EFAULT;

	/* read in the value */
	if (au1xxx_pm_setup->feedback_mode == ASCII_FEEDBACK) {
		if (sscanf(data_in, "%lu\n", &timeout_val) != 1) {
			printk(KERN_ERR "Kernel Error: Cannot read new power management settings.\n");
			return -EFAULT;
		}
	} else if (au1xxx_pm_setup->feedback_mode == BINARY_FEEDBACK) {
		DPRINTK("Attempting to write settings via BINARY_FEEDBACK method.\n");
		/* For some reason it doesn't work when you try and do this piece by piece
		 * so I've simply created a structure to do it all at once. */
		memcpy(&timeout_val, &data_in, sizeof(unsigned long));
		size = sizeof(unsigned long);
	} else {
		/* reset feedback mode in the error case */
		printk(KERN_WARNING "Invalid power management feedback setting; resetting to default.\n");
		spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
		au1xxx_pm_setup->feedback_mode = DEFAULT_FEEDBACK;
		spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);
		return -EFAULT;
	}

	/* convert the input from seconds into CPU ticks */
	timeout_val = (unsigned long) (timeout_val * HZ);
	
	/* actually make the changes */
	spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
	if ((timeout_val) && (dev->timeout)) {
		dev->timeout = timeout_val;
		mod_timer(&dev->timer, jiffies + dev->timeout);
	} else if ((!timeout_val) && (dev->timeout)) {
		dev->timeout = timeout_val;
		del_timer(&dev->timer);
	} else if ((timeout_val) && (!dev->timeout)) {
		dev->timeout = timeout_val;
		init_timer(&dev->timer);
		dev->timer.function = &au1xxx_pm_service_timer;
		dev->timer.expires = jiffies + dev->timeout;
		dev->timer.data = (unsigned long)dev;
		add_timer(&dev->timer);
	} else if ((!timeout_val) && (!dev->timeout)) {
		/* do nothing in this case, as the timeout is still nothing */
	}
	spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);

	return len;
}

static int proc_read_pending(char *page, char **start, off_t off, 
	int count, int *eof, void *data)
{
	int current_pending = 0;
	int size = 0;
	int i;
	au1xxx_pm_bin_header_t my_bin_hdr;

	/* add up all the pending flags for registered devices */
	for (i = 0; i < MAX_SUPPORTED_DEVICES; i++) {
		if (sys_power_pointers[i]->valid)
			current_pending += sys_power_pointers[i]->state_change_pending;
	}

	if (au1xxx_pm_setup->feedback_mode == ASCII_FEEDBACK) {
		size += sprintf((page + size), "Current pending transactions: %d\n", 
			current_pending);

		/* handle the buffer */
  	if (size <= off + count) 
			*eof = 1;
  	*start = page + off;
  	size -= off;
  	if (size > count) 
			size = count;
  	if (size < 0) 
			size = 0;
	} else if (au1xxx_pm_setup->feedback_mode == BINARY_FEEDBACK) {
		my_bin_hdr.report_type = AU1XXX_PM_PENDING;
		my_bin_hdr.num_reports = 1;
		memcpy((page + size), &my_bin_hdr, sizeof(au1xxx_pm_bin_header_t));
		size += sizeof(au1xxx_pm_bin_header_t);
		memcpy((page + size), &current_pending, sizeof(int));
		size += sizeof(int);
	} else {
		/* reset feedback mode in the error case */
		printk(KERN_WARNING "Invalid power management feedback setting; resetting to default.\n");
		spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
		au1xxx_pm_setup->feedback_mode = DEFAULT_FEEDBACK;
		spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);
		return -EFAULT;
	}

	return size;
}

static int proc_read_error(char *page, char **start, off_t off, 
	int count, int *eof, void *data)
{
	int size = 0;
	int i;
	au1xxx_pm_error_t *error = (au1xxx_pm_error_t *)data;
	au1xxx_pm_bin_header_t my_bin_hdr;

	if (au1xxx_pm_setup->feedback_mode == ASCII_FEEDBACK) {
		/* print all error information */

		if (!error->dev) {
			size += sprintf((page + size), "There are no errors registered with the power management system.\n");
		}
		else {
			for (i = 0; i < MAX_SUPPORTED_ERRORS; i++) {
				if (sys_error_pointers[i]->dev) {
					size += sprintf((page + size), "-----------------------------\n");
					size += sprintf((page + size), "Device Name: %s\nError Code: %d\nOccurances: %d\nFirst Occurance: %d\nLatest Occurance: %d\n",
						sys_error_pointers[i]->dev->name,
						sys_error_pointers[i]->error_code,
						sys_error_pointers[i]->error_count,
						(int) sys_error_pointers[i]->first_timestamp.tv_sec,
						(int) sys_error_pointers[i]->last_timestamp.tv_sec);
				}
			}
		}
		/* handle the buffer */
  	if (size <= off + count) 
			*eof = 1;
  	*start = page + off;
  	size -= off;
  	if (size > count) 
			size = count;
  	if (size < 0) 
			size = 0;
	} else if (au1xxx_pm_setup->feedback_mode == BINARY_FEEDBACK) {
		/* output all error info in binary format */
		for (i = 0; i < MAX_SUPPORTED_ERRORS; i++) {
			memcpy((page + size), &my_bin_hdr, sizeof(au1xxx_pm_bin_header_t));
			size += sizeof(au1xxx_pm_bin_header_t);
			memcpy((page + size), sys_error_pointers[i], sizeof(au1xxx_pm_error_t));
			size += sizeof(au1xxx_pm_error_t);
		}
	} else {
		/* reset feedback mode in the error case */
		printk(KERN_WARNING "Invalid power management feedback setting; resetting to default.\n");
		spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
		au1xxx_pm_setup->feedback_mode = DEFAULT_FEEDBACK;
		spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);
		return -EFAULT;
	}
	
	/* clear all error information */
	spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
	for (i = 0; i < MAX_SUPPORTED_ERRORS; i++) {
		remove_au1xxx_pm_error(sys_error_pointers[i]);
	}
	spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);

	return size;
}


/***********************************************************************
 *                     PM DEVICE FUNCTIONS
 ***********************************************************************/
au1xxx_power_dev_t *new_au1xxx_power_device(const char *name, 
		au1xxx_power_callback cb_func, void *data)
{
	int i;
	au1xxx_power_dev_t *new_dev = NULL;

	DPRINTK("Entering new_au1xxx_power_device()\n");

	spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);

	/* find the first free dev node */
	for (i = 0; i < MAX_SUPPORTED_DEVICES; i++) {
		if (sys_power_pointers[i]->valid == 0) {
			new_dev = sys_power_pointers[i];
			break;
		}
	}

	if (new_dev == NULL) {
		printk(KERN_ERR "Kernel Error: No memory for a new power device.\n");
		return NULL;
	}

	/* initialze the new device node */
	strncpy(new_dev->name, name, MAX_NAME_WIDTH);
	new_dev->callback = *cb_func;
	new_dev->valid = 1;
	new_dev->prev_state = SLEEP_STATE;
	do_gettimeofday(&(new_dev->prev_state_timestamp));
	new_dev->cur_state = AWAKE_STATE;
	do_gettimeofday(&(new_dev->cur_state_timestamp));
	new_dev->data = data;
	new_dev->timeout = 0;
	
	/* Create a /proc/pm entry for both read and write if the system has loaded
	 * the proc_fs driver. */
	if (system_dev->proc_entry != NULL) {
		DPRINTK("Creating a /proc/pm entry for the new device.\n");
		new_dev->proc_entry = create_proc_entry(name, 0644, power_dir);
		if (new_dev->proc_entry == NULL) 
			goto no_new_proc_entry;
		new_dev->proc_entry->data = new_dev;
		new_dev->proc_entry->read_proc = proc_read;
		new_dev->proc_entry->write_proc = proc_write;
	} else {
		/* This is not an error condition.  It is the case that occurs when a
		 * driver loads a power management device before the power management
		 * driver itself is loaded. */
		new_dev->proc_entry = NULL;
	}

	/* Create a /proc/pm/timeout entry for both read and write if the system has
	 * loaded the proc_fs driver. */

	if (system_dev->timer_proc_entry != NULL) {
		DPRINTK("Creating a /proc/pm/timeout entry for the new device.\n");
		new_dev->timer_proc_entry = create_proc_entry(name, 0644, timeout_dir);
		if (new_dev->timer_proc_entry == NULL) 
			goto no_new_timer_proc_entry;
		new_dev->timer_proc_entry->data = new_dev;
		new_dev->timer_proc_entry->read_proc = proc_read_timer;
		new_dev->timer_proc_entry->write_proc = proc_write_timer;
	} else {
		/* This is not an error condition.  It is the case that occurs when a
		 * driver loads a power management device before the power management
		 * driver itself is loaded. */
		new_dev->timer_proc_entry = NULL;
	}

	spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);
	
	DPRINTK("Returning from new_au1xxx_power_device() with a new device.\n");
	return new_dev;

	remove_proc_entry(new_dev->name, timeout_dir);
no_new_timer_proc_entry:
	remove_proc_entry(new_dev->name, power_dir);
no_new_proc_entry:

	printk(KERN_WARNING "Unable to create a /proc entry for a new power management device.\n");
	return NULL;
}

int remove_au1xxx_power_device(au1xxx_power_dev_t *dev) {
	int retval = 0;

	spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);

	if ((system_dev->proc_entry == NULL) || (dev == NULL)) {
		retval = ERROR_INVALID_DEVICE;
		printk(KERN_WARNING "Power management error: accessing invalid device.\n");
		return retval;
	}

	/* clear the device from the array */
	dev->valid = 0;
	strncpy(dev->name, "", strlen(""));
	dev->callback = NULL;
	dev->cur_state = SLEEP_STATE;
	do_gettimeofday(&(dev->cur_state_timestamp));
	dev->prev_state = SLEEP_STATE;
	do_gettimeofday(&(dev->prev_state_timestamp));
	dev->data = NULL;
	dev->timeout = 0;

	/* remove the proc entries if they exist */
	if (dev->proc_entry) 
		remove_proc_entry(dev->name, power_dir);
	if (dev->timer_proc_entry) 
		remove_proc_entry(dev->name, timeout_dir);

	spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);

	return retval;
}

au1xxx_pm_error_t *new_au1xxx_pm_error(int error_code, au1xxx_power_dev_t *dev) {
	au1xxx_pm_error_t *ret_error_t = NULL;
	int i;

	if (dev == NULL) {
		ret_error_t = NULL;
		printk(KERN_ERR "Invalid Au1xxx power management device access.\n");
		return ret_error_t;
	}

	/* find a matching error structure or a free error struct */
	for (i = 0; i < MAX_SUPPORTED_ERRORS; i++) {
		if (sys_error_pointers[i]->dev == dev)
			break;
		/* each device is allowed one error structure, so if it has an allocated
		 * structre, return it */
		else if (sys_error_pointers[i]->dev == NULL)
			break;
	}
	if (i == MAX_SUPPORTED_ERRORS) {
		ret_error_t = NULL;
		printk(KERN_ERR "No available resources for power management error reporting.\n");
		return ret_error_t;
	}

	/* update the error structure */
	if ((sys_error_pointers[i]->dev == dev) &&
		(sys_error_pointers[i]->error_code == error_code)) {
		sys_error_pointers[i]->error_count++;
		do_gettimeofday(&(sys_error_pointers[i]->last_timestamp));
	}
	/* the 'else' case handles new error structure as well as devices
	 * experiencing a new error */
	else {
		sys_error_pointers[i]->dev = dev;
		sys_error_pointers[i]->error_code = error_code;
		sys_error_pointers[i]->error_count = 1;
		do_gettimeofday(&(sys_error_pointers[i]->first_timestamp));
		do_gettimeofday(&(sys_error_pointers[i]->last_timestamp));
	}

	return ret_error_t;
}

int remove_au1xxx_pm_error(au1xxx_pm_error_t *error) {
	int retval = 0;

	if (error == NULL) {
		retval = -EFAULT;
		printk("Invalid Au1xxx power management error access.\n");
		return retval;
	}

	error->dev = NULL;
	error->error_code = 0;
	error->error_count = 0;
	error->first_timestamp = zero_timeval;
	error->last_timestamp = zero_timeval;

	return retval;
}

int au1xxx_pm_access(au1xxx_power_dev_t *dev) {
	int retval = 0;

	if ((system_dev->proc_entry == NULL) || 
			(dev == NULL) || 
			!system_dev->valid) {
		retval = ERROR_INVALID_DEVICE;
		printk(KERN_WARNING "Power management error: accessing invalid device.\n");
		return retval;
	}

	spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
	/* turn the device on if it's not already */
	if (!dev->cur_state) {
		dev->prev_state = dev->cur_state;
		dev->prev_state_timestamp = dev->cur_state_timestamp;
		dev->cur_state = AWAKE_STATE;
		do_gettimeofday(&(dev->cur_state_timestamp));
		retval = dev->callback(dev, AU1XXX_PM_ACCESS, NULL);
	}
	spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);

	/* reset the timeout, if enabled */
	if (dev->timeout)
		mod_timer(&dev->timer, jiffies + dev->timeout);

	/* always reset the system device's timeout, if enabled */
	if (system_dev->timeout)
		mod_timer(&system_dev->timer, jiffies + system_dev->timeout);

	/* handle errors */
	if (retval) 
		new_au1xxx_pm_error(retval, dev);

	/* setup locking/timed/discrete event sim system for keeping hardware on */
	/* maybe just disable interrupts? */
	
	return retval;
}

int au1xxx_pm_idle(au1xxx_power_dev_t *dev) {
	int retval = 0;

	/* return a warning if the main driver isn't loaded */
	if ((system_dev->proc_entry == NULL) || 
			(dev == NULL) || 
			!system_dev->valid) {
		retval = ERROR_INVALID_DEVICE;
		printk(KERN_WARNING "Power management error: accessing invalid device.\n");
		return retval;
	}

	spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);
	/* turn the device off if it's currently on */
	if (dev->cur_state) {
		dev->prev_state = dev->cur_state;
		dev->prev_state_timestamp = dev->cur_state_timestamp;
		dev->cur_state = SLEEP_STATE;
		do_gettimeofday(&(dev->cur_state_timestamp));
		retval = dev->callback(dev, AU1XXX_PM_IDLE, NULL);
	}
	spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);

	/* reset the timeout, if enabled */
	if (dev->timeout)
		mod_timer(&dev->timer, jiffies + dev->timeout);

	/* always reset the system device's timeout, if enabled */
	if (system_dev->timeout)
		mod_timer(&system_dev->timer, jiffies + system_dev->timeout);	

	/* handle errors */
	if (retval) 
		new_au1xxx_pm_error(retval, dev);

	return retval;
}

static void au1xxx_pm_service_timer(unsigned long var) { 
	int power_level = SLEEP_STATE;
	au1xxx_power_dev_t *dev = (au1xxx_power_dev_t *)var;

	/* handle the timeout */
	if (dev->cur_state != SLEEP_STATE)
		dev->callback(dev, AU1XXX_PM_SLEEP, &power_level);	

	/* reinstall timer */
	if (dev->timeout)
		mod_timer(&dev->timer, jiffies + dev->timeout);

	return;
}

int au1xxx_pm_set_timer(au1xxx_power_dev_t *dev, unsigned long timeout_val) { 
	int retval = 0;

	if (dev == NULL) {
		retval = -EFAULT;
	} else if ((timeout_val) && (dev->timeout)) {
		/* modify the existing timer */
		dev->timeout = timeout_val;
		mod_timer(&dev->timer, jiffies + dev->timeout);
	} else if ((!timeout_val) && (dev->timeout)) {
		/* delete the existing timer */
		dev->timeout = timeout_val;
		del_timer(&dev->timer);
	} else if ((timeout_val) && (!dev->timeout)) {
		/* install the timer if there was no previous timer */
		dev->timeout = timeout_val;
		init_timer(&dev->timer);
		dev->timer.function = &au1xxx_pm_service_timer;
		dev->timer.expires = jiffies + dev->timeout;
		dev->timer.data = (unsigned long)dev;
		add_timer(&dev->timer);
	} else if ((!timeout_val) && (!dev->timeout)) {
		/* do nothing in this case, as the timeout is still nothing */
	}

	return retval;
}

int au1xxx_pm_reset_timer(au1xxx_power_dev_t *dev) { 
	int retval = 0;

	if (dev == NULL) {
		retval = -EFAULT;
	} else if (dev->timeout) {
		/* reset the existing timer */
		mod_timer(&dev->timer, jiffies + dev->timeout);
	} else if (!dev->timeout) {
		/* do nothing in this case, as the timeout is nothing */
	}

	return retval;
}

int au1xxx_pm_del_timer(au1xxx_power_dev_t *dev) { 
	int retval = 0;

	if (dev == NULL) {
		retval = -EFAULT;
	} else if (dev->timeout == 0) {
		/* do nothing */
	} else {
		retval = del_timer(&dev->timer);
	}

	return retval;
}
/***********************************************************************
 *                     INITIALIZE POWER INTERFACE
 ***********************************************************************/
/* PM Catch All Init funtion */

void pm_catchall_pm_init(void)
{
	int i;

	printk("Power Managed Modules Support:\n");
	for(i=0; i<num_modules; i++) {
		pm_module *mod;
	
		mod = &pm_modules[i];
		printk("      %s %p\n", mod->modname, mod);
	    mod->dev = new_au1xxx_power_device( mod->pmname, pm_catchall_callback, (void*)mod );
		if (mod->dev==NULL){printk("Error: Module Power Management failed to initialize \n");}
		/*Assigning Sleep State as current state*/
	     mod->dev->cur_state=SLEEP_STATE;
	     mod->dev->prev_state=AWAKE_STATE;
	}
	  	   	     
	return;
}


/* The structures are layered so that there's an 'external' and an 'internal'
 * structure in the code.  The external structure is really just the first
 * part of the internal structure so you can cast an internal structure as an
 * external structure and basically mask off the parts that should remain
 * internal to the driver.
 */
int au1xxx_pm_struct_init(void)
{
	int i;
	static int initialized = 0;
	int retval = 0;

	/* check to see if structures are already initialized */
	if (initialized)
		return -EFAULT;
	else
		initialized = 1;

	spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);

	/* setup system device array */
	for (i = 0; i < MAX_SUPPORTED_DEVICES; i++) {
		sys_power_pointers[i] = &sys_power_array[i];
		strncpy(sys_power_pointers[i]->name, "NULL", strlen(""));
		sys_power_pointers[i]->callback = NULL;
		sys_power_pointers[i]->proc_entry = NULL;
		sys_power_pointers[i]->valid = 0;
		sys_power_pointers[i]->state_change_pending = 0;
		sys_power_pointers[i]->last_error = 0;
		sys_power_pointers[i]->cur_state = SLEEP_STATE;
		do_gettimeofday(&(sys_power_pointers[i]->cur_state_timestamp));
		sys_power_pointers[i]->prev_state = SLEEP_STATE;
		do_gettimeofday(&(sys_power_pointers[i]->prev_state_timestamp));
		sys_power_pointers[i]->data = NULL;
		sys_power_pointers[i]->timeout = 0;
		sys_power_pointers[i]->timer_proc_entry = NULL;
	}
	DPRINTK("sys_power_pointers array is now valid.\n");

	/* setup error array */
	for (i = 0; i < MAX_SUPPORTED_ERRORS; i++) {
		sys_error_pointers[i] = &sys_error_array[i];
		sys_error_pointers[i]->dev = NULL;
		sys_error_pointers[i]->error_code = 0;
		sys_error_pointers[i]->error_count = 0;
		sys_error_pointers[i]->first_timestamp = zero_timeval;
		sys_error_pointers[i]->last_timestamp = zero_timeval;
	}
	
	/* setup the system device structure */
	system_dev->valid = 1;
	system_dev->state_change_pending = 0;
	system_dev->last_error = 0;
	strncpy(system_dev->name, "system", MAX_NAME_WIDTH);
	system_dev->callback = &system_power_callback;
	system_dev->cur_state = AWAKE_STATE;
	do_gettimeofday(&(system_dev->cur_state_timestamp));
	system_dev->proc_entry = NULL;
	system_dev->data = NULL;
	system_dev->timeout = SYSTEM_TIMEOUT;

	au1xxx_pm_setup->valid = 0;
	au1xxx_pm_setup->wakeup_mode = DEFAULT_WAKEUP;
	au1xxx_pm_setup->sleep_time = 0;
	au1xxx_pm_setup->feedback_mode = DEFAULT_FEEDBACK;

	zero_timeval.tv_sec = 0;
	zero_timeval.tv_usec = 0;

	spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);

	au1xxx_pm_debug();

	return retval;
}

static int au1xxx_pm_proc_init(void) {
	int i = 0;
	int retval = 0;
#if ENABLE_TEST_DEV
	au1xxx_power_dev_t *test_dev;
#endif

	/* create power folder in /proc file system */
	DPRINTK("Attempting to create /proc/pm file system entires.\n");
	power_dir = proc_mkdir("pm", NULL);
	if (power_dir == NULL) {
		retval = -ENOMEM;
		goto no_power_dir;
	} else { DPRINTK("power_dir successfully created!\n"); }

	/* create the timeout folder in the /proc/pm/ directory */
	DPRINTK("Attempting to create /proc/pm file system entires.\n");
	timeout_dir = proc_mkdir("timeout", power_dir);
	if (timeout_dir == NULL) {
		retval = -ENOMEM;
		goto no_timeout_dir;
	} else { DPRINTK("timeout_dir successfully created!\n"); }

	spin_lock_irqsave(&au1xxx_pm_lock, spin_lock_flags);

	/* create the proc entry for both read and write for the system_dev */
	system_dev->proc_entry = create_proc_entry("system", 0644, power_dir);
	if (system_dev->proc_entry == NULL) {
		retval = -ENOMEM;
		goto no_system_power;
	}
	system_dev->proc_entry->data = system_dev;
	system_dev->proc_entry->read_proc = proc_read;
	system_dev->proc_entry->write_proc = proc_write;

	/* create the proc entry for both read and write for the settings */
	au1xxx_pm_setup->proc_entry = create_proc_entry("settings", 0644, power_dir);
	if (au1xxx_pm_setup->proc_entry == NULL) {
		retval = -ENOMEM;
		goto no_settings;
	}
	au1xxx_pm_setup->proc_entry->data = au1xxx_pm_setup;
	au1xxx_pm_setup->proc_entry->read_proc = proc_read_settings;
	au1xxx_pm_setup->proc_entry->write_proc = proc_write_settings;
	au1xxx_pm_setup->valid = 1;

	/* create the proc entry for read only for the pending entry */
	pending_proc_entry = create_proc_entry("pending", 0444, power_dir);
	if (pending_proc_entry == NULL) {
		retval = -ENOMEM;
		goto no_pending;
	}
	pending_proc_entry->read_proc = proc_read_pending;

	/* create the proc entry for both read only for the error entry */
	error_proc_entry = create_proc_entry("error", 0444, power_dir);
	if (error_proc_entry == NULL) {
		retval = -ENOMEM;
		goto no_error;
	}
	error_proc_entry->data = first_error;
	error_proc_entry->read_proc = proc_read_error;

	/* setup the proc entry for the devs that registered before proc_fs did */
	for (i = 0; i < MAX_SUPPORTED_DEVICES; i++) {
		if ((sys_power_pointers[i]->valid) && 
				(sys_power_pointers[i]->proc_entry == NULL))
		{
			sys_power_pointers[i]->proc_entry = create_proc_entry(
					sys_power_pointers[i]->name, 
					0644, 
					power_dir);
			if (sys_power_pointers[i]->proc_entry == NULL) {
				retval = -ENOMEM;
				remove_au1xxx_power_device(sys_power_pointers[i]);
				return retval;
			}
			sys_power_pointers[i]->proc_entry->data = sys_power_pointers[i];
			sys_power_pointers[i]->proc_entry->read_proc = proc_read;
			sys_power_pointers[i]->proc_entry->write_proc = proc_write;
		}
		if ((sys_power_pointers[i]->valid) &&
				(sys_power_pointers[i]->timer_proc_entry == NULL))
		{
			sys_power_pointers[i]->timer_proc_entry = create_proc_entry(
					sys_power_pointers[i]->name,
					0644,
					timeout_dir);
			if (sys_power_pointers[i]->timer_proc_entry == NULL) {
				retval = -ENOMEM;
				remove_au1xxx_power_device(sys_power_pointers[i]);
				return retval;
			}
			sys_power_pointers[i]->timer_proc_entry->data = sys_power_pointers[i];
			sys_power_pointers[i]->timer_proc_entry->read_proc = proc_read_timer;
			sys_power_pointers[i]->timer_proc_entry->write_proc = proc_write_timer;
		}
	}

	/* setup the periodic timers */
	DPRINTK("Initializing the au1xxx_pm_timers\n");
	for (i = 0; i < MAX_SUPPORTED_TIMERS; i++) {
		if (sys_power_pointers[i]->timeout) {
			init_timer(&sys_power_pointers[i]->timer);
			sys_power_pointers[i]->timer.function = &au1xxx_pm_service_timer;
			sys_power_pointers[i]->timer.expires = jiffies + sys_power_pointers[i]->timeout;
			sys_power_pointers[i]->timer.data = (unsigned long)sys_power_pointers[i];
			add_timer(&sys_power_pointers[i]->timer);
		}
	}

	spin_unlock_irqrestore(&au1xxx_pm_lock, spin_lock_flags);
	
	/* test code to create a new driver */
#if ENABLE_TEST_DEV
	test_dev = new_au1xxx_power_device("test1", &test_callback, NULL);
	test_dev = new_au1xxx_power_device("test2", &test_callback, NULL);
	test_dev = new_au1xxx_power_device("test3", &test_callback, NULL);
	test_dev = new_au1xxx_power_device("test4", &test_callback, NULL);
	test_dev = new_au1xxx_power_device("test5", &test_callback, NULL);
	au1xxx_pm_debug();

	remove_au1xxx_power_device(test_dev);
	au1xxx_pm_debug();
#endif

    pm_catchall_pm_init();
	 
	if ((retval == 0))
		printk(KERN_INFO "Au1xxx Power Management Driver v%s\n", AU1XXX_PM_VERSION);
	else
		printk(KERN_ERR "Unable to load Au1xxx Power Management Driver.\n");
		DPRINTK("retval: %d\n", retval);
	return retval;

	// TODO: does this need locking?
	remove_proc_entry("error", power_dir);
no_error:
	remove_proc_entry("pending", power_dir);
no_pending:
	remove_proc_entry("settings", power_dir);
no_settings:
	remove_proc_entry(system_dev->name, power_dir);
no_system_power:
	remove_proc_entry("timeout", NULL);
no_timeout_dir:
	remove_proc_entry("pm", NULL);
no_power_dir:
	printk(KERN_ERR "Unable to load Au1xxx Power Management Driver.\n");
	return retval;
}


static void au1xxx_pm_debug(void) {
#if DEBUG
	int i = 0;

	for (i = 0; i < MAX_SUPPORTED_DEVICES; i++) {
		DPRINTK(" ---- au1xxx_pm_struct #%d ---- \n", i);
		if (!sys_power_pointers[i]->valid) {
			DPRINTK("   Invalid device.\n");
		} else {
			DPRINTK("   Name: %s\n", sys_power_pointers[i]->name);
			DPRINTK("   Current State: %d\n", sys_power_pointers[i]->cur_state);
			DPRINTK("   Previous State: %d\n", sys_power_pointers[i]->prev_state);
			if (sys_power_pointers[i]->proc_entry)
				DPRINTK("   /proc entry exists\n");
			else 
				DPRINTK("   /proc entry does NOT exist\n");
			DPRINTK("   Pending Transition: %d\n", sys_power_pointers[i]->state_change_pending);
			DPRINTK("   Error Code: %d\n", sys_power_pointers[i]->last_error);
		}
	}
	DPRINTK("\n");
	for (i = 0; i < MAX_SUPPORTED_ERRORS; i++) {
		DPRINTK(" ---- au1xxx_pm_error #%d ---- \n", i);
		if (!sys_error_pointers[i]->dev) {
			DPRINTK("   No stored error.\n");
		} else {
			DPRINTK("   Device Name: %s\n", sys_error_pointers[i]->dev->name);
			DPRINTK("   Error Code: %d\n", sys_error_pointers[i]->error_code);
			DPRINTK("   Occurances: %d\n", sys_error_pointers[i]->error_count);
			DPRINTK("   First Occurance: %d\n", (int) sys_error_pointers[i]->first_timestamp.tv_sec);
			DPRINTK("   Latest Occurance: %d\n", (int) sys_error_pointers[i]->last_timestamp.tv_sec);
		}
	}
#endif
	
	return;
}

__initcall(au1xxx_pm_proc_init);

#endif  /* CONFIG_PM */
