/*
 * AMD 5536 USB OTG controller driver state machine
 */

/*
 * Copyright (C) 2005 AMD (http://www.amd.com)
 * Author: Karsten Boge 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/*****************************************************************************
 *  Includes
 *****************************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/unaligned.h>

/* #define DEBUG */
/* #define VERBOSE */

#ifndef CONFIG_USB_OTG
#define CONFIG_USB_OTG
#endif
#undef CONFIG_USB_OTG_ALWAYS_END_SESSION

#ifdef CONFIG_SOC_AU1X00
#if	LINUX_VERSION_CODE < KERNEL_VERSION(2,5,59)
#include <asm/au1000.h>
#include <asm/au1000_gpio.h>
#else
#include <asm/mach-au1x00/au1000.h>
#include <asm/mach-au1x00/au1000_gpio.h>
#endif
#endif

#define DRIVER_REVISION         "01.00.0200 - $Revision: 1.1 $"
#define DRIVER_NAME_FOR_PRINT   OTG_DRIVER_NAME
#include <linux/usb.h>
#include <linux/usb_gadget.h>
#include <linux/usb_otg.h>
#include "amd5536otg.h"
#include "amd5536uoc.h"


/*****************************************************************************
 *  Data
 *****************************************************************************/

extern u32                             otg_tmr_high_count;
extern struct usb_otg_gadget_extension otg_gadget_extension;


/*****************************************************************************
 *  Function Definitions
 *****************************************************************************/

/**
 * \brief
 * Process SRP (B-device)
 *
 * Initiate or continue a Session Request
 *
 * \param  otg   OTG controller info
 * \param  state actual HW state
 * \param  event actual event state
 * \param  pMask  
 *
 * \return       void
 */
inline void process_srp (struct otg *otg, u32 state, u32 event, u32 *pMask)
{
	switch (otg->transceiver.state) {

	case OTG_STATE_B_SRP_INIT :

		if (IS_BIT_RES (OTG_STS_SESSEND, state) ||
		    IS_BIT_SET (OTG_STS_LST, state)) {

			/* VBus too high | line state != SE0 => B_IDLE       */

			SET_FLAG (otg, OTG_B_SRP_ERROR);
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pMask);
		}
		else {

			/* start timer => SE0 wait                           */

			set_srp_conditions (otg);
			SET_OTG_TIMER (otg, SRP_WAIT_SE0);
			CHANGE_STATE (otg, OTG_STATE_B_SRP_WAIT_SE0, pMask);
		}
		break;

	case OTG_STATE_B_SRP_WAIT_SE0 :

		CHECK_STATE (otg, OTG_STATE_B_SRP_WAIT_SE0, pMask);
		if (GOT_EVENT (OTG_INT_TMX, event)) {

			/* timer exp. => D_PULSE                             */
			/*               start timer, set state              */

			SET_OTG_TIMER (otg, SRP_WAIT_DP);
			CHANGE_STATE (otg, OTG_STATE_B_SRP_D_PULSE, pMask);
		}
		else if (IS_BIT_RES (OTG_STS_SESSEND, state) ||
		         GOT_EVENT (OTG_INT_LSTC, event)) {

			/* VBus too high | line state != SE0 => B_IDLE       */
			/*       this should not happen here,                */
			/*       it's just to avoid a deadlock scene         */
			/*   Stop timer, set state                           */

			SET_FLAG (otg, OTG_B_SRP_ERROR);
			reset_srp_conditions (otg);
			ERR ("SRP: bus is not idle\n");
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pMask);

			/* TODO kboge: system notification                   */
			/*   otg->transceiver.state & OTG_STATE_B_SRP_ERROR  */
		}
		break;

	case OTG_STATE_B_SRP_D_PULSE :

		CHECK_STATE (otg, OTG_STATE_B_SRP_D_PULSE, pMask);
		if (GOT_EVENT (OTG_INT_TMX, event)) {

			/* D_PULSE done => V_PULSE                           */
			/*                 start timer, set state            */

			SET_OTG_TIMER_LONG (otg, SRP_WAIT_VP);
			CHANGE_STATE (otg, OTG_STATE_B_SRP_V_PULSE, pMask);
		}
		break;

	case OTG_STATE_B_SRP_V_PULSE :

		CHECK_STATE (otg, OTG_STATE_B_SRP_V_PULSE, pMask);
		if (GOT_EVENT (OTG_INT_TMX, event)) {

			/* V_PULSE done => wait VBus (session valid)         */
			/*                 start timer, set state            */

			SET_OTG_TIMER_LONG (otg, SRP_DCHRG_V);
			CHANGE_STATE (otg, OTG_STATE_B_SRP_V_DCHRG, pMask);
		}
		break;

	case OTG_STATE_B_SRP_V_DCHRG :

		CHECK_STATE (otg, OTG_STATE_B_SRP_V_DCHRG, pMask);
		if (GOT_EVENT (OTG_INT_TMX, event)) {

			/* V_PULSE done => wait VBus (session valid)         */
			/*                 start timer, set state            */

			if (IS_BIT_SET (OTG_STS_SESSVLD, state)) {

				/* Session valid => B_IDLE                   */

				RES_FLAG (otg, OTG_B_aSSN_REQ);
				reset_srp_conditions (otg);
				CHANGE_STATE (otg, OTG_STATE_B_IDLE, pMask);
			}
			else {
				/* Timer expired => B_SRP_WAIT_VBUS          */

				SET_OTG_TIMER_LONG (otg, SRP_WAIT_VRS);
				CHANGE_STATE (otg, OTG_STATE_B_SRP_WAIT_VBUS,
				              pMask);
			}
		}
		break;

	case OTG_STATE_B_SRP_WAIT_VBUS :

		CHECK_STATE (otg, OTG_STATE_B_SRP_WAIT_VBUS, pMask);
		if (IS_BIT_SET (OTG_STS_SESSVLD, state)) {

			/* Session valid => B_IDLE                           */
			/*                  stop timer                       */

			RES_FLAG (otg, OTG_B_aSSN_REQ);
			reset_srp_conditions (otg);
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pMask);
		}
		else if (GOT_EVENT (OTG_INT_TMX, event)) {

			/* VBus TO => B_IDLE                                 */
			/*            system notification                    */

			SET_FLAG (otg, OTG_B_SRP_ERROR);
			RES_FLAG (otg, OTG_B_aSSN_REQ);
			reset_srp_conditions (otg);
			ERR ("SRP: Device Not Responding, giving up\n");
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pMask);

			/* TODO kboge: system notification                   */
			/*   otg->transceiver.state & OTG_STATE_B_SRP_ERROR  */
		}
		break;

	/* unlikely */ default :

		/* something went wrong */

		BUG ();
		break;
	}
}

/**
 * \brief
 * Detect SRP (A-device)
 *
 * Detect a valid completion of a Session Request
 *
 * \param  otg   OTG controller info
 * \param  state actual HW state
 * \param  event actual event state
 * \param  pMask  
 *
 * \return       1: detected, 0: idle or still running
 */
inline int detect_srp (struct otg *otg, u32 state, u32 event, u32 *pMask)
{
	int  ret = 0;

	switch (otg->transceiver.state) {

	case OTG_STATE_A_IDLE :

		if (IS_BIT_SET (OTG_STS_LST, state)) {

			/* DP pulse detected => A_IDLE_WAIT_DPULSE           */
			/*     Start timer, set state                        */

			SET_OTG_TIMER (otg, ASRP_WAIT_MP);
			CHANGE_STATE (otg, OTG_STATE_A_IDLE_WAIT_DP, pMask);
		}
		else if (IS_BIT_SET (OTG_STS_SESSVLD, state)) {

			/* VBus pulse detected => A_IDLE_WAIT_VPULSE         */
			/*     Start timer, set state                        */

			SET_OTG_TIMER (otg, ASRP_WAIT_MP);
			CHANGE_STATE (otg, OTG_STATE_A_IDLE_WAIT_VP, pMask);
		}
		break;

	case OTG_STATE_A_IDLE_WAIT_DP :

		CHECK_STATE (otg, OTG_STATE_A_IDLE_WAIT_DP, pMask);
		if (GOT_EVENT (OTG_INT_LSTC, event) &&
		    IS_BIT_SET (OTG_STS_LST, state)) {

			/* D pulse changed => restart detection              */
			/*                    Set state                      */

			reset_timer (otg);
		}
		else if (IS_BIT_RES (OTG_STS_LST, state)) {

			/* DP pulse too short => A_IDLE                      */
			/*                       Set state                   */

			WARN ("short D pulse detected\n");
			CHANGE_STATE (otg, OTG_STATE_A_IDLE, pMask);
		}
		else if (GOT_EVENT (OTG_INT_TMX, event)) {

			/* DP pulse still active => A_IDLE_WAIT_DV           */
			/*                    Start timer, Set state         */

			SET_OTG_TIMER_LONG (otg, ASRP_WAIT_DV);
			CHANGE_STATE (otg, OTG_STATE_A_IDLE_WAIT_DV, pMask);
		}
		break;

	case OTG_STATE_A_IDLE_WAIT_VP :

		CHECK_STATE (otg, OTG_STATE_A_IDLE_WAIT_VP, pMask);
		if (IS_BIT_RES (OTG_STS_SESSVLD, state)) {

			/* VBusP pulse too short => A_IDLE                   */
			/*                          Set state                */

			WARN ("short VBus pulse detected\n");
			CHANGE_STATE (otg, OTG_STATE_A_IDLE, pMask);
		}
		else if (GOT_EVENT (OTG_INT_TMX, event)) {

			/* VBus pulse still active => A_IDLE_WAIT_DV         */
			/*                    Start timer, set state         */

			SET_OTG_TIMER_LONG (otg, ASRP_WAIT_DV);
			CHANGE_STATE (otg, OTG_STATE_A_IDLE_WAIT_DV, pMask);
		}
		break;

	case OTG_STATE_A_IDLE_WAIT_DV :

		CHECK_STATE (otg, OTG_STATE_A_IDLE_WAIT_DV, pMask);
		if (GOT_EVENT (OTG_INT_TMX, event)) {

			/* SRP detected => ready to activate the bus         */
			/*                 Set state                         */

			ret = 1;
			CHANGE_STATE (otg, OTG_STATE_A_IDLE, pMask);
			VDBG ("  OTG event: SRP detected\n");

		/* TODO kboge: pulses that are too long will be accepted     */
		}
		break;

	/* unlikely */ default :

		/* something went wrong */

		BUG ();
		break;
	}
	return (ret);
}

/**
 * \brief
 * OTG state change
 *
 * Basic state machine controlling the OTG states described in the
 * OTG supplement to the USB 2.0 spec, section 6.8.
 * This can be called either from the driver ISR or as response to
 * an application request, where application does not neccessarily
 * mean a user space application, more likely this routine will be
 * called from another kernel driver (gadget or device drivers).
 *
 * \param  otg          OTG controller info
 * \param  event_code   event that requested a state change
 * \param  pEvt_mask
 *
 * \return              events that were not handled here
 */
u32 otg_change_state (struct otg *otg, u32 _event, u32 *pEvt_mask)
{
	u32  event_code = _event;
	u32  uoc_status = get_status (otg);

	if (GOT_EVENT (OTG_A_CLR_ERR, event_code) &&
	    ((otg->transceiver.state & OTG_STATE_MASK) !=
	     OTG_STATE_A_VBUS_ERR)) {

		/* This request doesn't make sense: clear it                 */

		RES_FLAG (otg, OTG_A_CLR_ERR);
		RES_EVENT (OTG_A_CLR_ERR, event_code);
	}
	
	if (GOT_EVENT (OTG_GADGET_READY, event_code)) {

		/* Switch from "NO_B_DEVICE" states to normal operation or   */
		/*       deactivate operations in case gadget was unloaded   */

		if ((otg->transceiver.state & OTG_STATE_MASK) !=
		    OTG_STATE_UNDEFINED) {

			CHANGE_STATE (otg, OTG_STATE_UNDEFINED, pEvt_mask);
		}
	}

	if ((OTG_INT_TMX & event_code) && otg_tmr_high_count) {

		/* a long timer is running : decrement the high part         */

		restart_timer (otg);
		otg_tmr_high_count--;
		RES_EVENT (OTG_INT_TMX, event_code);
	}

	if (((otg->transceiver.state & OTG_STATE_MASK) !=
	     OTG_STATE_A_SUSPEND) &&
	    ((otg->transceiver.state & OTG_STATE_MASK) !=
	     OTG_STATE_B_HOST) &&
	    (GOT_EVENT (OTG_INT_DPRC, event_code))) {

		/* Host drives port reset => reset bHNP flag                 */

		if (is_b_hnp_enabled (otg)) {
			reset_b_hnp_enable (otg);
		}
		RES_EVENT (OTG_INT_DPRC, event_code);
	}

	if (GOT_EVENT (OTG_INT_HNPFC, event_code)) {

		/* HNP feature changed, just report it to the system         */

		RES_EVENT (OTG_INT_HNPFC, event_code);
#ifdef DEBUG
		if (IS_BIT_SET (OTG_STS_HNPSTS, uoc_status)) {

			/* status change: only SET_FEATURE is possible      */
			/*                flags cannot be reset here        */

			otg->transceiver.params |= uoc_status &
			                           OTG_STS_HNPSTS;
		}
		else {

			/* status reset: this must have been a               */
			/*               port reset or session end           */

			otg->transceiver.params &= ~OTG_STS_HNPSTS;
		}
		DBG ("    HNP feature change (%08x)\n",
		     (OTG_STS_HNPSTS & uoc_status));
		if (IS_BIT_SET (OTG_STS_HNP_EN, uoc_status)) {}
		if (IS_BIT_SET (OTG_STS_HNP_SUPP, uoc_status)) {}
		if (IS_BIT_SET (OTG_STS_HNP_ALTSUPP, uoc_status)) {}
#endif  /* DEBUG */

		/* TODO kboge: system notification                           */
		/*             IS_BIT_SET (OTG_STS_HNPSTS, uoc_status)       */
	}

#ifdef VERBOSE
/* 	if (GOT_EVENT (OTG_INT_LSTC, event_code)) { */

		/* Line state changed, just report it to the system          */

/* 		RES_EVENT (OTG_INT_LSTC, event_code); */
/* 		VDBG ("    line state: %2x\n", */
/* 		      ((OTG_STS_LST & uoc_status) >> 4)); */
/* 	} */

	if (GOT_EVENT (OTG_INT_PSPDC, event_code)) {

		/* Port speed changed, just report it to the system          */

		RES_EVENT (OTG_INT_PSPDC, event_code);

		if (OTG_STS_PCON & uoc_status) {

			if (!(OTG_STS_PSPD & uoc_status)) {
				DBG ("    port is in HIGH_SPEED\n");
			}
			else if (OTG_STS_PSPD_LS & uoc_status) {
				DBG ("    port is in LOW_SPEED\n");
			}
			else {
				DBG ("    port is in FULL_SPEED\n");
			}
		}
	}
#endif  /* VERBOSE */

	if (event_code) {

	DBG ("OTG-state change, event code: %08x\n",
	     otg->transceiver.params | event_code);

#ifdef DEBUG
		if (GOT_EVENT (OTG_INT_OCD, event_code)) {

			/* Port overcurrent detected, just report it here    */

			if (OTG_STS_OC & uoc_status) {
				DBG ("    over-current detected\n");
			}
			else {
				DBG ("    spurious over-current event\n");
			}
		}
		if (GOT_EVENT (OTG_INT_PSC, event_code)) {

			/* Port suspend state changed, just report it        */

			if (OTG_STS_PSUS & uoc_status) {
				DBG ("    port is suspended\n");
			}
			else {
				DBG ("    port is active\n");
			}
		}
		if (GOT_EVENT (OTG_INT_PCC, event_code)) {

			/* Port connect state changed, just report it        */

			if (OTG_STS_PCON & uoc_status) {
				DBG ("    port is connected\n");
			}
			else {
				DBG ("    port is not connected\n");
			}
		}
#endif  /* DEBUG */

	do switch (otg->transceiver.state & OTG_STATE_MASK) {

	/* STATE_UNDEFINED (yet): init state, 1st time after loading         */
	/* =========================================================         */

	case OTG_STATE_UNDEFINED :

		CHECK_STATE (otg, OTG_STATE_UNDEFINED, pEvt_mask);

		if (IS_FLAG_RES (otg, OTG_FLAGS_ACTIV)) {

			/* seems to be the first time: let it run !          */

			SET_FLAG (otg, OTG_FLAGS_ACTIV);
		}

		/*      muxer is still neutral                               */

		RES_EVENT ((OTG_INT_IDC | OTG_INT_TMX), event_code);

		if (IS_FLAG_RES (otg, OTG_GADGET_READY)) {

			/* NOT_READY (substate): gadget drivers not ready    */
			/* ==============================================    */
			/*                                                   */
			/*     check ID pin and enable host as needed        */

			if (IS_BIT_RES (OTG_STS_ID, uoc_status)) {

				/* ID pin connected: A-device                */
				/*         muxer -> host                     */

				if (OTG_STATE_NO_B_DEVICE_A !=
				    otg->transceiver.state) {
					CHANGE_STATE (otg,
					              OTG_STATE_NO_B_DEVICE_A,
					              pEvt_mask);
				}
			}
			else if (IS_BIT_SET (OTG_STS_ID, uoc_status)) {

				/* ID pin not connected: disable             */
				/*         muxer -> neutral                  */

				if (OTG_STATE_NO_B_DEVICE_B !=
				    otg->transceiver.state) {
					CHANGE_STATE (otg,
					              OTG_STATE_NO_B_DEVICE_B,
					              pEvt_mask);
				}
			}
		}
		else if ((OTG_STATE_NO_B_DEVICE_A ==
		          otg->transceiver.state) ||
		         (OTG_STATE_NO_B_DEVICE_B ==
		          otg->transceiver.state)) {

			/* Exit "not ready" state                            */
			/* ----------------------                            */

			RES_EVENT (OTG_GADGET_READY, event_code);
			CHANGE_STATE (otg, OTG_STATE_UNDEFINED, pEvt_mask);
		}

		/* ==================================================        */

		else if (IS_BIT_RES (OTG_STS_ID, uoc_status)) {

			/* ID pin connected: A-device                        */
			/*         muxer -> host                             */

			CHANGE_STATE (otg, OTG_STATE_A_IDLE, pEvt_mask);
		}
		else {

			/* ID pin is not connected: B-device                 */
			/*         muxer -> device (peripheral)              */

			RES_FLAG (otg, OTG_A_VBUS_FAILED);
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);
		}
		break;

	/* A: ID pin is connected, A-device drives VBus                      */
	/* ============================================                      */

	case OTG_STATE_A_IDLE :

		/* A_IDLE: start state for A-devices                         */
		/*         no VBus, no device connect, no activity           */

		CHECK_STATE (otg, OTG_STATE_A_IDLE, pEvt_mask);

		if (IS_BIT_SET (OTG_STS_ID, uoc_status)) {

			/* ID pin changed => B_IDLE                          */
			/*                   Switch muxer, set state         */

			RES_EVENT (OTG_INT_IDC, event_code);
			RES_FLAG (otg, OTG_A_VBUS_FAILED);
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);
		}
		else if (IS_FLAG_RES (otg, OTG_A_BUS_DROP) &&
		         (IS_FLAG_SET (otg, OTG_A_BUS_REQ) ||
		          detect_srp (otg, uoc_status,
		                      event_code, pEvt_mask))) {

			/* Srp detect | bus request => A_WAIT_VRISE          */
			/*             VBus on, start VRise timer, set state */

			SET_OTG_TIMER_LONG (otg, A_WAIT_VRISE);
			CHANGE_STATE (otg, OTG_STATE_A_WAIT_VRISE, pEvt_mask);
		}
		else if (IS_FLAG_SET (otg, OTG_A_BUS_DROP)) {

			RES_FLAG (otg, OTG_A_BUS_DROP);
		}
		break;

	case OTG_STATE_A_WAIT_VRISE :

		/* A_WAIT_VRISE: A-device waits for VBus valid               */
		/*        drive VBus, no device connect, no activity         */

		CHECK_STATE (otg, OTG_STATE_A_WAIT_VRISE, pEvt_mask);

		if (IS_BIT_SET ((OTG_STS_ID | OTG_STS_VBUSVLD), uoc_status) ||
		    GOT_EVENT (OTG_INT_TMX, event_code) ||
		    IS_FLAG_SET (otg, OTG_A_BUS_DROP)) {

			/* ID change | VBus valid | VRise TO | Bus drop req  */
			/*                                    => A_WAIT_BCON */
			/*                       Start BCon timer, set state */

			RES_EVENT (OTG_INT_TMX, event_code);

			/* Spec notice: don`t need to have a TO condition    */
			/*              here, the host just waits for a      */
			/* (#01)        device connect or ID change          */

			/* (#01)  s...g (dev, OTG_TMR_A_WAIT_BCON); */
			CHANGE_STATE (otg, OTG_STATE_A_WAIT_BCON, pEvt_mask);
		}
		break;

	case OTG_STATE_A_WAIT_BCON :

		/* A_WAIT_BCON: A-device waits for B-device connect          */
		/*              drive VBus, no device connect, no activity   */

		CHECK_STATE (otg, OTG_STATE_A_WAIT_BCON, pEvt_mask);

		if (IS_BIT_SET (OTG_STS_ID, uoc_status) ||
		    GOT_EVENT (OTG_INT_TMX, event_code) ||
		    IS_FLAG_SET (otg, OTG_A_BUS_DROP)) {

			/* ID change | BCon TO | Bus drop req                */
			/*           => A_WAIT_VFALL                         */
			/*              Stop timer, turn off VBus, set state */

			RES_EVENT ((OTG_INT_IDC | OTG_INT_TMX), event_code);
			RES_FLAG (otg, OTG_A_BUS_DROP);
			CHANGE_STATE (otg, OTG_STATE_A_WAIT_VFALL_DN,
			              pEvt_mask);
		}
		else if (IS_BIT_RES (OTG_STS_VBUSVLD, uoc_status)) {

			RES_EVENT (OTG_INT_VBVC, event_code);
			if (OTG_STATE_A_WAIT_BCON_VB !=
			    otg->transceiver.state) {

				/* VBus drop => start timer and re-check     */

				RES_EVENT (OTG_INT_TMX, event_code);
				SET_OTG_TIMER (otg, A_BCON_VB);
				CHANGE_STATE (otg, OTG_STATE_A_WAIT_BCON_VB,
				              pEvt_mask);
			}
			else if (GOT_EVENT (OTG_INT_TMX, event_code)) {

				/* VBus drop => A_VBUS_ERR                   */
				/*      Stop timer, turn off VBus, set state */

				RES_EVENT (OTG_INT_TMX, event_code);
				CHANGE_STATE (otg, OTG_STATE_A_VBUS_ERR,
				              pEvt_mask);
			}
		}
		else if (IS_BIT_SET (OTG_STS_PCON, uoc_status)) {

			/* Port connect => A_HOST                            */
			/*                 Stop timer, set state             */

			RES_EVENT (OTG_INT_PCC, event_code);
			CHANGE_STATE (otg, OTG_STATE_A_HOST, pEvt_mask);

		}
		else if (OTG_STATE_A_WAIT_BCON_VB ==
		         otg->transceiver.state) {
			CHECK_STATE (otg, OTG_STATE_A_WAIT_BCON_VB, pEvt_mask);
			RES_EVENT (OTG_INT_TMX, event_code);
			CHANGE_STATE (otg, OTG_STATE_A_WAIT_BCON, pEvt_mask);
		}
		break;

	case OTG_STATE_A_HOST :

		/* A_HOST: B-peripheral connected, ready for USB ops         */
		/*         drive VBus, bus is activ (SOF)                    */

		CHECK_STATE (otg, OTG_STATE_A_HOST, pEvt_mask);

		if (IS_BIT_RES (OTG_STS_VBUSVLD, uoc_status)) {

			/* VBus drop => A_VBUS_ERR                           */
			/*              Turn off VBus, set state             */
 
			RES_EVENT (OTG_INT_VBVC, event_code);
			reset_b_hnp_enable (otg);
			CHANGE_STATE (otg, OTG_STATE_A_VBUS_ERR, pEvt_mask);
		}
		else if (IS_BIT_SET (OTG_STS_ID, uoc_status) ||
		         IS_BIT_RES (OTG_STS_PCON, uoc_status) ||
			 IS_FLAG_SET (otg, OTG_A_BUS_DROP)) {

			/* ID change | Port dc  => A_WAIT_BCON               */
			/*                         Set state                 */

			RES_EVENT ((OTG_INT_IDC | OTG_INT_PCC), event_code);
			reset_b_hnp_enable (otg);
			CHANGE_STATE (otg, OTG_STATE_A_WAIT_BCON, pEvt_mask);
		}
		/* Spec issue: don't want an immediate resume here so the    */
		/*             OTG_A_BUS_REQ flag can't be checked !         */
		/* (#02)  suspend/resume handling should be under HC control */

		/* (#02) if (IS_BIT_SET (OTG_STS_PSUS, uoc_status) ||        */
		/* (#02)     IS_BIT_RES (OTG_A_BUS_REQ,                      */
		/*                       otg->transceiver.params)) {         */
		else if (IS_BIT_SET (OTG_STS_PSUS, uoc_status)) {  /* (#02)  */

			/* Port suspend | Bus drop req => A_SUSPEND          */
			/*                 Start disconnect timer, set state */

			RES_EVENT (OTG_INT_PSC, event_code);
			/* (#02) RES_FLAG (otg, OTG_A_BUS_REQ); */

		/* Spec notice: load the disconnect wait timer only if HNP   */
		/*              is enabled, otherwise don't end the session  */
		/* (#03)        on A_AIDL_BDIS_TO                            */

#ifndef CONFIG_USB_OTG_ALWAYS_END_SESSION                            /* (#3) */
			if (is_b_hnp_enabled (otg))                  /* (#3) */
#endif                                                               /* (#3) */
			  SET_OTG_TIMER_LONG (otg, A_IDLE_BDIS);     /* (#3) */

			CHANGE_STATE (otg, OTG_STATE_A_SUSPEND, pEvt_mask);
		}
#ifdef VERBOSE
		else {
			if (!(OTG_STS_PSPD & uoc_status)) {
				DBG ("    port is in HIGH_SPEED\n");
			}
			else if (OTG_STS_PSPD_LS & uoc_status) {
				DBG ("    port is in LOW_SPEED\n");
			}
			else {
				DBG ("    port is in FULL_SPEED\n");
			}
		}
#endif
		break;

	case OTG_STATE_A_SUSPEND :

		/* A_SUSPEND: B-peripheral connected, port suspended         */
		/*            drive VBus, no bus activity                    */

		CHECK_STATE (otg, OTG_STATE_A_SUSPEND, pEvt_mask);

		if (IS_BIT_SET (OTG_STS_ID, uoc_status) ||
		    IS_FLAG_SET (otg, OTG_A_BUS_DROP)) {

			/*  ID change | Bus drop req | B discon. TO          */
			/*           => A_WAIT_VFALL                         */
			/*              Turn off VBus, stop timer, set state */

			RES_EVENT ((OTG_INT_IDC | OTG_INT_TMX), event_code);
			reset_b_hnp_enable (otg);
			RES_FLAG (otg, OTG_A_BUS_DROP);
			CHANGE_STATE (otg, OTG_STATE_A_WAIT_VFALL_DN,
			              pEvt_mask);
		}
		else if (IS_BIT_RES (OTG_STS_VBUSVLD, uoc_status)) {

			/* VBus drop => A_VBUS_ERR                           */
			/*              Turn off VBus, set state             */

			RES_EVENT ((OTG_INT_VBVC | OTG_INT_TMX), event_code);
			reset_b_hnp_enable (otg);
			CHANGE_STATE (otg, OTG_STATE_A_VBUS_ERR,
			              pEvt_mask);
		}
		else if (IS_BIT_RES (OTG_STS_PCON, uoc_status)) {

			if (is_b_hnp_enabled (otg)) {

				/* HNP enabled & B discon. => A_PERIPHERAL   */
				/*                 Enable device, set state  */

				RES_EVENT ((OTG_INT_PCC | OTG_INT_TMX),
				           event_code);
				reset_b_hnp_enable (otg);
				CHANGE_STATE (otg, OTG_STATE_A_PERIPHERAL,
				              pEvt_mask);
#ifndef CONFIG_TEST_MODE_ONLY
				if (otg_gadget_extension.notify) {
					otg_gadget_extension.notify (
						OTG_GADGET_EVT_SVALID);
				}
#endif
			}
			else {
				/* HNP disabled & B discon. => A_WAIT_BCON   */
				/*                             Set state     */

				RES_EVENT ((OTG_INT_PCC | OTG_INT_TMX),
				           event_code);
				/* (#01)  s...g (otg, OTG_TMR_A_WAIT_BCON); */
				CHANGE_STATE (otg, OTG_STATE_A_WAIT_BCON,
				              pEvt_mask);
			}
		}
		/* (#02) if (IS_BIT_RES (OTG_STS_PSUS, uoc_status) ||        */
		/* (#02)     IS_BIT_SET (OTG_A_BUS_REQ,                      */
		/*                       otg->transceiver.params)) {         */
		else if (IS_BIT_RES (OTG_STS_PSUS, uoc_status)) {  /* (#02)  */

			/* Port resume => A_HOST                             */
			/*                Set state                          */

			RES_EVENT ((OTG_INT_PSC | OTG_INT_TMX), event_code);
			CHANGE_STATE (otg, OTG_STATE_A_HOST, pEvt_mask);
		}
		else if (GOT_EVENT (OTG_INT_TMX, event_code)) {

			/*  B discon. TO => A_WAIT_VFALL                     */
			/*                  Turn off VBus, set state         */

			RES_EVENT (OTG_INT_TMX, event_code);
			reset_b_hnp_enable (otg);
			RES_FLAG (otg, OTG_A_BUS_DROP);
			CHANGE_STATE (otg, OTG_STATE_A_WAIT_VFALL_DN,
			              pEvt_mask);
		}
		break;

	case OTG_STATE_A_PERIPHERAL :

		/* A_PERIPHERAL: connected to B-host, responding             */
		/*               drive VBus, remote bus activity             */

		CHECK_STATE (otg, OTG_STATE_A_PERIPHERAL, pEvt_mask);

		if (IS_BIT_SET (OTG_STS_ID, uoc_status) ||
		    IS_FLAG_SET (otg, OTG_A_BUS_DROP)) {

			/* ID on | Bus drop req => A_WAIT_VFALL              */
			/*                         Turn off VBus, set state  */

			RES_EVENT (OTG_INT_IDC, event_code);
			reset_b_hnp_enable (otg);
			RES_FLAG (otg, OTG_A_BUS_DROP);
			CHANGE_STATE (otg, OTG_STATE_A_WAIT_VFALL_DN,
			              pEvt_mask);
#ifndef CONFIG_TEST_MODE_ONLY
			if (otg_gadget_extension.notify) {
				otg_gadget_extension.notify (
					OTG_GADGET_EVT_SVDROP);
			}
#endif
		}
		else if (IS_BIT_RES (OTG_STS_VBUSVLD, uoc_status) ||
		         GOT_EVENT (OTG_INT_OCD, event_code)) {

			/* VBus off | over-current => A_VBUS_ERR             */
			/*                         Turn off VBus, set state  */

			RES_EVENT ((OTG_INT_VBVC | OTG_INT_OCD), event_code);
			reset_b_hnp_enable (otg);
			CHANGE_STATE (otg, OTG_STATE_A_VBUS_ERR, pEvt_mask);
#ifndef CONFIG_TEST_MODE_ONLY
			if (otg_gadget_extension.notify) {
				otg_gadget_extension.notify (
					OTG_GADGET_EVT_SVDROP);
			}
#endif
		}
		else if (OTG_STS_PSUS & uoc_status) {

			/* Port suspended => A_WAIT_BCON                     */
			/*                   Set state                       */

			RES_EVENT ((OTG_INT_PSC | OTG_INT_PCC), event_code);
			SET_OTG_TIMER_LONG (otg, A_WAIT_BCON);
			CHANGE_STATE (otg, OTG_STATE_A_WAIT_BCON, pEvt_mask);
#ifndef CONFIG_TEST_MODE_ONLY
			if (otg_gadget_extension.notify) {
				otg_gadget_extension.notify (
					OTG_GADGET_EVT_SVDROP);
			}
#endif
		}
#ifdef VERBOSE
		else {
			if (!(OTG_STS_PSPD & uoc_status)) {
				DBG ("    port is in HIGH_SPEED\n");
			}
			else {
				DBG ("    port is in FULL_SPEED\n");
			}
		}
#endif
		break;

	case OTG_STATE_A_WAIT_VFALL :

		/* A_WAIT_VFALL: A-device waits for Session end              */
		/*               no VBus, no connect, no activity            */

		CHECK_STATE (otg, OTG_STATE_A_WAIT_VFALL, pEvt_mask);

		if (IS_BIT_RES (OTG_STS_SESSEND, uoc_status)) {

			/* VBus on => no state change                        */
			/*            Discharge VBus                         */
			/* TODO kboge: need a time-out here                  */

			if (OTG_STATE_A_WAIT_VFALL_DN !=
			    otg->transceiver.state) {
				CHANGE_STATE (otg, OTG_STATE_A_WAIT_VFALL_DN,
				              pEvt_mask);
			}

		}
		else {

			/* Session end => A_IDLE (=> B_IDLE)                 */
			/*                Set state                          */

			RES_EVENT ((OTG_INT_IDC | OTG_INT_PCC), event_code);
			CHANGE_STATE (otg, OTG_STATE_A_IDLE, pEvt_mask);

		}
		break;

	case OTG_STATE_A_VBUS_ERR :

		/* A_VBUS_ERR: VBus valid was de-asserted                    */
		/*             no VBus, no connect, no activity              */

		CHECK_STATE (otg, OTG_STATE_A_VBUS_ERR, pEvt_mask);
		SET_FLAG (otg, OTG_A_VBUS_FAILED);

		if (IS_BIT_SET (OTG_STS_ID, uoc_status) ||
		    IS_FLAG_SET (otg, OTG_A_BUS_DROP | OTG_A_CLR_ERR)) {

			/* Set state                                         */

			RES_EVENT (OTG_INT_IDC, event_code);
			RES_FLAG (otg, OTG_A_CLR_ERR | OTG_A_BUS_DROP);
			CHANGE_STATE (otg, OTG_STATE_A_WAIT_VFALL, pEvt_mask);
		}
		break;

	/* B: ID pin is not connected                                        */
	/* ==========================                                        */

	case OTG_STATE_B_IDLE :

		/* B_IDLE: init state for B-devices                          */
		/*         monitor VBus, no connection, no activity          */

		CHECK_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);

		if (IS_BIT_RES (OTG_STS_ID, uoc_status)) {

			/* ID pin changed => A_IDLE                          */
			/*                   Set state, request "A-bus"      */

                    if (IS_FLAG_SET (otg, OTG_HOST_READY)) {

				SET_FLAG (otg, OTG_A_BUS_REQ);
			}
			RES_EVENT (OTG_INT_IDC, event_code);
			CHANGE_STATE (otg, OTG_STATE_A_IDLE, pEvt_mask);
		}
		else if (IS_BIT_SET (OTG_STS_SESSVLD, uoc_status)) {

			/* Session valid => B_PERIPHERAL                     */
			/*                  Set state                        */

			RES_EVENT (OTG_INT_SVC, event_code);
			CHANGE_STATE (otg, OTG_STATE_B_PERIPHERAL, pEvt_mask);
#ifndef CONFIG_TEST_MODE_ONLY
			if (otg_gadget_extension.notify) {
				otg_gadget_extension.notify (
					OTG_GADGET_EVT_SVALID);
			}
#endif
		}
		else if (IS_BIT_SET (OTG_STS_SESSEND, uoc_status) &&
		         IS_FLAG_SET (otg, OTG_GADGET_READY) &&
		         IS_FLAG_SET (otg, OTG_B_aSSN_REQ)) {

			/* Session end & Gadget ready & SRP request          */
			/*                => B_SRP_INIT  Set state           */

			RES_FLAG (otg, OTG_B_SRP_ERROR);
			CHANGE_STATE (otg, OTG_STATE_B_SRP_INIT, pEvt_mask);
		}
		else if (IS_FLAG_SET (otg, OTG_B_aSSN_REQ)) {

			RES_FLAG (otg, OTG_B_aSSN_REQ);
		}
		break;

	case OTG_STATE_B_SRP_INIT :

		/* B_SRP_INIT: device initiates a session request            */
		/*             pulse charge VBus, no activity                */

		CHECK_STATE (otg, OTG_STATE_B_SRP_INIT, pEvt_mask);

		if (IS_BIT_RES (OTG_STS_ID, uoc_status)) {

			/* ID pin changed => B_IDLE                          */
			/*               Stop timer, set resistors and state */
			RES_EVENT ((OTG_INT_IDC | OTG_INT_TMX), event_code);
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);
		}
		else {

			/* Process the SRP as required                       */

			process_srp (otg, uoc_status, event_code, pEvt_mask);
			RES_EVENT ((OTG_INT_SEC | OTG_INT_TMX), event_code);
		}
		break;

	case OTG_STATE_B_PERIPHERAL :

		/* B_PERIPHERAL: connected to A-host, responding             */
		/*               VBus driven by A, remote activity           */

		CHECK_STATE (otg, OTG_STATE_B_PERIPHERAL, pEvt_mask);

		if (IS_BIT_RES (OTG_STS_ID, uoc_status) ||
		    IS_BIT_RES (OTG_STS_SESSVLD, uoc_status)) {

			/* ID pin changed | ~Session valid => B_IDLE         */
			/*                                    Set state      */

			RES_EVENT ((OTG_INT_IDC | OTG_INT_SVC), event_code);
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);
#ifndef CONFIG_TEST_MODE_ONLY
			if (otg_gadget_extension.notify) {
				otg_gadget_extension.notify (
					OTG_GADGET_EVT_SVDROP);
			}
#endif
		}
		else if (OTG_STATE_B_PERIPHERAL_DC ==
		         otg->transceiver.state) {

			CHECK_STATE (otg, OTG_STATE_B_PERIPHERAL_DC,
			             pEvt_mask);
			/* TODO kboge: do we need a better TO condition here */

			/* Waiting for clean disconnect state                */

			if (IS_BIT_RES (OTG_STS_PCON, uoc_status) ||
			    IS_BIT_RES (OTG_STS_LST, uoc_status)) {

				/* now disconnected, set new state           */

				RES_EVENT ((OTG_INT_PCC | OTG_INT_TMX),
				           event_code);
				SET_OTG_TIMER (otg, B_ASE0_BRST);
				CHANGE_STATE (otg, OTG_STATE_B_WAIT_ACON,
				              pEvt_mask);
			}
			else if (GOT_EVENT (OTG_INT_TMX, event_code)) {

				/* Time over, still connected, asume this is */
				/*            already the A-peripheral       */

				RES_EVENT (OTG_INT_TMX, event_code);
				CHANGE_STATE (otg, OTG_STATE_B_HOST,
				              pEvt_mask);
				otg_hnp_start_enum (&otg->transceiver);
#ifndef CONFIG_TEST_MODE_ONLY
				if (otg_gadget_extension.notify) {
					otg_gadget_extension.notify (
						OTG_GADGET_EVT_SVDROP);
				}
#endif
			}
		}
		else if (IS_BIT_SET (OTG_STS_HNP_EN, uoc_status) &&
		         IS_FLAG_SET (otg, OTG_HOST_READY) &&
		         IS_FLAG_SET (otg, OTG_B_BUS_REQ) &&
		         IS_BIT_SET (OTG_STS_PSUS, uoc_status)) {

			/* BBus req & BHnp en & Bus suspended => B_WAIT_ACON */
			/*                      Wait, enable host, set state */

			if (OTG_STATE_B_PERIPHERAL_WT !=
			    otg->transceiver.state) {

				/* Wait _Ta_idle_bdis_ before disconnect     */

				RES_EVENT (OTG_INT_PSC, event_code);
				/* SET_OTG_TIMER_LONG (otg, B_AIDL_BDIS); */
				SET_OTG_TIMER (otg, B_AIDL_BDIS);
				CHANGE_STATE (otg, OTG_STATE_B_PERIPHERAL_WT,
				              pEvt_mask);
			}
			else {
				CHECK_STATE (otg, OTG_STATE_B_PERIPHERAL_WT,
				             pEvt_mask);

				if (GOT_EVENT (OTG_INT_TMX, event_code)) {

					/* Time over,                        */
					/* disconnect and change state       */

					RES_EVENT (OTG_INT_TMX, event_code);
					SET_OTG_TIMER_SHORT (otg,
					                     B_WAIT_ADISCON);

					/* dis-re-connect time could be ~0   */
					CHANGE_STATE (otg,
					              OTG_STATE_B_PERIPHERAL_DC,
						      pEvt_mask);
				}
			}
		}
		else if (OTG_STATE_B_PERIPHERAL_WT ==
		         otg->transceiver.state) {

			CHECK_STATE (otg, OTG_STATE_B_PERIPHERAL_WT,
			             pEvt_mask);
			if (IS_BIT_RES (OTG_STS_PSUS, uoc_status)) {

				/* Resumed during wait time                  */

				CHANGE_STATE (otg, OTG_STATE_B_PERIPHERAL,
				              pEvt_mask);
			}
		}
#ifdef VERBOSE
		else {
			if (!(OTG_STS_PSPD & uoc_status)) {
				DBG ("    port is in HIGH_SPEED\n");
			}
			else {
				DBG ("    port is in FULL_SPEED\n");
			}
		}
#endif
		break;

	case OTG_STATE_B_WAIT_ACON :

		/* B_WAIT_ACON: B-device waits for A-device connect          */
		/*              no device connect, no bus activity           */

		CHECK_STATE (otg, OTG_STATE_B_WAIT_ACON, pEvt_mask);

		if (IS_BIT_RES (OTG_STS_ID, uoc_status) ||
		    IS_BIT_RES (OTG_STS_SESSVLD, uoc_status)) {

			/* ID pin changed | ~Session valid => B_IDLE         */
			/*                         Enable device, set state  */

			RES_EVENT ((OTG_INT_IDC | OTG_INT_SVC | OTG_INT_TMX),
			            event_code);
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);
#ifndef CONFIG_TEST_MODE_ONLY
			if (otg_gadget_extension.notify) {
				otg_gadget_extension.notify (
					OTG_GADGET_EVT_SVDROP);
			}
#endif
		}
		else if ((IS_BIT_SET (OTG_STS_PCON, uoc_status)) &&
		         !(IS_BIT_RES (OTG_STS_LST, uoc_status))) {

			/* Port connect => B_HOST                            */
			/*       Set wait time before reset, set state       */

			RES_EVENT ((OTG_INT_PCC | OTG_INT_TMX), event_code);
			SET_OTG_TIMER_SHORT (otg, B_ACON_BRST);
			CHANGE_STATE (otg, OTG_STATE_B_HOST_WT, pEvt_mask);
		}
		else if (GOT_EVENT (OTG_INT_TMX, event_code)) {

			/* Time out => B_PERIPHERAL                          */
			/*             Set state                             */

			RES_EVENT ((OTG_INT_PSC | OTG_INT_HNPFC | OTG_INT_TMX),
			           event_code);

			/* NOTE: no UDC notification here                    */
			/*       don't want to confuse the UDC               */

			ERR ("HNP: Device Not Responding, giving up\n");
			CHANGE_STATE (otg, OTG_STATE_B_PERIPHERAL, pEvt_mask);
		}
		break;

	case OTG_STATE_B_HOST :

		/* B_HOST: A-peripheral connected, ready for USB ops         */
		/*         VBus is driven by A, bus is activ (SOF)           */

		CHECK_STATE (otg, OTG_STATE_B_HOST, pEvt_mask);

		if (OTG_STATE_B_HOST_WT == otg->transceiver.state) {

			CHECK_STATE (otg, OTG_STATE_B_HOST_WT, pEvt_mask);

			if (GOT_EVENT (OTG_INT_TMX, event_code)) {

				/* time over, now starting reset ...         */
			
				RES_EVENT (OTG_INT_TMX, event_code);
				CHANGE_STATE (otg, OTG_STATE_B_HOST,
				              pEvt_mask);
				otg_hnp_start_enum (&otg->transceiver);
#ifndef CONFIG_TEST_MODE_ONLY
				if (otg_gadget_extension.notify) {
					otg_gadget_extension.notify (
						OTG_GADGET_EVT_SVDROP);
				}
#endif
			}
		}
		else if (IS_BIT_RES (OTG_STS_ID, uoc_status) ||
		         IS_BIT_RES (OTG_STS_SESSVLD, uoc_status)) {

			/* ID pin changed | ~Session valid => B_IDLE         */
			/*                          Enable device, set state */

			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);
		}
		else if (IS_BIT_RES (OTG_STS_PCON, uoc_status)) {

			/* Port disconnect & Session valid                   */
			/*         => B_PERIPHERAL, Enable device, set state */

			CHANGE_STATE (otg, OTG_STATE_B_PERIPHERAL, pEvt_mask);
#ifndef CONFIG_TEST_MODE_ONLY
			if (otg_gadget_extension.notify) {
				otg_gadget_extension.notify (
					OTG_GADGET_EVT_SVALID);
			}
#endif
		}
#ifdef VERBOSE
		else {
			if (!(OTG_STS_PSPD & uoc_status)) {
				DBG ("    port is in HIGH_SPEED\n");
			}
			else if (OTG_STS_PSPD_LS & uoc_status) {
				DBG ("    port is in LOW_SPEED\n");
			}
			else {
				DBG ("    port is in FULL_SPEED\n");
			}
		}
#endif
		break;

	/* unlikely */ default :

		/* something went wrong */

		BUG ();
		break;
	}
	while ((otg->transceiver.state ^ otg->transceiver.prv_state) &
	       OTG_STATE_MASK);

	DBG ("OTG-state change done\n");
	}

	return event_code;
}


/* This comment closes the module definition from above. There can be multiple 
   definitions of this kind in a file. See the doxygen documentation for more
   information. */
/** \}*/
