/*
 * AMD 5536 USB OTG driver
 */

/*
 * Copyright (C) 2005 AMD (http://www.amd.com)
 * Author: Karsten Boge 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#ifndef AMD5536OTG_H
#define AMD5536OTG_H


/*****************************************************************************
*  Constants
*****************************************************************************/


/**********************************
 * OTG sub-state definitions
***********************************/

#define OTG_STATE_MASK                  0x0F

#define OTG_STATE_NO_B_DEVICE_A         (0x60 | OTG_STATE_UNDEFINED)
#define OTG_STATE_NO_B_DEVICE_B         (0x40 | OTG_STATE_UNDEFINED)

#define OTG_STATE_B_HOST_WT             (0x10 | OTG_STATE_B_HOST)
#define OTG_STATE_B_PERIPHERAL_WT       (0x10 | OTG_STATE_B_PERIPHERAL)
#define OTG_STATE_B_PERIPHERAL_DC       (0x20 | OTG_STATE_B_PERIPHERAL)
#define OTG_STATE_B_SRP_WAIT_SE0        (0x10 | OTG_STATE_B_SRP_INIT)
#define OTG_STATE_B_SRP_D_PULSE         (0x20 | OTG_STATE_B_SRP_INIT)
#define OTG_STATE_B_SRP_V_PULSE         (0x30 | OTG_STATE_B_SRP_INIT)
#define OTG_STATE_B_SRP_V_DCHRG         (0x40 | OTG_STATE_B_SRP_INIT)
#define OTG_STATE_B_SRP_WAIT_VBUS       (0x50 | OTG_STATE_B_SRP_INIT)

#define OTG_STATE_A_IDLE_WAIT_DP        (0x10 | OTG_STATE_A_IDLE)
#define OTG_STATE_A_IDLE_WAIT_VP        (0x20 | OTG_STATE_A_IDLE)
#define OTG_STATE_A_IDLE_WAIT_MP        (0x30 | OTG_STATE_A_IDLE)
#define OTG_STATE_A_IDLE_WAIT_DV        (0x40 | OTG_STATE_A_IDLE)
#define OTG_STATE_A_WAIT_BCON_VB        (0x10 | OTG_STATE_A_WAIT_BCON)
#define OTG_STATE_A_WAIT_VFALL_DN       (0x10 | OTG_STATE_A_WAIT_VFALL)


/**********************************
*  typical timer values
**********************************/

#define OTG_TMR_WAIT_VFALL     10   /* (  ) A waits for VBus                 */
#define OTG_TMR_A_WAIT_VRISE   100  /* (  ) A waits for VBus                 */
#define OTG_TMR_A_WAIT_BCON    200  /* (  ) A waits for B-connect (1.. s)    */
#define OTG_TMR_A_IDLE_BDIS    250  /* (ms) A waits for B-disc (200.. ms)    */
#define OTG_TMR_B_WAIT_ADISCON 600  /* (us) B waits for A to disconnect <1ms */
#define OTG_TMR_B_ACON_BRST    200  /* (us) B waits before starting reset    */
#define OTG_TMR_B_ASE0_BRST    5    /* (ms) B waits for A-conn (3.125.. ms)  */
#define OTG_TMR_B_AIDL_BDIS    50   /* (ms) B waits before dc (5..150ms)     */
#define OTG_TMR_SRP_WAIT_SE0   2    /* (  ) B SRP idle wait                  */
#define OTG_TMR_SRP_WAIT_DP    7    /* (ms) B SRP D_PULSE (5..10ms)          */
#define OTG_TMR_SRP_WAIT_VP    80   /* (ms) B SRP V_PULSE (5..100ms)         */
#define OTG_TMR_SRP_DCHRG_V    30   /* (  ) B SRP VBus discharge             */
#define OTG_TMR_SRP_WAIT_VRS   5800 /* (ms) B SRP waits for VBus (5..6s)     */
#define OTG_TMR_ASRP_WAIT_MP   4    /* (  ) A SRP min. pulse                 */
#define OTG_TMR_ASRP_WAIT_DP   10   /* (ms) A SRP D_PULSE TO                 */
#define OTG_TMR_ASRP_WAIT_VP   200  /* (ms) A SRP V_PULSE TO                 */
#define OTG_TMR_ASRP_WAIT_DV   200  /* (  ) A SRP waits for V_PULSE          */
#define OTG_TMR_A_BCON_VB      50   /* (  ) A waits for VBus after connect   */

#define OTG_TMR_IDSNS_WAIT     10   /* (ms) ID sense wait                    */

/*********************************/

#define TIMER_PERIOD           1000      /* 10 ms, if longer than 10ms       */

/**********************************
 * OTG state parameters
 **********************************/

#define OTG_HOST_READY         (1<<20)   /* indicates a USB host driver is   */
                                         /* running                          */
#define OTG_GADGET_READY       (1<<21)   /* indicates a USB gadget driver is */
                                         /* running                          */
#define OTG_A_BUS_REQ          (1<<22)   /* used by appl-SW to request a     */
                                         /* VBus rise, auto-reset by driver  */
#define OTG_A_BUS_DROP         (1<<23)   /* used by appl-SW to request a     */
                                         /* VBus drop, auto-reset by driver  */
#define OTG_A_CLR_ERR          (1<<24)   /* used by appl-SW to request VBerr */
                                         /* clean-up, auto-reset by driver   */
#define OTG_AB_HNP_REQ         (1<<25)   /* used by appl-SW to initiate      */
                                         /* HNP, auto-reset by driver        */
#define OTG_B_BUS_REQ          (1<<26)   /* used by appl-SW to request       */
                                         /* B-device functionality, ...      */
#define OTG_B_BUS_DIS          (1<<27)   /* used by appl-SW to request       */
                                         /* disable B-device functionality   */
#define OTG_B_aSSN_REQ         (1<<28)   /* used by appl-SW to initiate SRP, */
                                         /* auto-reset by the driver         */
#define OTG_B_SRP_ERROR        (1<<29)   /* indicates invalid HW conditions  */
                                         /* during SRP, reset by writing "1" */
#define OTG_A_VBUS_FAILED      (1<<30)   /* indicates a VBus error, reset by */
                                         /* writing "1", when setting        */
                                         /* CLR_ERR or when leaving A-states */
#define OTG_UDC_RWK_REQ        (1<<31)   /* call UDC function to force a     */
                                         /* remote wake-up                   */

#define SW_REQUEST_MASK        (OTG_A_BUS_REQ | OTG_A_BUS_DROP | \
                                OTG_A_CLR_ERR | OTG_B_aSSN_REQ | \
                                OTG_B_BUS_REQ | OTG_B_BUS_DIS | \
                                OTG_UDC_RWK_REQ)

/*********************************************************************/

/*
 * gadget events for notify function
 */
#define OTG_GADGET_EVT_SVDROP  (1<<0)    /* Session valid drop       */
#define OTG_GADGET_EVT_SVALID  (1<<1)    /* Session valid            */
#define OTG_GADGET_REQ_WAKE    (1<<2)    /* Request remote wake-up   */

/*** HS-A0 WA: gadget phy suspend issue                            ***/
#define OTG_FLAGS_UDC_SUSP     (1<<17)   /* gadget phy suspended     */


/*****************************************************************************
*  Types
*****************************************************************************/


/*****************************************************************************
*  Macros
*****************************************************************************/


/*****************************************************************************
*  Data
*****************************************************************************/
#ifdef __cplusplus
extern "C"
{
#endif

struct usb_otg_gadget_extension {
	int (*request) (u32);	/* function call for state change requests */
	u32 (*query) (int);	/* function call to query state            */
	void (*notify) (u32);   /* filled in by gadget for notification    */
};

#ifdef __cplusplus
} /* extern "C" */
#endif


/*****************************************************************************
*  Functions
*****************************************************************************/
#ifdef __cplusplus
extern "C"
{
#endif


#ifdef __cplusplus
} /* extern "C" */
#endif


/*****************************************************************************
*  Inline Functions
*****************************************************************************/


#endif /* AMD5536OTG_H */
