/*
 * AMD 5536 UDC high/full speed USB device controller.
 */

/*
 * Copyright (C) 2005 AMD (http://www.amd.com)
 * Author: Thomas Dahlmann
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*****************************************************************************
 * Defines
 *****************************************************************************/

/* debug control */
#undef DEBUG
//DISABLEDamd #define DEBUG     1
#undef VERBOSE
//DISABLEDamd #define   VERBOSE         
//DISABLEDamd #define UDC_REGISTER_DUMP

/* device driver registration of kernel 2.6.x usage */
//#define UDC_USE_DRIVER_REGISTER

/* prefix name for print */
#define DRIVER_NAME_FOR_PRINT   UDC_DRIVER_NAME

/* Driver name */
#define DRIVER_DESC             "AMD 5536 UDC USB Peripheral Controller"
#define DRIVER_VERSION          "01.00.0200 - $Revision: 1.1 $"

/* kernel version of new gadget stack generation (for 2.6.x) */
#define UDC_NEW_GADGET_KERNEL KERNEL_VERSION(2,5,59)
/*****************************************************************************
 *  Includes
 *****************************************************************************/

/* system */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/ioctl.h>
#include <linux/fs.h>
#if     LINUX_VERSION_CODE < UDC_NEW_GADGET_KERNEL
#include <linux/wrapper.h>
#else
#include <linux/dmapool.h>
#include <linux/moduleparam.h>
#include <linux/device.h>
#endif
#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/unaligned.h>

/* MIPS config */
#ifdef CONFIG_SOC_AU1200
#if     LINUX_VERSION_CODE < UDC_NEW_GADGET_KERNEL
#include <asm/au1000.h>
#else
#include <asm/mach-au1x00/au1000.h>
#endif
#ifndef CONFIG_USB_NON_PCI_OTGDEVICE
#define CONFIG_USB_NON_PCI_OTGDEVICE
#endif
#endif

/* gadget stack */
#include <linux/usb_ch9.h>
#include <linux/usb_gadget.h>
#include <linux/usb_otg.h>
#include "amd5536otg.h"

/* udc specific */
#include "amd5536udc.h"

/*****************************************************************************
 *  Static Function Declarations
 *****************************************************************************/

void udc_tasklet_disconnect(unsigned long);
static int udc_read_fifo_dwords(struct udc* dev, u32* buf, int dwords);
static int udc_write_fifo(struct udc* dev, int ep_ix, u32* buf, int bytes);
static void nuke (struct udc_ep *);
static int udc_probe (struct pci_dev *pdev, const struct pci_device_id *id);
static void udc_remove (struct pci_dev *pdev);
static void usb_reset (struct udc *dev);
static void usb_reinit (struct udc *dev);
static void udc_soft_reset(struct udc* dev);
#ifdef DEBUG
#ifdef NOT_USED
static void dump_buffer(u8* buf, u32 bytes);
#endif
static int udc_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int udc_open (struct inode *inode, struct file *file);
static int udc_release (struct inode *inode, struct file *file);
#endif
static struct udc_data_dma* udc_get_last_dma_desc(struct udc_request* req);
static u32 udc_get_ppbdu_rxbytes(struct udc_request* req);
static int udc_free_dma_chain(struct udc* dev, struct udc_request* req);
static int udc_create_dma_chain(struct udc_ep* ep, struct udc_request* req, unsigned long buf_len, int gfp_flags);
static inline int startup_registers(struct udc* dev);
#ifdef CONFIG_SOC_AU1200
void au1200_sync(void);
#endif
static int udc_remote_wakeup(struct udc* dev);
#ifdef UDC_USE_DRIVER_REGISTER
static int udc_suspend(struct udc* dev);
static int udc_resume(struct udc* dev);
#endif

/*****************************************************************************
 *  Data
 *****************************************************************************/

/* description */
static const char driver_name [] = UDC_DRIVER_NAME; 
static const char driver_desc [] = DRIVER_DESC;

/* structure to hold endpoint function pointers */
static struct usb_ep_ops udc_ep_ops;

/* received setup data */
static union udc_setup_data setup_data;

/* pointer to device object */      
static struct udc       *the_controller;

/* irq spin lock for soft reset */
spinlock_t udc_irq_spinlock;

/* TODO this is used for dma chaining, global gfp not good */
static int td_gfp_flags = 0;
        
/* slave mode: pending bytes in rx fifo after nyet,
used if EPIN irq came but no req was available */
static unsigned int udc_rxfifo_pending = 0;

/* count soft resets after suspend to avoid loop */
static int soft_reset_occured = 0;
#ifdef UDC_IPBUG_3958_WORKAROUND_SOFT_RESET_ON_USBRESET
static int soft_reset_after_usbreset_occured = 0;
#endif
/* tasklet for usb disconnect */
DECLARE_TASKLET(disconnect_tasklet, udc_tasklet_disconnect, (unsigned long) &the_controller);

#ifdef CONFIG_USB_NON_PCI_OTGDEVICE
static struct pci_dev pdev_dummy;
static struct pci_dev* pdev = &pdev_dummy;
#endif

#ifdef UDC_IPBUG_3958_WORKAROUND
/* CNAK pending field: bit0 = ep0in, bit16 = ep0out */
static u32 cnak_pending = 0;
#define UDC_QUEUE_CNAK(ep, num) \
        if (readl(&((ep)->regs->ctl)) & UDC_EPCTL_NAK) { \
                INFO("NAK could not be cleared for ep%d\n", num); \
                cnak_pending |= 1 << (num); \
                (ep)->naking = 1; \
        }
#else
#define UDC_QUEUE_CNAK(ep, num) {} 
#endif
#ifdef UDC_IPBUG_3958_WORKAROUND_RXFIFO_FLUSH
/* rxfifo cleari/trash buffer */
static u8 udc_rxfifo_trash[UDC_RXFIFO_SIZE];
#endif

#ifdef UDC_IPBUG_3943_WORKAROUND
/* otg registering count */
static u32 otg_reg_count = 0;
#endif
/* gadget registering count */
static u32 gadget_bind_count = 0;

#ifdef DEBUG
/* driver operations */
static struct file_operations udc_fops = {
        owner:          THIS_MODULE,

        read:           NULL,
        write:          NULL,
        ioctl:          udc_ioctl,
        open:           udc_open,
        release:        udc_release,
};
#endif

/* PCI device parameters */
static struct pci_device_id pci_ids [] = { {
        .class =        ((PCI_CLASS_SERIAL_USB << 8) | 0xfe),
        .class_mask =   ~0,
        .vendor =       UDC_PCI_VENDOR_ID,
        .device =       UDC_PCI_DEVICE_ID,
        .subvendor =    PCI_ANY_ID,
        .subdevice =    PCI_ANY_ID,

}, { /* zero */ }
};

#ifndef CONFIG_USB_NON_PCI_OTGDEVICE
/* PCI functions */
static struct pci_driver udc_pci_driver = {
        .name =         (char *) driver_name,
        .id_table =     pci_ids,

        .probe =        udc_probe,
        .remove =       udc_remove,

        /* TODO power management */
};
#endif

/* endpoint names used for print */
static const char ep0name [] = "ep0in";
static const char *ep_name [] = {
        ep0name,
        "ep1in-int", "ep2in-bulk", "ep3in-bulk", "ep4in-bulk", "ep5in-bulk", "ep6in-bulk", "ep7in-bulk", "ep8in-bulk",
        "ep9in-bulk", "ep10in-bulk", "ep11in-bulk", "ep12in-bulk", "ep13in-bulk", "ep14in-bulk", "ep15in-bulk", "ep0out",
        "ep1out-bulk", "ep2out-bulk", "ep3out-bulk", "ep4out-bulk", "ep5out-bulk", "ep6out-bulk", "ep7out-bulk", "ep8out-bulk",
        "ep9out-bulk", "ep10out-bulk", "ep11out-bulk", "ep12out-bulk", "ep13out-bulk", "ep14out-bulk", "ep15out-bulk" 
};

#if defined(CONFIG_MIPS)
/* util's sync only performs sync w/o refilling the pipe
   => read issued after write might appear before write on bus */
#define au1200_sync_delay() {\
  asm(".set noreorder"); \
  asm("b l1"); \
  asm("l1:");  \
  asm("sync");  \
  asm("nop");   \
  asm("nop");   \
  asm("nop");   \
  asm("nop");   \
  asm("b l2");   \
  asm("l2:");   \
  asm("nop");   \
  asm("nop");   \
  asm(".set reorder"); \
}

//DISABLEDamd #define wb_flush() au1200_sync() 
#define wb_flush() __asm__ volatile ("sync")
#else
#define wb_flush() {} 
#endif

#ifdef DEBUG
/* data for debuging only */
static unsigned long no_pref_req = 0;
static unsigned long no_req = 0;
static u32 same_cfg = 0;
static u32 num_enums = 0;
#ifdef UDC_REGISTER_DUMP
#ifdef VERBOSE
#define writel(d,a) {VDBG("write32: address=%08lx data=%08lx (%s)\n",a,d,#a);\
                     (*(volatile unsigned int *) __io_virt(a) = (d));}
#define writeb(d,a) {VDBG("write8 : address=%08lx data=%08lx (%s)\n",a,d,#a);\
                     (*(volatile unsigned char *) __io_virt(a) = (d));}
#define readl(a)   read32(a) 
#define readb(a)   read8(a) 
#endif
#endif
#endif

#if defined(CONFIG_MIPS)
#undef writel
#undef writeb
#if     LINUX_VERSION_CODE < UDC_NEW_GADGET_KERNEL
#define writel(d,a) {((*(volatile unsigned int *)(a)) = (__ioswab32(d))); wb_flush();}
#else
#define writel(d,a) {((*(volatile unsigned int *)(a)) = (d)); wb_flush();}
#endif
#define writeb(d,a) {((*(volatile unsigned char *)(a)) = (d)); wb_flush();}
//DISABLEDamd #define writel(d,a) {((*(volatile unsigned int *)(a)) = (__ioswab32(d))); au1200_sync();}
//DISABLEDamd #define writeb(d,a) {((*(volatile unsigned char *)(a)) = (d)); au1200_sync();}
#endif 
/****** following flags can be set by module parameters */
/* DMA usage flag */
static int use_dma = 1;
/* packet per buffer dma */
static int use_dma_ppb = 1;
/* with per descr. update */
static int use_dma_ppb_du = 0;
/* buffer fill mode */
static int use_dma_bufferfill_mode = 0;
/* full speed only mode */
static int use_fullspeed = 0;
/* tx buffer size for high speed */
static unsigned long hs_tx_buf = UDC_EPIN_BUFF_SIZE;

/* module parameters */
MODULE_PARM (use_dma, "i");
MODULE_PARM_DESC (use_dma, "true to use dma");
MODULE_PARM (use_dma_ppb, "i");
MODULE_PARM_DESC (use_dma_ppb, "true to use dma in packet per buffer mode");
MODULE_PARM (use_dma_ppb_du, "i");
MODULE_PARM_DESC (use_dma_ppb_du, "true to use dma in packet per buffer mode with descriptor update");
MODULE_PARM (use_fullspeed, "i");
MODULE_PARM_DESC (use_fullspeed, "true to use fullspeed only");
MODULE_PARM (hs_tx_buf, "l");
MODULE_PARM_DESC (hs_tx_buf, "high speed tx buffer size for data endpoints in dwords");

MODULE_DESCRIPTION (DRIVER_DESC);
MODULE_AUTHOR ("Thomas Dahlmann");
MODULE_LICENSE("GPL");

#ifndef CONFIG_USB_NON_PCI_OTGDEVICE
MODULE_DEVICE_TABLE (pci, pci_ids);
#endif

/*****************************************************************************
 *  Function Definitions
 *****************************************************************************/

/* AU specific */
#ifdef CONFIG_SOC_AU1200
void au1200_sync(void)
{
        au1200_sync_delay();
}
#endif

/* Debugging ----------------------------------------------------------------*/
#if defined(USE_SYSFS_DEBUG_FILES) || defined (DEBUG)
/**
 * Gets traffic type string of endpoint
 * 
 * \param bmAttributes  endpoint transfer type
 * \return endpoint transfer type
 */
static char *type_string (u8 bmAttributes)
{
        switch ((bmAttributes) & USB_ENDPOINT_XFERTYPE_MASK) {
        case USB_ENDPOINT_XFER_BULK:    return "bulk";
        case USB_ENDPOINT_XFER_ISOC:    return "iso";
        case USB_ENDPOINT_XFER_INT:     return "intr";
        };
        return "control";
}
#endif

/**
 * Dumps dword read access
 * 
 * \param a address pointer
 * \return read dword
 */
#ifdef UDC_REGISTER_DUMP
#ifdef VERBOSE
unsigned int read32(void* a)
{
        unsigned int  d;
        d = *(volatile unsigned int *) __io_virt(a);
        DBG("read32 : address=%08lx data=%08lx\n",a,d);
        return d;

}
/**
 * Dumps byte read access
 * 
 * \param a address pointer
 * \return read byte
 */
unsigned char read8(void* a)
{
        unsigned char  d;
        d = *(volatile unsigned char *) __io_virt(a);
        DBG("read8  : address=%08lx data=%08x\n",a,d);
        return d;

}
#endif
#endif

#ifdef DEBUG
/**
 * Dumps byte read access
 * 
 * \param a address pointer
 * \return read byte
 */
static void print_td(struct udc_data_dma *p)
{
        INFO("td = %08lx: dmastat=%08lx bufptr=%08lx next=%08lx\n", 
             (unsigned long) p,
             (unsigned long) p->dmastat,
             (unsigned long) p->bufptr,
             (unsigned long) p->next);
}
#endif

/* printing registers --------------------------------------------------------*/
/**
 * Prints UDC device registers and endpoint irq registers
 * 
 * \param dev pointer to device struct 
 */
static void print_regs(struct udc* dev)
{
#ifndef UDC_IPBUG_3943_WORKAROUND
        DBG( "------- Device registers -------\n");
        DBG( "dev config     = %08lx\n", (unsigned long) dev->regs->cfg);
        DBG( "dev control    = %08lx\n", (unsigned long) dev->regs->ctl);
        DBG( "dev status     = %08lx\n", (unsigned long) dev->regs->sts);
        DBG( "\n");
        DBG( "dev int's      = %08lx\n", (unsigned long) dev->regs->irqsts);
        DBG( "dev intmask    = %08lx\n", (unsigned long) dev->regs->irqmsk);
        DBG( "\n");
        DBG( "dev ep int's   = %08lx\n", (unsigned long) dev->regs->ep_irqsts);
        DBG( "dev ep intmask = %08lx\n", (unsigned long) dev->regs->ep_irqmsk);
        DBG( "\n");
#endif
        DBG( "USE DMA        = %d\n", use_dma);
        if (use_dma && use_dma_ppb && !use_dma_ppb_du) {
                DBG( "DMA mode       = PPBNDU (packet per buffer WITHOUT desc. update)\n");
#ifndef DEBUG
                INFO( "DMA mode (PPBNDU)\n");
#endif
        } 
        else if (use_dma && use_dma_ppb_du && use_dma_ppb_du) {
                DBG( "DMA mode       = PPBDU (packet per buffer WITH desc. update)\n");
#ifndef DEBUG
                INFO( "DMA mode (PPBDU)\n");
#endif
        } 
        if (use_dma && use_dma_bufferfill_mode) {
                DBG( "DMA mode       = BF (buffer fill mode)\n");
#ifndef DEBUG
                INFO( "DMA mode (BF)\n");
#endif
        } 
#ifndef DEBUG
        if (!use_dma) {
                INFO( "FIFO mode\n");
        } 
#endif
        DBG("-------------------------------------------------------\n");       
}

/**
 * Prints snapshot of ep registers
 * 
 * \param dev   pointer to device struct
 */
#ifdef DEBUG
static void print_epregs(struct udc* dev, struct udc_epregs* epregs)
{
        INFO( "ep control  = %08lx\n", (unsigned long) epregs->ctl);
        INFO( "ep status   = %08lx\n", (unsigned long) epregs->sts);
        INFO( "--------------------------------\n");
}
#endif

/**
 * Prints ep registers
 * 
 * \param dev           pointer to device struct
 * \param ep_ix         endpoint index
 */
#ifdef NOT_USED
static void print_current_epregs(struct udc* dev, int ep_ix)
{
        INFO( "------- ep %d registers -------\n", ep_ix);
        print_epregs(dev, dev->ep[ep_ix].regs);
        INFO( "--------------------------------\n");
}
#endif

/* TODO tdahlman: remove */
/**
 * Prints misc information, to be removed
 * 
 * \param dev           pointer to device struct
 */
#ifdef DEBUG
static void print_misc(struct udc* dev)
{
        print_regs(dev);
 
        /******* for DMA IN problem */  
        if (use_dma) {

                INFO("no_req=%ld no_pref_req=%ld\n", no_req, no_pref_req);
        }
}
#endif

/* driver functions ----------------------------------------------------------*/
/**
 * Called by OS for insmod 
 * 
 * \param inode         file node
 * \param               file struct
 * \return read byte
 */
#ifdef DEBUG
static int udc_open (struct inode *inode, struct file *file)
{
        int retval = 0;

#if     LINUX_VERSION_CODE < UDC_NEW_GADGET_KERNEL
        /* Increment our usage count for the module.*/
        MOD_INC_USE_COUNT;
#endif

        return retval;
}

/**
 * Called by OS for rmmod 
 * 
 * \param inode         file node
 * \param file          file struct
 * \return read byte
 */
static int udc_release (struct inode *inode, struct file *file)
{
        int retval = 0;

#if     LINUX_VERSION_CODE < UDC_NEW_GADGET_KERNEL
        /* decrement our usage count for the module */
        MOD_DEC_USE_COUNT;
#endif
        return retval;
}

/**
 * Called by OS for ioctl() from user space
 * 
 * \param inode         file node
 * \param file          file struct
 * \param cmd           ioctl command code
 * \param arg           ioctl argument
 * \return 0 if success 
 */
static int udc_ioctl (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
        struct udc* dev = the_controller;
        
        if (!dev)
                return -EINVAL;
     
        /* for testing PM */ 
        switch (cmd)
        {
                case 1:
                        udc_remote_wakeup(the_controller);
                        break;
                case 5:
                        udc_suspend(the_controller);
                        break;
                case 6:
                        udc_resume(the_controller);
                        break;
        }
 
        INFO( "ioctl called\n");

        //print_misc(dev);
        return 0;
}

/**
 *  Prints a buffers contents 
 *  \param buf          pointer to buffer
 *  \param bytes        number bytes to print
 */
#ifdef NOT_USED
static void dump_buffer(u8* buf, u32 bytes)
{
        int i;

        printk("\nbuffer %lx = %d bytes:\n", (unsigned long) buf, bytes);
        for (i = 0; i < bytes; i++) {
                printk("%02x", *(buf + i));
                if ((i+1)%16 == 0)
                        printk("\n");
        }
        printk("\n");
}
#endif
#endif

/**
 * Masks unused interrupts 
 * 
 * \param dev           pointer to device struct
 * \return 0 if success 
 */
static int udc_mask_unused_interrupts(struct udc* dev)
{
        u32 tmp;

        /* mask all dev interrupts */
        tmp =   AMD_BIT(UDC_DEVINT_SVC) |
                AMD_BIT(UDC_DEVINT_ENUM) |
                AMD_BIT(UDC_DEVINT_US) |
                AMD_BIT(UDC_DEVINT_UR) |
                AMD_BIT(UDC_DEVINT_ES) |
                AMD_BIT(UDC_DEVINT_SI) |
                AMD_BIT(UDC_DEVINT_SOF)|
                AMD_BIT(UDC_DEVINT_SC);
        writel(tmp, &dev->regs->irqmsk);        
        
        /* mask all ep interrupts */
        writel(UDC_EPINT_MSK_DISABLE_ALL, &dev->regs->ep_irqmsk);        

        return 0;
}

/**
 * Enables endpoint 0 interrupts 
 * 
 * \param dev           pointer to device struct
 * \return 0 if success 
 */
static int udc_enable_ep0_interrupts(struct udc* dev)
{
        u32 tmp;

        DBG("udc_enable_ep0_interrupts()\n");

        /* read irq mask */
        tmp = readl(&dev->regs->ep_irqmsk);
        /* enable ep0 irq's */
        tmp &= AMD_UNMASK_BIT(UDC_EPINT_IN_EP0)
               & AMD_UNMASK_BIT(UDC_EPINT_OUT_EP0);  
        writel(tmp, &dev->regs->ep_irqmsk);        

        return 0;
}

/**
 * Enables device interrupts for SET_INTF and SET_CONFIG
 * 
 * \param dev           pointer to device struct
 * \return 0 if success 
 */
static int udc_enable_dev_setup_interrupts(struct udc* dev)
{
        u32 tmp;

        DBG("enable device interrupts for setup data\n");

        /* read irq mask */
        tmp = readl(&dev->regs->irqmsk);
       
        /* enable SET_INTERFACE, SET_CONFIG and other needed irq's */
        tmp &= AMD_UNMASK_BIT(UDC_DEVINT_SI)
               & AMD_UNMASK_BIT(UDC_DEVINT_SC)  
#ifdef UDC_IPCASE_8000009363_WORKAROUND
               & AMD_UNMASK_BIT(UDC_DEVINT_SOF)  
#endif
               & AMD_UNMASK_BIT(UDC_DEVINT_UR)  
#ifndef UDC_IPBUG_3943_WORKAROUND
               & AMD_UNMASK_BIT(UDC_DEVINT_US)  
#endif
#ifndef UDC_IPBUG_3950_WORKAROUND
               & AMD_UNMASK_BIT(UDC_DEVINT_SVC)
#endif
               & AMD_UNMASK_BIT(UDC_DEVINT_ENUM);
        writel(tmp, &dev->regs->irqmsk);        

        return 0;
}

/**
 * Calculates fifo start of endpoint based on preceeding endpoints
 * 
 * \param ep           pointer to ep struct
 * \return 0 if success 
 */
static int udc_set_txfifo_addr(struct udc_ep *ep)
{
        struct udc      *dev;
        u32 tmp;
        int i;

        if (!ep || !(ep->is_in))
                return -EINVAL;
        
        dev = ep->dev;
        ep->txfifo = dev->txfifo;

        /* traverse ep's */
        for (i = 0; i < ep->num; i++) {
                if (dev->ep[i].regs) {
                        /* read fifo size */
                        tmp = readl(&dev->ep[i].regs->bufin_framenum);
                        tmp = AMD_GETBITS(tmp, UDC_EPIN_BUFF_SIZE);
                        ep->txfifo+= tmp;
                }
        }
        return 0;
}

/**
 * Enables endpoint, is called by gadget driver
 * 
 * \param ep            pointer to ep struct
 * \param desc          pointer to endpoint descriptor
 * \return 0 if success 
 */
static int
udc_enable (struct usb_ep *_ep, const struct usb_endpoint_descriptor *desc)
{
        struct udc              *dev;
        struct udc_ep   *ep;
        u32                     tmp;
        unsigned long           flags;
        u8 udc_csr_epix;

        VDBG("udc_enable()\n");        
        
        ep = container_of (_ep, struct udc_ep, ep);
        /* FIXME some gadget do not call udc_disable before udc_enable so
           ep->desc may still be set */
        //if (!_ep || !desc || ep->desc || _ep->name == ep0name
        if (!_ep || !desc || _ep->name == ep0name
                        || desc->bDescriptorType != USB_DT_ENDPOINT) {
                ERR("udc_enable: !_ep=%d !desc=%d ep->desc!=NULL=%d _ep->name==ep0name=%d desc->bDescriptorType!=USB_DT_ENDPOINT=%d\n",
                     !_ep, !desc, ep->desc != NULL, _ep->name == ep0name, desc->bDescriptorType != USB_DT_ENDPOINT);
                return -EINVAL;
        }
        
        DBG("udc_enable() ep %d\n", ep->num);

        dev = ep->dev;

        /* exit on suspend */
        if (dev->sys_suspended) 
              return -ESHUTDOWN;

        if (!dev->driver || dev->gadget.speed == USB_SPEED_UNKNOWN)
                return -ESHUTDOWN;

        spin_lock_irqsave (&dev->lock, flags);
        ep->desc = desc;

        /* ep_reset() has already been called */
        ep->stopped = 0;

        /* set traffic type */
        tmp = readl(&dev->ep[ep->num].regs->ctl);
        tmp = AMD_ADDBITS(tmp, desc->bmAttributes, UDC_EPCTL_ET);
        writel(tmp, &dev->ep[ep->num].regs->ctl);
        ep->is_iso = (desc->bmAttributes == USB_ENDPOINT_XFER_ISOC) ? 1 : 0;

        /* set max packet size */
        tmp = readl(&dev->ep[ep->num].regs->bufout_maxpkt);
        tmp = AMD_ADDBITS(tmp, desc->wMaxPacketSize, UDC_EP_MAX_PKT_SIZE);
        ep->ep.maxpacket = desc->wMaxPacketSize;
        writel(tmp, &dev->ep[ep->num].regs->bufout_maxpkt);

        /* IN ep */
        if (ep->is_in) {
                
                /* ep ix in UDC CSR register space */
                udc_csr_epix = ep->num;
                
                /* set buffer size (tx fifo entries) */
                tmp = readl(&dev->ep[ep->num].regs->bufin_framenum);
                /* double buffering: fifo size = 2 x max packet size */
                tmp = AMD_ADDBITS(
                              tmp,
                              desc->wMaxPacketSize * UDC_EPIN_BUFF_SIZE_MULT /
                                                     UDC_DWORD_BYTES,
                              UDC_EPIN_BUFF_SIZE);                 
                writel(tmp, &dev->ep[ep->num].regs->bufin_framenum);

                /* calc. tx fifo base addr */
                udc_set_txfifo_addr(ep); 

                /* flush fifo */
                tmp = readl(&ep->regs->ctl);
                tmp |= AMD_BIT(UDC_EPCTL_F); 
                writel(tmp, &ep->regs->ctl);
                
        } /* OUT ep */
        else {
                /* ep ix in UDC CSR register space */
                udc_csr_epix = ep->num - UDC_CSR_EP_OUT_IX_OFS;

                /* set max packet size UDC CSR  */
                tmp = readl(&dev->csr->ne[ep->num - UDC_CSR_EP_OUT_IX_OFS]);
                tmp = AMD_ADDBITS(tmp, desc->wMaxPacketSize, UDC_CSR_NE_MAX_PKT);
                writel(tmp, &dev->csr->ne[ep->num - UDC_CSR_EP_OUT_IX_OFS]);
        }

        /***** UDC CSR reg ****************************/
        /* set ep values  */
        tmp = readl(&dev->csr->ne[udc_csr_epix]);
        /* max packet */
        tmp = AMD_ADDBITS(tmp, desc->wMaxPacketSize, UDC_CSR_NE_MAX_PKT);
        /* ep number */
        tmp = AMD_ADDBITS(tmp, desc->bEndpointAddress, UDC_CSR_NE_NUM);
        /* ep direction */
        tmp = AMD_ADDBITS(tmp, ep->is_in, UDC_CSR_NE_DIR);
        /* ep type */
        tmp = AMD_ADDBITS(tmp, desc->bmAttributes, UDC_CSR_NE_TYPE);
        /* ep config */
        tmp = AMD_ADDBITS(tmp, ep->dev->cur_config, UDC_CSR_NE_CFG);
        /* ep interface */
        tmp = AMD_ADDBITS(tmp, ep->dev->cur_intf, UDC_CSR_NE_INTF);
        /* ep alt */
        tmp = AMD_ADDBITS(tmp, ep->dev->cur_alt, UDC_CSR_NE_ALT);
        /* write reg */
        writel(tmp, &dev->csr->ne[udc_csr_epix]);
        
        /* enable ep irq */
        tmp = readl(&dev->regs->ep_irqmsk);
        tmp &= AMD_UNMASK_BIT(ep->num);
        writel(tmp, &dev->regs->ep_irqmsk);        

        /* clear NAK by writing CNAK */
        /* avoid BNA for DMA,  dont clear NAK until DMA desc. written */
        if (!use_dma) {
                tmp = readl(&ep->regs->ctl);
                tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
                writel(tmp, &ep->regs->ctl);
                ep->naking = 0;
                UDC_QUEUE_CNAK(ep, ep->num);
        }
        tmp = desc->bEndpointAddress;
        DBG( "enabled %s (ep%d%s-%s) %s\n",
                _ep->name, tmp & 0x0f, DIR_STRING (tmp),
                type_string (desc->bmAttributes),
                ep->dma ? "dma" : "pio");

        /* pci writes may still be posted */
        spin_unlock_irqrestore (&dev->lock, flags);
        return 0;
}
/**
 * Resets endpoint
 * 
 * \param regs          pointer to device register struct
 * \param ep            pointer to endpoint
 */
static void ep_reset (struct udc_regs *regs, struct udc_ep *ep)
{
        u32             tmp;

        VDBG("ep-%d reset\n", ep->num);
        ep->desc = 0;
        INIT_LIST_HEAD (&ep->queue);

        ep->ep.maxpacket = ~0;
        ep->ep.ops = &udc_ep_ops;

        if (!(ep->dev->sys_suspended)) {
                /* set NAK  */
                tmp = readl(&ep->regs->ctl);
                tmp |= AMD_BIT(UDC_EPCTL_SNAK); 
                writel(tmp, &ep->regs->ctl);
                ep->naking = 1;
                
                /* disable interrupt */
                tmp = readl(&regs->ep_irqmsk);
                tmp |= AMD_BIT(ep->num);
                writel(tmp, &regs->ep_irqmsk);        

                if (ep->is_in) {
                        /* unset P and IN bit of potential former DMA */
                        tmp = readl(&ep->regs->ctl);
                        tmp &= AMD_UNMASK_BIT(UDC_EPCTL_P);
                        writel(tmp, &ep->regs->ctl);

                        tmp = readl(&ep->regs->sts);
                        tmp |= AMD_BIT(UDC_EPSTS_IN);
                        writel(tmp, &ep->regs->sts);

                        /* flush the fifo */
                        tmp = readl(&ep->regs->ctl);
                        tmp |= AMD_BIT(UDC_EPCTL_F); 
                        writel(tmp, &ep->regs->ctl);

                }
                /* reset desc pointer */
                writel(0, &ep->regs->desptr);
        }


}

/**
 * Disables endpoint, is called by gadget driver
 * 
 * \param ep            pointer to ep struct
 * \return 0 if success 
 */
static int udc_disable (struct usb_ep *_ep)
{
        struct udc_ep   *ep = NULL;
        unsigned long           flags;

        ep = container_of (_ep, struct udc_ep, ep);
        if (!_ep || !ep->desc || _ep->name == ep0name)
                return -EINVAL;

        DBG("Disable ep-%d\n", ep->num);

        spin_lock_irqsave (&ep->dev->lock, flags);
        nuke (ep);
        ep_reset (ep->dev->regs, ep);

        VDBG( "disabled %s %s\n",
                        ep->dma ? "dma" : "pio", _ep->name);

        spin_unlock_irqrestore (&ep->dev->lock, flags);
        return 0;
}

/**
 * Allocates request packet, called by gadget driver
 * 
 * \param _ep           pointer to usb ep struct
 * \param gfp_flags     flags for kmalloc
 * \return allocated request packet, 0 if error
 */
static struct usb_request *
udc_alloc_request (struct usb_ep *_ep, int gfp_flags)
{
        struct udc_ep   *ep;
        struct udc_request      *req;

        VDBG("udc_alloc_req()\n");
        if (!_ep)
                return 0;
        ep = container_of (_ep, struct udc_ep, ep);
        /* TODO gfp flags used for dma chaining later - find better way */
        td_gfp_flags = gfp_flags;

        VDBG("udc_alloc_req(): ep%d\n", ep->num);
        req = kmalloc (sizeof *req, gfp_flags);
        if (!req)
                return 0;

        memset (req, 0, sizeof *req);
        req->req.dma = DMA_ADDR_INVALID;
        INIT_LIST_HEAD (&req->queue);

        if (ep->dma) {
                struct udc_data_dma     *td;

#if defined(CONFIG_MIPS)
                gfp_flags = GFP_ATOMIC | GFP_DMA;
#endif
                /* FIXME ep0 in requests are allocated from data pool here */
                td = pci_pool_alloc (ep->dev->data_requests, gfp_flags,
                                &req->td_dma);
                if (!td) {
                        kfree (req);
                        return 0;
                }

                VDBG("udc_alloc_req: req = %lx td = %lx, req->td_dma = %lx\n",
                     (unsigned long) req, (unsigned long) td, (unsigned long)req->td_dma);
                /* prevent from using desc. - set HOST BUSY */
                td->dmastat = AMD_ADDBITS(td->dmastat,
                                          UDC_DMA_STP_STS_BS_HOST_BUSY,
                                          UDC_DMA_STP_STS_BS);  
                td->bufptr = __constant_cpu_to_le32 (DMA_ADDR_INVALID);
                req->td_data = td;
                req->td_data_last = NULL;
                req->chain_len = 1;
        }
        return &req->req;
}

/**
 * Frees request packet, called by gadget driver
 * 
 * \param _ep   pointer to usb ep struct
 * \param _req  pointer to request packet to be freed
 */
static void
udc_free_request (struct usb_ep *_ep, struct usb_request *_req)
{
        struct udc_ep   *ep;
        struct udc_request      *req;

        if (!_ep || !_req)
                return;

        ep = container_of (_ep, struct udc_ep, ep);
        req = container_of (_req, struct udc_request, req);
        VDBG("free_req req=%lx\n", (unsigned long) req);
        WARN_ON (!list_empty (&req->queue));
        if (req->td_data) {
                VDBG("req->td_data=%lx\n", (unsigned long) req->td_data);

#ifdef UDC_IPDEFECT_9000004946_WORKAROUND
                /* re-link broken chain  */
                if (req->td_data_last) {
                        req->td_data_last->next = req->td_data_last_next;
                }
#endif
                /* free dma chain if created */
                if (req->chain_len > 1) {
                        udc_free_dma_chain(ep->dev, req);
                }

                pci_pool_free (ep->dev->data_requests, req->td_data, req->td_dma);
        }
        kfree (req);
}

/*-------------------------------------------------------------------------*/
#undef USE_KMALLOC

/* many common platforms have dma-coherent caches, which means that it's
 * safe to use kmalloc() memory for all i/o buffers without using any
 * cache flushing calls.  (unless you're trying to share cache lines
 * between dma and non-dma activities, which is a slow idea in any case.)
 *
 * other platforms need more care, with 2.5 having a moderately general
 * solution (which falls down for allocations smaller than one page)
 * that improves significantly on the 2.4 PCI allocators by removing
 * the restriction that memory never be freed in_interrupt().
 */
#if     defined(CONFIG_X86)
#define USE_KMALLOC

#elif   defined(CONFIG_PPC) && !defined(CONFIG_NOT_COHERENT_CACHE)
#define USE_KMALLOC

/* FIXME TMP26: cached dma buffers seem no to work for 2.6.x to be
 *  examined */
#elif   defined(CONFIG_MIPS)
/* MIPS kernel 2.4 */
#if     LINUX_VERSION_CODE < UDC_NEW_GADGET_KERNEL 
#if !defined(CONFIG_NONCOHERENT_IO)
#define USE_KMALLOC
#endif /* !defined(CONFIG_NONCOHERENT_IO) */
#endif /* LINUX_VERSION_CODE < UDC_NEW_GADGET_KERNEL */
#if   LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
#ifdef CONFIG_DMA_COHERENT
#define USE_KMALLOC
#endif /* CONFIG_DMA_COHERENT */
#endif /* LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL */
/* FIXME there are other cases, including an x86-64 one ...  */
#endif

/* allocating buffers this way eliminates dma mapping overhead, which
 * on some platforms will mean eliminating a per-io buffer copy.  with
 * some kinds of system caches, further tweaks may still be needed.
 */

/**
 * Allocated data buffer for request packet, called by gadget driver
 * 
 * \param _ep           pointer to usb ep struct
 * \param bytes         number bytes to allocate
 * \param dma           pointer to dma struct
 * \param gfp_flags     flags for allocate 
 * \return pointer to allocated buffer, 0 if error
 */
static void *
udc_alloc_buffer (
        struct usb_ep           *_ep,
        unsigned                bytes,
        dma_addr_t              *dma,
        int                     gfp_flags
)
{
        void                    *retval;
        struct udc_ep   *ep;
        VDBG("alloc_buffer(): ep-%s\n", _ep->name);
        ep = container_of (_ep, struct udc_ep, ep);
        if (!_ep)
                return 0;
        *dma = DMA_ADDR_INVALID;

        if (ep->dma) {
#if     defined(USE_KMALLOC)

#if defined(CONFIG_MIPS)
                gfp_flags = GFP_ATOMIC | GFP_DMA;
#endif
                retval = kmalloc (bytes, gfp_flags);

                if (retval)
                        *dma = virt_to_phys (retval);

#elif   LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
//DISABLEDamd #warning Using dma_alloc_coherent even with sub-page allocations
                /* the main problem with this call is that it wastes memory
                 * on typical 1/N page allocations: it allocates 1-N pages.
                 */
                retval = dma_alloc_coherent (&ep->dev->pdev->dev,
                                bytes, dma, gfp_flags);
#elif defined (CONFIG_MIPS)
//DISABLEDamd #warning Using pci_alloc_consistent even with buffers smaller than a page.
                retval = pci_alloc_consistent(ep->dev->pdev, bytes, dma);
#else
#error No dma-coherent memory allocator is available
                /* pci_alloc_consistent works, but pci_free_consistent()
                 * isn't safe in_interrupt().  plus, in addition to the
                 * 1/Nth page weakness, it doesn't understand gfp_flags.
                 */
#endif
        } else
                retval = kmalloc (bytes, gfp_flags);

        VDBG("alloc_buffer() retval = %lx dma=%lx\n", (unsigned long) retval, (unsigned long) *dma);
        return retval;
}


/**
 * Free data buffer for request packet, called by gadget driver
 * 
 * \param _ep           pointer to usb ep struct
 * \param buf           pointer to buffer to be freed
 * \param dma           pointer to dma struct
 * \param bytes         number bytes to be freed 
 * \return pointer to allocated buffer, 0 if error
 */
static void
udc_free_buffer (
        struct usb_ep *_ep,
        void *buf,
        dma_addr_t dma,
        unsigned bytes
) {
        struct udc_ep   *ep;

        VDBG ("ep %s - free buffer %lx\n", _ep->name, (unsigned long) buf);

        /* null pointer ? */
        if (!buf) {
                return;
        }

        ep = container_of(_ep, struct udc_ep, ep);
        if (!ep)
                return;

        /* free memory into the right allocator */
#ifndef USE_KMALLOC
        if (dma != DMA_ADDR_INVALID) {
#if   LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
                dma_free_coherent (&ep->dev->pdev->dev, bytes, buf, dma);
#else
                /* one problem with this call is that some platforms
                 * don't allow it to be used in_irq().
                 */
                pci_free_consistent(ep->dev->pdev, bytes, buf, dma);
#endif
        } else
#endif
        {
                if (!(ep->dma))
                        kfree (buf);
        }

}

/**
 * Write data to TX fifo for IN packets 
 * 
 * \param ep            pointer to ep struct
 * \param req           pointer to request packet
 * \return allocated request packet, 0 if error
 */
static void
write_fifo (struct udc_ep *ep, struct usb_request *req)
{
        u8                      *buf;
        unsigned                count, total;

        VDBG("write_fifo()\n");
        if (req) {
                buf = req->buf + req->actual;
                prefetch (buf);
                total = req->length - req->actual;
        } else {
                total = 0;
                buf = 0;
        }
        
        /* write just one packet at a time */
        count = ep->ep.maxpacket;
	if (count > total)
		count = total;
	
        /* TODO tdahlman - get dev pointer from ep */
        udc_write_fifo(the_controller, ep->num, (u32*) buf, count); 
}

/**
 * Read dwords from RX fifo for OUT transfers
 * 
 * \param dev           pointer to device struct
 * \param buf           pointer to buffer to be filled
 * \param dwords        number of dwords to be read
 * \return allocated request packet, 0 if error
 */
static int udc_read_fifo_dwords(struct udc* dev, u32* buf, int dwords)
{
        int i;
        
        VDBG("udc_read_dwords(): %d dwords\n", dwords);

        for (i = 0; i < dwords; i++)
        {
                *(buf + i) = readl(dev->rxfifo);
        }        
        return 0;
}

/**
 * Read bytes from RX fifo for OUT transfers
 * 
 * \param dev           pointer to device struct
 * \param buf           pointer to buffer to be filled
 * \param bytes         number of bytes to be read
 * \return allocated request packet, 0 if error
 */
static int udc_read_fifo_bytes(struct udc* dev, u8* buf, int bytes)
{
        int i,j;
        u32 tmp;
        
        VDBG("udc_read_bytes(): %d bytes\n", bytes);

        /* dwords first */
        for (i = 0; i < bytes / UDC_DWORD_BYTES; i++) {
               *((u32*) (buf + (i<<2))) = readl(dev->rxfifo); 
        }

        /* remaining bytes must be read by byte access */
        if (bytes % UDC_DWORD_BYTES) {
                tmp = readl(dev->rxfifo);
                for (j = 0; j < bytes % UDC_DWORD_BYTES; j++) {
                        *(buf + (i<<2) + j) = (u8) (tmp & UDC_BYTE_MASK);
                        tmp = tmp >> UDC_BITS_PER_BYTE;
                }
        }
        
        return 0;
}

/**
 * Write bytes to TX fifo for IN packets 
 * 
 * \param dev           pointer to device struct
 * \param ep_ix         endpoint index
 * \param buf           pointer to buffer to be write 
 * \param bytes         number of bytes to be write 
 * \return allocated request packet, 0 if error
 */
static int udc_write_fifo(struct udc* dev, int ep_ix, u32* buf, int bytes)
{
        int i,j;
        struct udc_ep *ep;

        VDBG("udc_write_fifo: %d bytes\n", bytes);

        ep = &dev->ep[ep_ix];

        /* dwords first */
        for (i = 0; i < bytes / UDC_DWORD_BYTES; i++) {
                writel(*(buf + i), ep->txfifo);
        }

        /* remaining bytes must be written by byte access */
        for (j = 0; j < bytes % UDC_DWORD_BYTES; j++) {
                writeb((u8) (*(buf + i) >> (j << UDC_BITS_PER_BYTE_SHIFT)),
                       ep->txfifo);
        }

#ifdef UDC_IPBUG_2253_WORKAROUND
        {
                u32 tmp;

                /* NAK if small packet until write confirm completed */
                if (bytes < UDC_SMALL_PACKET) {
                        /* set NAK */
                        tmp = readl(&ep->regs->ctl);
                        tmp |= AMD_BIT(UDC_EPCTL_SNAK); 
                        writel(tmp, &ep->regs->ctl);
                        ep->naking = 1;
                        wb_flush();
                }

                /* dummy write confirm */
                writel(0, &ep->regs->confirm); 

                /* stop NAKing after small packet DMA */
                if (ep->naking) {
                        /* clear NAK by writing CNAK */
                        tmp = readl(&ep->regs->ctl);
                        tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
                        writel(tmp, &ep->regs->ctl);
                        ep->naking = 0;
                        UDC_QUEUE_CNAK(ep, ep->num);
                }
        }
#else
        /* dummy write confirm */
        writel(0, &ep->regs->confirm); 
#endif
        
        return 0;
}

/**
 * Read data from RX fifo for OUT transfers
 * 
 * \param ep            pointer to ep struct
 * \param req           pointer to request packet
 * \return true if request completes for short or max packet, false otherwise   
 */
static int
read_fifo (struct udc_ep *ep, struct udc_request *req)
{
        u8                      *buf = req->req.buf + req->req.actual;
        unsigned                count=0, tmp, is_short;
        unsigned                cleanup = 0;


        /* TODO needed ? */
        prefetchw (buf);
        
        /* received number bytes */
        count = readl(&ep->regs->sts);
        count = AMD_GETBITS(count, UDC_EPSTS_RX_PKT_SIZE);
        
        tmp = req->req.length - req->req.actual;
        if (count > tmp) {
                /* as with DMA, data overflow gets flushed */
                if ((tmp % ep->ep.maxpacket) != 0) {
                        ERR( "%s out fifo %d bytes, expected %d\n",
                                ep->ep.name, count, tmp);
                        req->req.status = -EOVERFLOW;
                        cleanup = 1;
                }
                count = tmp;
        }
        req->req.actual += count;
        is_short = (count == 0) || ((count % ep->ep.maxpacket) != 0);
        VDBG( "read %s fifo (OUT) %d bytes%s%s req %p %d/%d\n",
                        ep->ep.name, count, is_short ? " (short)" : "",
                        cleanup ? " flush" : "", 
                        req, req->req.actual, req->req.length);
        
        /* read rx fifo bytes */
        udc_read_fifo_bytes(ep->dev, buf, count); 
       
        return is_short || ((req->req.actual == req->req.length)
                                && !req->req.zero);
}

/**
 *  Temporary left in the code for later reusing or removing
 */
static inline void spin_stop_dma (u32 *dma)
{
}

/**
 *  Temporary left in the code for later reusing or removing
 */
static inline void stop_dma (u32 *dma)
{
        spin_stop_dma (dma);
}

/**
 * create/re-init a DMA descriptor or a DMA descriptor chain
 *
 * \param ep            pointer to endpoint struct 
 * \param req           pointer to request packet
 */
static int prep_dma (struct udc_ep *ep, struct udc_request *req)
{
        int retval = 0;
        u32                     tmp;
        VDBG("prep_dma\n");
        VDBG("prep_dma ep%d req->td_data=%lx\n",
              ep->num, (unsigned long) req->td_data);

        /* set buffer pointer */
        req->td_data->bufptr   = req->req.dma;

        /* set last bit */
        req->td_data->dmastat |= AMD_BIT(UDC_DMA_IN_STS_L);
       
        /* build/re-init dma chain if maxpkt scatter mode, not for EP0 */
        if (use_dma_ppb && ep->num != UDC_EP0OUT_IX && ep->num != UDC_EP0IN_IX) {

                retval = udc_create_dma_chain(ep, req, ep->ep.maxpacket, td_gfp_flags);
                if (retval != 0)
                {
                        if (retval == -ENOMEM)
                                INFO("Out of DMA memory (allocation failed)\n");       
                        return retval;
                }
                /*TODO better place ? */
                if (ep->is_in) {
                        if (req->req.length == ep->ep.maxpacket) {
                                /* write tx bytes */
                                req->td_data->dmastat = AMD_ADDBITS(req->td_data->dmastat,
                                                ep->ep.maxpacket,
                                                UDC_DMA_IN_STS_TXBYTES); 

                        }
                }
                
        }
#ifdef UDC_IPDEFECT_9000004946_WORKAROUND
        else { /* EP0 */
                /* point to itself */
                req->td_data->next = req->td_dma;
        }
#endif
        
        if (ep->is_in) {
                VDBG("IN: use_dma_ppb=%d req->req.length=%d ep->ep.maxpacket=%d ep%d\n",
                      use_dma_ppb, req->req.length, ep->ep.maxpacket, ep->num);
                /* if bytes < max packet then tx bytes must */
                /* be written in packet per buffer mode */
                if (!use_dma_ppb || req->req.length < ep->ep.maxpacket
                    || ep->num == UDC_EP0OUT_IX || ep->num == UDC_EP0IN_IX) {
                        /* write tx bytes */
                        req->td_data->dmastat = AMD_ADDBITS(req->td_data->dmastat,
                                                            req->req.length,
                                                            UDC_DMA_IN_STS_TXBYTES); 
                        /* reset frame num */
                        req->td_data->dmastat = AMD_ADDBITS(req->td_data->dmastat,
                                                            0,
                                                            UDC_DMA_IN_STS_FRAMENUM); 
                }
                /* set HOST BUSY */
                req->td_data->dmastat 
                      = AMD_ADDBITS(req->td_data->dmastat,
                                    UDC_DMA_STP_STS_BS_HOST_BUSY,
                                    UDC_DMA_STP_STS_BS);
        }
        else {
                VDBG("OUT set host ready\n");
                /* set HOST READY */
                req->td_data->dmastat 
                      = AMD_ADDBITS(req->td_data->dmastat,
                                    UDC_DMA_STP_STS_BS_HOST_READY,
                                    UDC_DMA_STP_STS_BS);

 
                        /* clear NAK by writing CNAK */
                        if (ep->naking) {
                                tmp = readl(&ep->regs->ctl);
                                tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
                                writel(tmp, &ep->regs->ctl);
                                ep->naking = 0;
                                UDC_QUEUE_CNAK(ep, ep->num);
                        }

        }

        return retval;
}

/**
 * Completes request packet 
 * 
 * \param ep            pointer to ep struct
 * \param req           pointer to request packet
 * \param status        status of request
 */
static void
done (struct udc_ep *ep, struct udc_request *req, int status)
{
        struct udc              *dev;
        unsigned                stopped = ep->stopped;

        VDBG("done(): ep%d\n", ep->num);

        list_del_init (&req->queue);

        if (req->req.status == -EINPROGRESS)
                req->req.status = status;
        else
                status = req->req.status;

        dev = ep->dev;
        if (req->mapped) {
                pci_unmap_single (dev->pdev, req->req.dma, req->req.length,
                        ep->is_in ? PCI_DMA_TODEVICE : PCI_DMA_FROMDEVICE);
                req->req.dma = DMA_ADDR_INVALID;
                req->mapped = 0;
        }

        if (status && status != -ESHUTDOWN)
                VDBG( "complete %s req %p stat %d len %u/%u\n",
                        ep->ep.name, &req->req, status,
                        req->req.actual, req->req.length);

        /* don't modify queue heads during completion callback */
        ep->stopped = 1;
        VDBG( "complete %s req %p stat %d len %u/%u\n",
                        ep->ep.name, &req->req, status,
                        req->req.actual, req->req.length);
        if (spin_is_locked(&dev->lock)) {
                spin_unlock (&dev->lock);
                req->req.complete (&ep->ep, &req->req);
                spin_lock (&dev->lock);
        }
        else {
                req->req.complete (&ep->ep, &req->req);
        }

        ep->stopped = stopped;
}

/**
 * frees pci pool descriptors of a DMA chain 
 * 
 * \param dev           pointer to device struct
 * \param req           pointer to request packet
 * \return 0 if success
 */
static int udc_free_dma_chain(struct udc* dev, struct udc_request* req)
{

        int ret_val = 0;
        struct udc_data_dma     *td;
        struct udc_data_dma     *td_last = NULL;
        unsigned int i;

        DBG("free chain req = %lx\n", (unsigned long) req);

#ifdef UDC_IPDEFECT_9000004946_WORKAROUND
        /* re-link broken chain  */
        if (req->td_data_last) {
                req->td_data_last->next = req->td_data_last_next;
        }
#endif
        /* do not free first desc., will be done by free for request */
        td_last = req->td_data;
#if defined (CONFIG_MIPS) && !defined(CONFIG_DMA_COHERENT)
        td = UNCAC_ADDR(phys_to_virt(td_last->next));
#else
        td = phys_to_virt(td_last->next);
#endif

        for (i = 1; i < req->chain_len; i++) {

                pci_pool_free (dev->data_requests, td,
                                (dma_addr_t) td_last->next);
                td_last = td;
#if defined (CONFIG_MIPS) && !defined(CONFIG_DMA_COHERENT)
                td = UNCAC_ADDR(phys_to_virt(td_last->next));
#else
                td = phys_to_virt(td_last->next);
#endif
        }

#ifdef UDC_IPDEFECT_9000004946_WORKAROUND
        req->td_data_last = NULL;
#endif
        return ret_val;        
}

/**
 * Iterates to the end of a DMA chain and returns last descriptor 
 * 
 * \param req           pointer to request packet
 * \return pointer to last descriptori of chain
 */
static struct udc_data_dma* udc_get_last_dma_desc(struct udc_request* req)
{
        struct udc_data_dma     *td;

        td = req->td_data;
        while (td && !(td->dmastat & AMD_BIT(UDC_DMA_IN_STS_L))) {
#if defined (CONFIG_MIPS) && !defined(CONFIG_DMA_COHERENT)
                td = UNCAC_ADDR(phys_to_virt(td->next));
#else
                td = phys_to_virt(td->next);
#endif
        }

        return td;

}

/**
 * Iterates to the end of a DMA chain and counts bytes received 
 * 
 * \param req           pointer to request packet
 * \return number of received bytes
 */
static u32 udc_get_ppbdu_rxbytes(struct udc_request* req)
{
        struct udc_data_dma     *td;
        u32 count;

        td = req->td_data;
        /* received number bytes */
        count = AMD_GETBITS(td->dmastat, UDC_DMA_OUT_STS_RXBYTES); 
        
        while (td && !(td->dmastat & AMD_BIT(UDC_DMA_IN_STS_L))) {
#if defined (CONFIG_MIPS) && !defined(CONFIG_DMA_COHERENT)
                td = UNCAC_ADDR(phys_to_virt(td->next));
#else
                td = phys_to_virt(td->next);
#endif
                /* received number bytes */
                if (td) {
                        count += AMD_GETBITS(td->dmastat, UDC_DMA_OUT_STS_RXBYTES);
                }
        }

        return count;

}

/**
 * Creates or re-inits a DMA chain 
 *
 * \param ep            pointer to endpoint struct 
 * \param req           pointer to request packet
 * \param buf_len       number of buffer bytes per descriptor (except last short)
 */
static int udc_create_dma_chain(struct udc_ep* ep, struct udc_request* req, unsigned long buf_len, int gfp_flags)
{
        unsigned long bytes = req->req.length;
        unsigned int i;
        dma_addr_t dma_addr;
        struct udc_data_dma     *td = NULL;
        struct udc_data_dma     *last = NULL;
        unsigned long txbytes;
        unsigned create_new_chain = 0;
        unsigned len;
#ifdef UDC_IPDEFECT_9000004946_WORKAROUND
        dma_addr_t last_next = DMA_ADDR_INVALID;
#endif

        VDBG("udc_create_dma_chain: bytes=%ld buf_len=%ld\n", bytes, buf_len);
        dma_addr = DMA_ADDR_INVALID;

        /* unset L bit in first desc for OUT */
        if (!ep->is_in) {
                req->td_data->dmastat &= AMD_CLEAR_BIT(UDC_DMA_IN_STS_L);
        }
        
        /* alloc only new desc's if not already available */
        len = req->req.length / ep->ep.maxpacket;
        if (req->req.length % ep->ep.maxpacket) {
                len++;
        }

        if (len > req->chain_len) {
                /* shorter chain already allocated before */
                if (req->chain_len > 1) {
                        udc_free_dma_chain(ep->dev, req);
                }
                req->chain_len = len;
                create_new_chain = 1;
        }

#ifdef UDC_IPDEFECT_9000004946_WORKAROUND
        /* re-link broken chain  */
        if (req->td_data_last) {
                req->td_data_last->next = req->td_data_last_next;
        }
        /* if only one td then last_next is root td */
        last_next = req->td_dma;
#endif

        td = req->td_data; 
        /* gen. required number of descriptors and buffers */
        for (i = buf_len; i < bytes; i += buf_len)
        {
                /* create or determine next desc. */
                if (create_new_chain) {

#if defined(CONFIG_MIPS)
                        gfp_flags = GFP_ATOMIC | GFP_DMA;
#endif
                        td = pci_pool_alloc (ep->dev->data_requests, gfp_flags,
                                             &dma_addr);
                        if (!td) 
                                return -ENOMEM;

                        td->dmastat = 0;
                }
                else if (i == buf_len)
                {
                        /* first td */
#if defined (CONFIG_MIPS) && !defined(CONFIG_DMA_COHERENT)
                        td = (struct udc_data_dma*) UNCAC_ADDR(phys_to_virt(req->td_data->next));
#else
                        td = (struct udc_data_dma*) phys_to_virt(req->td_data->next);
#endif
                        td->dmastat = 0;
                }
                else {
#if defined (CONFIG_MIPS) && !defined(CONFIG_DMA_COHERENT)
                        td = (struct udc_data_dma*) UNCAC_ADDR(phys_to_virt(last->next));
#else
                        td = (struct udc_data_dma*) phys_to_virt(last->next);
#endif
                        td->dmastat = 0;
                }
                        

                if (td) {
                        /* assign buffer */
                       td->bufptr = req->req.dma + i; 
                }
                else {
                        break;
                }

                /* short packet ? */
                if ((bytes - i) >= buf_len) {
                        txbytes = buf_len;
                }
                else {
                        /* short packet */
                        txbytes = bytes - i;
                }
         
                /* link td and assign tx bytes */
                if (i == buf_len) {
                        if (create_new_chain) {
                                req->td_data->next = dma_addr;
                        }
                        else {
                                //req->td_data->next = virt_to_phys(td);
                        }
                        /* write tx bytes */
                        if (ep->is_in) {
                                /* first desc */
                                req->td_data->dmastat = AMD_ADDBITS(req->td_data->dmastat,
                                                                    ep->ep.maxpacket,
                                                                    UDC_DMA_IN_STS_TXBYTES);
                                /* second desc */
                                td->dmastat = AMD_ADDBITS(td->dmastat,
                                                          txbytes,
                                                          UDC_DMA_IN_STS_TXBYTES); 
                        } 

#ifdef UDC_IPDEFECT_9000004946_WORKAROUND
                        last_next = req->td_data->next;
#endif
                }
                else {
                        if (create_new_chain) {
                                last->next = dma_addr;
                        }
                        else { 
                                //last->next = virt_to_phys(td);
                        }
                        if (ep->is_in) {
                                /* write tx bytes */
                                td->dmastat = AMD_ADDBITS(td->dmastat,
                                                          txbytes,
                                                          UDC_DMA_IN_STS_TXBYTES); 
                        }
#ifdef UDC_IPDEFECT_9000004946_WORKAROUND
                        last_next = last->next;
#endif
                }
                last = td;
        }
        /* set last bit */
        if (td) {
                td->dmastat |= AMD_BIT(UDC_DMA_IN_STS_L);
                /* last desc. points to itself */
#ifdef UDC_IPDEFECT_9000004946_WORKAROUND
                /* remember broken chain next pointer */
                req->td_data_last_next = td->next;
                /* point to itself */
                td->next = last_next;
#endif
                req->td_data_last = td;
        }

        return 0;
}

/**
 * Queues a request packet, called by gadget driver
 * 
 * \param _ep           pointer to usb ep struct
 * \param _req          pointer to request packet to be freed
 * \param gfp_flags     flags for alloc
 * \return 0 if success
 */
static int
udc_queue (struct usb_ep *_ep, struct usb_request *_req, int gfp_flags)
{
        int retval = 0;
        struct udc_request      *req;
        struct udc_ep   *ep;
        struct udc              *dev;
        unsigned long           flags;
        u32 tmp;
        

        VDBG ("udc_queue()\n");

        /* check the inputs */
        req = container_of (_req, struct udc_request, req);
        VDBG("!req=%d !compl=%d !buf=%d !empty_list=%d \n",
                !_req, !_req->complete, !_req->buf, !list_empty(&req->queue));

        if (!_req || !_req->complete || !_req->buf
                        || !list_empty (&req->queue))
                return -EINVAL;

        ep = container_of (_ep, struct udc_ep, ep);
        if (!_ep || (!ep->desc && ep->num != 0))
                return -EINVAL;

        VDBG("udc_queue(): ep%d-in=%d\n", ep->num, ep->is_in);
        dev = ep->dev;

        /* exit on suspend */
        if (dev->sys_suspended) 
              return -ESHUTDOWN;

        if (!dev->driver || dev->gadget.speed == USB_SPEED_UNKNOWN)
                return -ESHUTDOWN;

        /* set up dma mapping in case the caller didn't */
        if (ep->dma && _req->length != 0 && _req->dma == DMA_ADDR_INVALID) {
                VDBG("DMA unmapped: mapping req %lx\n", (unsigned long) req);
                _req->dma = pci_map_single (dev->pdev, _req->buf, _req->length,
                        ep->is_in ? PCI_DMA_TODEVICE : PCI_DMA_FROMDEVICE);
                req->mapped = 1;
        }

        VDBG( "%s queue req %p, len %d req->td_data=%lx buf %p\n",
                        _ep->name, _req, _req->length, (unsigned long) req->td_data, _req->buf);

        spin_lock_irqsave (&dev->lock, flags);

        _req->status = -EINPROGRESS;
        _req->actual = 0;
        req->dma_done = 0;

        /* kickstart this i/o queue? */
        if (list_empty (&ep->queue) && !ep->stopped) {
                /* maybe there's no control data, just status ack */
                if (ep->is_in && _req->length == 0) {
                        //allow_status (ep);
                        done (ep, req, 0);
                        VDBG( "%s status ack\n", ep->ep.name);
                        goto done;
                }
                /* use DMA if the endpoint supports it, else pio */
                if (ep->dma) {

                        retval = prep_dma(ep, req);
                        if (retval != 0)
                                goto done;
                        /* write desc pointer to enable DMA */
                        if (ep->is_in) {
                                /* set HOST READY */
                                req->td_data->dmastat 
                                      = AMD_ADDBITS(req->td_data->dmastat,
                                                    UDC_DMA_IN_STS_BS_HOST_READY, 
                                                    UDC_DMA_IN_STS_BS);
                        }

                        /* write desc pointer */
                        writel(req->td_dma, &ep->regs->desptr);
                        /* clear NAK by writing CNAK */
                        if (ep->naking) {
                                tmp = readl(&ep->regs->ctl);
                                tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
                                writel(tmp, &ep->regs->ctl);
                                ep->naking = 0;
                                UDC_QUEUE_CNAK(ep, ep->num);
                        }
                
                        if (!ep->is_in) {
                                /* enable DMA */
                                /* set RDE */
                                tmp = readl(&dev->regs->ctl);
                                tmp |= AMD_BIT(UDC_DEVCTL_RDE);
                                writel(tmp, &dev->regs->ctl);                
                                tmp = readl(&dev->regs->ctl);
                        }
#ifdef UDC_DISABLE_IRQ_IF_EMPTY_IN_QUEUE
                        else {
                                /* enable ep irq */
                                tmp = readl(&dev->regs->ep_irqmsk);
                                tmp &= AMD_UNMASK_BIT(ep->num);
                                writel(tmp, &dev->regs->ep_irqmsk);        
                        }
#endif
                }
                else {
                        /* maybe there's no control data, just status ack */
                        /*if (ep->num == 0 && _req->length == 0) {
                                allow_status (ep);
                                done (ep, req, 0);
                                VDBG( "%s status ack\n", ep->ep.name);
                                goto done;
                        }*/
                }

        } else if (ep->dma) {

                /* TODO use prep_dma also for OUT ep's, this is not possible
                   for PPB modes currently, because of chain creation reasons */
                if (ep->is_in) {
                        retval = prep_dma(ep, req);
                        if (retval != 0)
                                goto done;
                }

        } /* else the irq handler advances the queue. */
        VDBG("list_add\n");
        if (req) {
        
                list_add_tail (&req->queue, &ep->queue);

                /* stop OUT naking */
                if (!ep->is_in) {
                        if (!use_dma && udc_rxfifo_pending) {
                                DBG("udc_queue(): pending bytes in rxfifo after nyet\n");
                                /* read pending bytes afer nyet: referring to isr */
                                if (read_fifo(ep, req)) {
                                        /* finish */
                                        done (ep, req, 0);
                                }
                                udc_rxfifo_pending = 0;

                        }
                }

        }


done:
        spin_unlock_irqrestore (&dev->lock, flags);
        return retval;
}

/**
 * Completes a request packet in DMA context 
 * 
 * \param ep            pointer to ep struct
 * \param req           pointer to request packet
 * \param status        request packet status
 */
static inline void
dma_done (
        struct udc_ep *ep,
        struct udc_request *req,
        int status
)
{
        rmb ();
        done (ep, req, status);
}

/**
 * Scan DMA completions 
 * 
 * \param ep            pointer to ep struct
 */
#ifdef NOT_USED
static void scan_dma_completions (struct udc_ep *ep)
{
        while (!list_empty (&ep->queue)) {
                struct udc_request      *req;

                req = list_entry (ep->queue.next,
                                struct udc_request, queue);

                if (!ep->is_in && (req->req.length % ep->ep.maxpacket) != 0) {
                        req->dma_done = 1;
                        break;
                }
                dma_done (ep, req, 0);
        }
}
#endif

/**
 * Abort current endpoint DMA 
 * 
 * \param ep            pointer to ep struct
 */
static inline void abort_dma (struct udc_ep *ep)
{
        /* abort the current transfer */

        /* collect completed transfers (except the current one), */
        /* but we have no real DMA chains, only for max packets */
        /* for PPB modes */
        //scan_dma_completions (ep);
}

/**
 * Destroy current endpoint DMA 
 * 
 * \param ep            pointer to ep struct
 */
static void nuke (struct udc_ep *ep)
{
        struct udc_request      *req;
        /* called with spinlock held */

        ep->stopped = 1;
        if (ep->dma)
                abort_dma (ep);

        while (!list_empty (&ep->queue)) {
                req = list_entry (ep->queue.next,
                                struct udc_request,
                                queue);
                done (ep, req, -ESHUTDOWN);
        }
}

/**
 * Dequeues a request packet, called by gadget driver
 * 
 * \param _ep           pointer to usb ep struct
 * \param _req          pointer to request packet to be freed
 * \return 0 if success
 */
static int udc_dequeue (struct usb_ep *_ep, struct usb_request *_req)
{
        struct udc_ep   *ep;
        struct udc_request      *req;
        unsigned long           flags;
        int                     stopped;

        ep = container_of (_ep, struct udc_ep, ep);
        req = container_of (_req, struct udc_request, req);
        if (!_ep || (!ep->desc && ep->num != 0) || !_req)
                return -EINVAL;

        spin_lock_irqsave (&ep->dev->lock, flags);
        stopped = ep->stopped;

        /* pause dma while we scan the queue */
        ep->stopped = 1;
        if (ep->dma) {
                /* disable dma */
                /* force synch, clean any completed requests */
                //spin_stop_dma (ep->dma);
                /* collect completed transfers (except the current one), */
                /* but we have no real DMA chains, only for max packets */
                /* for PPB modes */
                //scan_dma_completions (ep);
        }

        /* queue head may be partially complete. */
        if (ep->queue.next == &req->queue) {
                if (ep->dma) {
                        DBG( "unlink (%s) dma\n", _ep->name);
                        _req->status = -ECONNRESET;
                        abort_dma (ep);
                        if (likely (ep->queue.next == &req->queue))
                                dma_done (ep, req,
                                        -ECONNRESET);
                } else {
                        DBG( "unlink (%s) pio\n", _ep->name);
                        done (ep, req, -ECONNRESET);
                }
                req = 0;

        }

        if (req)
                done (ep, req, -ECONNRESET);
        ep->stopped = stopped;

        if (ep->dma) {
                /* turn off dma on inactive queues */
                if (list_empty (&ep->queue)) {
                        //stop_dma (ep->dma);
                }
                else if (!ep->stopped) {
                        /* resume current request, or start new one */
                }
        }

        spin_unlock_irqrestore (&ep->dev->lock, flags);
        return req ? 0 : -EOPNOTSUPP;
}

/**
 * Halt endpoint 
 * 
 * \param ep            pointer to ep struct
 */
static void set_halt (struct udc_ep *ep)
{
        u32 tmp;
        struct udc_request* req=NULL;
        //unsigned dma_pend = 0;
        ep->stopped = 1;
        /* check for pending DMA */
        if (!list_empty (&ep->queue)) {
                       req = list_entry (ep->queue.next,
                                        struct udc_request, queue);
                       if (req) {
                                /* TODO spin lock ? */
                                //if (req->dma_pend) {
                                        /* ep is stopped by ISR after completion of last req */
                                        //dma_pend = 1;
                                //}
                        }
        }
        if (!req) {
                /* set STALL */
                /* flush the txfifo */
                tmp = readl(&ep->regs->ctl);
                tmp |= AMD_BIT(UDC_EPCTL_F);
                writel(tmp, &ep->regs->ctl);

                tmp = readl(&ep->regs->ctl);
                tmp |= AMD_BIT(UDC_EPCTL_S);
                writel(tmp, &ep->regs->ctl);
                ep->stopped = 0;
        }

}

/**
 * clear halt of endpoint 
 * 
 * \param ep            pointer to ep struct
 * \return 0 if success 
 */
static int clear_halt (struct udc_ep *ep)
{
        u32 tmp;
        int retval = 0;

        /* ep can be stopped by set_halt() before */
        if (ep->stopped) {
                tmp = readl(&ep->regs->ctl);
                DBG("check STALL for ep%d\n",ep->num);
                /* STALL cleared ? */
                if (!(tmp & AMD_BIT(UDC_EPCTL_S))) {
                        DBG("clear NAK after STALL\n");
                        /* clear NAK by writing CNAK */
                        tmp = readl(&ep->regs->ctl);
                        tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
                        writel(tmp, &ep->regs->ctl);
                        ep->stopped = 0;
                        UDC_QUEUE_CNAK(ep, ep->num);
                }
                else {
                        /* Host did not sent CLEAR_FEATURE HALT yet */
                        retval = -EINVAL;
                }
        }

        return retval;
}

/**
 * Halt or clear halt of endpoint 
 * 
 * \param ep            pointer to ep struct
 * \param value         1 - halt, 0 - clear halt
 * \return 0 if success 
 */
static int
udc_set_halt (struct usb_ep *_ep, int value)
{
        struct udc_ep   *ep;
        int retval = 0;

        ep = container_of (_ep, struct udc_ep, ep);
        if (!_ep || (!ep->desc && ep->num != 0))
                return -EINVAL;
        if (!ep->dev->driver || ep->dev->gadget.speed == USB_SPEED_UNKNOWN)
                return -ESHUTDOWN;
        if (ep->desc /* not ep0 */ && 
            (ep->desc->bmAttributes & USB_ENDPOINT_XFERTYPE_MASK)
                        == USB_ENDPOINT_XFER_ISOC)
                return -EINVAL;

        DBG( "%s %s halt\n", _ep->name, value ? "set" : "clear");

        /* exit on suspend */
        if (ep->dev->sys_suspended) 
              return -ESHUTDOWN;

        /* set/clear, then synch memory views with the device */
        if (value) {
                if (ep->num == 0)
                        ep->dev->protocol_stall = 1;
                else
                        set_halt (ep);
        } else
                retval = clear_halt (ep);

        return retval;
}

/**
 * Return fifo fill state (not used, not implemented)
 * 
 * \param _ep            pointer to usb ep struct
 * \return available bytes in fifo 
 */
static int
udc_fifo_status (struct usb_ep *_ep)
{
        struct udc_ep   *ep;

        ep = container_of (_ep, struct udc_ep, ep);
        if (!_ep || (!ep->desc && ep->num != 0))
                return -ENODEV;
        if (!ep->dev->driver || ep->dev->gadget.speed == USB_SPEED_UNKNOWN)
                return -ESHUTDOWN;

        return 0;
}

/**
 * Flush the endpoint fifo (not implemented)
 * 
 * \param _ep            pointer to usb ep struct
 */
static void
udc_fifo_flush (struct usb_ep *_ep)
{
        struct udc_ep   *ep;

        ep = container_of (_ep, struct udc_ep, ep);
        if (!_ep || (!ep->desc && ep->num != 0))
                return;
        if (!ep->dev->driver || ep->dev->gadget.speed == USB_SPEED_UNKNOWN)
                return;
}

/* gadget interface */
static struct usb_ep_ops udc_ep_ops = {
        .enable         = udc_enable,
        .disable        = udc_disable,

        .alloc_request  = udc_alloc_request,
        .free_request   = udc_free_request,

        .alloc_buffer   = udc_alloc_buffer,
        .free_buffer    = udc_free_buffer,

        .queue          = udc_queue,
        .dequeue        = udc_dequeue,

        .set_halt       = udc_set_halt,
        .fifo_status    = udc_fifo_status,
        .fifo_flush     = udc_fifo_flush,
};

/*-------------------------------------------------------------------------*/

/**
 * Get frame count fifo (not implemented)
 * 
 * \param _gadget            pointer to gadget struct
 * \return frame count
 */
static int udc_get_frame (struct usb_gadget *_gadget)
{
        struct udc              *dev;
        unsigned long           flags;
        u16                     retval=0;

        if (!_gadget)
                return -ENODEV;
        dev = container_of (_gadget, struct udc, gadget);
        spin_lock_irqsave (&dev->lock, flags);
        // get frame
        spin_unlock_irqrestore (&dev->lock, flags);
        return retval;
}

/**
 * Remote wakeup gadget interface
 * 
 * \param _gadget            pointer to gadget struct
 * \return 0 if success
 */
static int udc_wakeup (struct usb_gadget *_gadget)
{
        struct udc              *dev;

        if (!_gadget)
                return 0;
        dev = container_of (_gadget, struct udc, gadget);
        udc_remote_wakeup(dev);

        return 0;
}

/**
 * gadget ioctl, used for OTG support notification
 * 
 * \param _gadget            pointer to gadget struct
 * \return 1 if OTG supported, else 0
 */
static int udc_gadget_ioctl (struct usb_gadget *_gadget, unsigned code, unsigned long param)
{
        struct udc              *dev;
        unsigned long           flags;
        u16                     retval=0;
        u32 tmp;

        if (!_gadget)
                return -ENODEV;
        dev = container_of (_gadget, struct udc, gadget);
        spin_lock_irqsave (&dev->lock, flags);
        tmp = readl(&dev->regs->cfg);
        
        if (tmp & AMD_BIT(UDC_DEVCFG_HNPSFEN))
                retval = 1;
        else
                retval = 0;
        spin_unlock_irqrestore (&dev->lock, flags);
        return retval;
}

/* gadget operations */
static const struct usb_gadget_ops udc_ops = {
        .get_frame      = udc_get_frame,
        .wakeup         = udc_wakeup,
        .ioctl          = udc_gadget_ioctl, 
        // .set_selfpowered = udc_set_selfpowered,
};

/*-------------------------------------------------------------------------*/

#undef USE_SYSFS_DEBUG_FILES
#ifdef  USE_SYSFS_DEBUG_FILES

/* "function" sysfs attribute */
/**
 *  Temporary left in the code for later reusing or removing
 *  dequeue ALL requests
 */
static ssize_t
show_function (struct device *_dev, char *buf)
{
        struct udc      *dev = dev_get_drvdata (_dev);

        if (!dev->driver
                        || !dev->driver->function
                        || strlen (dev->driver->function) > PAGE_SIZE)
                return 0;
        return snprintf (buf, PAGE_SIZE, "%s\n", dev->driver->function);
}
static DEVICE_ATTR (function, S_IRUGO, show_function, NULL);

/**
 *  Temporary left in the code for later reusing or removing
 *  dequeue ALL requests
 */
static ssize_t
show_registers (struct device *_dev, char *buf)
{
        return 0;
}
static DEVICE_ATTR (registers, S_IRUGO, show_registers, NULL);

/**
 *  Temporary left in the code for later reusing or removing
 *  dequeue ALL requests
 */
static ssize_t
show_queues (struct device *_dev, char *buf)
{
        struct udc              *dev;
        unsigned long           flags;

        dev = dev_get_drvdata (_dev);
        spin_lock_irqsave (&dev->lock, flags);
        spin_unlock_irqrestore (&dev->lock, flags);
        return 0;
}
static DEVICE_ATTR (queues, S_IRUGO, show_queues, NULL);


#else

#define device_create_file(a,b) do {} while (0)
#define device_remove_file      device_create_file

#endif

/*-------------------------------------------------------------------------*/

/**
 * Setups endpoint parameters, adds endpoints to linked list
 * 
 * \param dev           pointer to dev struct
 * \param mode          not used
 */
static void set_fifo_mode (struct udc *dev, int mode)
{
        /* make list */
        INIT_LIST_HEAD (&dev->gadget.ep_list);
        list_add_tail (&dev->ep [UDC_EPIN_STATUS_IX].ep.ep_list, &dev->gadget.ep_list);
        list_add_tail (&dev->ep [UDC_EPIN_IX].ep.ep_list, &dev->gadget.ep_list);
        list_add_tail (&dev->ep [UDC_EPOUT_IX].ep.ep_list, &dev->gadget.ep_list);
        
        /* setup fifo size */
        dev->ep [UDC_EPIN_STATUS_IX].fifo_size = UDC_EPIN_SMALLINT_BUFF_SIZE;
        if (dev->gadget.speed == USB_SPEED_FULL)
                dev->ep [UDC_EPIN_IX].fifo_size = UDC_FS_EPIN_BUFF_SIZE; 
        else if (dev->gadget.speed == USB_SPEED_HIGH)
                dev->ep [UDC_EPIN_IX].fifo_size = hs_tx_buf; 
        dev->ep [UDC_EPOUT_IX].fifo_size = UDC_RXFIFO_SIZE;
}

/**
 * Sets fifo mode, called by gadget driver
 * 
 * \param gadget        pointer to gadget struct
 * \param mode          not used
 * \return 0 if success 
 */
int amd5536udc_set_fifo_mode (struct usb_gadget *gadget, int mode)
{
        struct udc              *dev;
        unsigned long           flags;

        if (!gadget)
                return -ENODEV;
        dev = container_of (gadget, struct udc, gadget);

        spin_lock_irqsave (&dev->lock, flags);
        set_fifo_mode (dev, 0);
        spin_unlock_irqrestore (&dev->lock, flags);

        return 0;
}
EXPORT_SYMBOL (amd5536udc_set_fifo_mode);

/**
 * \brief
 * init registers at driver load time 
 *
 * \param dev        pointer to device struc
 * \return 0 if success
 */
static int startup_registers(struct udc* dev)
{
        u32 tmp;

        /* init controller by soft reset */
        udc_soft_reset(dev);

        /* mask not needed interrupts */
        udc_mask_unused_interrupts(dev);

        /* put into initial config */
        usb_reset (dev);
        /* link up all endpoints */
        usb_reinit (dev);
       
        /* program speed */
        tmp = readl(&dev->regs->cfg);
        if (use_fullspeed) {
                tmp = AMD_ADDBITS(tmp, UDC_DEVCFG_SPD_FS, UDC_DEVCFG_SPD);
        }
        else {
                tmp = AMD_ADDBITS(tmp, UDC_DEVCFG_SPD_HS, UDC_DEVCFG_SPD);
        }
        writel(tmp, &dev->regs->cfg);

        return 0;
}

/**
 * Resets usb context
 * 
 * \param dev           pointer to device struct
 */
static void usb_reset (struct udc *dev)
{
        u32     tmp;

        DBG("usb_reset()\n");        
        
        /* force immediate bus disconnect, and synch through pci */
        // disconnect
        dev->gadget.speed = USB_SPEED_UNKNOWN;

        /* disable DMA */
        tmp = readl(&dev->regs->ctl);
        tmp &= AMD_UNMASK_BIT(UDC_DEVCTL_RDE);
        tmp &= AMD_UNMASK_BIT(UDC_DEVCTL_TDE);
        writel(tmp, &dev->regs->ctl);                

        /* enable dynamic CSR programming */
        tmp =  readl(&dev->regs->cfg);
        tmp |= AMD_BIT(UDC_DEVCFG_CSR_PRG);
        /* set self powered */
        tmp |= AMD_BIT(UDC_DEVCFG_SP);
        /* set remote wakeupable */
        tmp |= AMD_BIT(UDC_DEVCFG_RWKP);
        writel(tmp, &dev->regs->cfg);                     

        /* clear old dma and irq state */
        // clear old dma

        /* standard fifo and endpoint allocations */
        set_fifo_mode (dev, 0);
}

/**
 * Inits usb context, sets initial endpoint parameters
 * 
 * \param dev           pointer to device struct
 */
static void usb_reinit (struct udc *dev)
{
        u32     tmp;
        u32     reg;

        DBG("usb_reinit()\n");

        /* read enum speed */
        tmp = readl(&dev->regs->sts);
        tmp = AMD_GETBITS(tmp, UDC_DEVSTS_ENUM_SPEED);        
        if (tmp ==  UDC_DEVSTS_ENUM_SPEED_HIGH) {
                dev->gadget.speed = USB_SPEED_HIGH;
        }
        else if (tmp ==  UDC_DEVSTS_ENUM_SPEED_FULL) {
                dev->gadget.speed = USB_SPEED_FULL;
        }
        
        /* basic endpoint init */
        for (tmp = 0; tmp < UDC_EP_NUM; tmp++) {
                struct udc_ep   *ep = &dev->ep [tmp];

                ep->ep.name = ep_name [tmp];
                ep->dev = dev;
                ep->num = tmp;
                /* txfifo size is calculated at enable time */
                ep->txfifo = dev->txfifo;

                /* fifo size */
                if (tmp < UDC_EPIN_NUM) {
                        ep->fifo_size = UDC_TXFIFO_SIZE;
                        ep->is_in = 1;
                }
                else {
                        ep->fifo_size = UDC_RXFIFO_SIZE;
                        ep->is_in = 0;

                }

                ep->regs = &dev->epregs [tmp];
                /* TODO ep will be reset only if ep was not enabled before to avoid
                        disabling ep interrupts when ENUM interrupt occurs but ep is
                        not enabled by gadget driver 
                        => to be improved for general ep_reset() */
                if (!ep->desc) {
                        ep_reset (dev->regs, ep);
                }

                if (use_dma) {
                        /* TODO ep->dma is not really used, just to indicate that */
                        /* DMA is active: remove this */
                        /* dma regs = dev control regs */        
                        ep->dma = (u32*) &dev->regs->ctl;       

                        /* nak OUT endpoints until enable - not for ep0*/
                        if (tmp != UDC_EP0IN_IX
                            && tmp != UDC_EP0OUT_IX
                            && tmp > UDC_EPIN_NUM) {
                                /* set NAK  */
                                reg = readl(&dev->ep[tmp].regs->ctl);
                                reg |= AMD_BIT(UDC_EPCTL_SNAK); 
                                writel(reg, &dev->ep[tmp].regs->ctl);
                                dev->ep[tmp].naking = 1;
                                       
                        }
                }
        }
        /* EP0 max packet */
        if (dev->gadget.speed == USB_SPEED_FULL) {
                dev->ep [UDC_EP0IN_IX].ep.maxpacket = UDC_FS_EP0IN_MAX_PKT_SIZE;
                dev->ep [UDC_EP0OUT_IX].ep.maxpacket = UDC_FS_EP0OUT_MAX_PKT_SIZE;
        }
        else if (dev->gadget.speed == USB_SPEED_HIGH) {
                dev->ep [UDC_EP0IN_IX].ep.maxpacket = UDC_EP0IN_MAX_PKT_SIZE;
                dev->ep [UDC_EP0OUT_IX].ep.maxpacket = UDC_EP0OUT_MAX_PKT_SIZE;
        }

        /* with suspend bug workaround, ep0 params for gadget driver
        are set at gadget driver bind() call */
        dev->gadget.ep0 = &dev->ep [UDC_EP0IN_IX].ep;
        dev->ep [UDC_EP0IN_IX].stopped = 0;
        INIT_LIST_HEAD (&dev->gadget.ep0->ep_list);

        /* init cfg/alt/int */
        dev->cur_config = 0;
        dev->cur_intf = 0;
        dev->cur_alt = 0;
}

/**
 * Bringup after Connect event,
 * initial bringup to be ready for ep0 events 
 * 
 * \param dev           pointer to device struct
 */
static void usb_connect (struct udc *dev) {

                INFO("USB Connect\n");

                dev->connected = 1;

                /* put into initial config */
                usb_reset (dev);

                /* enable device setup interrupts */
                udc_enable_dev_setup_interrupts(dev);
}

/**
 * Calls gadget with disconnect event and resets the UDC and makes
 * initial bringup to be ready for ep0 events 
 * 
 * \param dev           pointer to device struct
 */
static void usb_disconnect (struct udc *dev) {

                INFO("USB Disconnect\n");

                dev->connected = 0;

                /* mask interrupts */
                udc_mask_unused_interrupts(dev);

                tasklet_schedule(&disconnect_tasklet);
}

/**
 * Tasklet for disconnect to be outside of interrupt 
 * context 
 * 
 * \param par   pointer to device struct pointer
 */
void udc_tasklet_disconnect(unsigned long par)
{
        struct udc* dev = (struct udc*) (*((struct udc**) par));
        u32 tmp;

        DBG("Tasklet disconnect\n");

        /* call gadget to reset configs etc. */
        if (spin_is_locked(&dev->lock)) {
                spin_unlock(&dev->lock);
                dev->driver->disconnect (&dev->gadget);
                spin_lock(&dev->lock);
        }
        else
                dev->driver->disconnect (&dev->gadget);

        /* empty queues */
        for (tmp = 0; tmp < UDC_EP_NUM; tmp++) {
                nuke (&dev->ep [tmp]);
        }
        /* disable ep0 */
        ep_reset (dev->regs,
                        &dev->ep [UDC_EP0IN_IX]);
        

        if (!soft_reset_occured) {
                /* init controller by soft reset */
#ifdef UDC_IPCASE_8000018724_WORKAROUND
                udc_soft_reset(dev);
#endif
                soft_reset_occured++;

#ifndef UDC_IPBUG_3950_WORKAROUND
                /* re-enable dev interrupts */
                udc_enable_dev_setup_interrupts(dev);
#endif
                /* back to full speed ? */
                if (use_fullspeed) {
                        tmp = readl(&dev->regs->cfg);
                        tmp = AMD_ADDBITS(tmp, UDC_DEVCFG_SPD_FS, UDC_DEVCFG_SPD);
                        writel(tmp, &dev->regs->cfg);
                }

                /* TODO for AU1200: disable UDC memory */
        }
}

/**
 * Reset the UDC core 
 * 
 * \param dev           pointer to device struct
 */
static void udc_soft_reset(struct udc* dev)
{
        DBG("Soft reset\n");
        /* reset possible waiting interrupts, because int.
           status is lost after soft reset */
        /* ep int. status reset */ 
        writel(UDC_EPINT_MSK_DISABLE_ALL, &dev->regs->ep_irqsts);
        /* device int. status reset */ 
        writel(UDC_DEV_MSK_DISABLE, &dev->regs->irqsts);

        /* FIXME: reset is the only way to restart DMA now,
           later if the hw bug is fixed, use a softer way */
        spin_lock_irq(&udc_irq_spinlock);
        writel(AMD_BIT(UDC_DEVCFG_SOFTRESET), &dev->regs->cfg);
        readl(&dev->regs->cfg);
        spin_unlock_irq(&udc_irq_spinlock);
        
}

#if defined(UDC_IPBUG_3950_WORKAROUND) \
    || defined(UDC_IPCASE_8000018724_WORKAROUND)
/**
 * Called by OTG driver to notify us regarding an OTG event
 * 
 * \param code           notify code
 */
void otg_notify(unsigned int code) {
        VDBG("OTG notify code=%d\n", code);
        switch (code) {
                case OTG_GADGET_EVT_SVDROP:
                        /* disconnect event */
                        usb_disconnect(the_controller);
                        break;
                case OTG_GADGET_EVT_SVALID:
                        /* connect event */
                        usb_connect(the_controller);
                        break;
                case OTG_GADGET_REQ_WAKE:
                        /* remote wakeup event */
                        udc_remote_wakeup(the_controller);
                        break;
        }  
}
#endif

/**
 * Inits endpoint 0 so that SETUP packets are processed
 * 
 * \param dev           pointer to device struct
 */
static void ep0_start (struct udc *dev)
{
        u32 tmp;

        DBG("ep0_start\n");

        /* flush fifo */
        tmp = readl(&dev->ep[UDC_EP0IN_IX].regs->ctl);
        tmp |= AMD_BIT(UDC_EPCTL_F); 
        writel(tmp, &dev->ep[UDC_EP0IN_IX].regs->ctl);

        /* set ep0 directions */
        dev->ep[UDC_EP0IN_IX].is_in = 1;
        dev->ep[UDC_EP0OUT_IX].is_in = 0;

        /* set buffer size (tx fifo entries) of EP0_IN */
        tmp = readl(&dev->ep[UDC_EP0IN_IX].regs->bufin_framenum);
        if (dev->gadget.speed == USB_SPEED_FULL)
                tmp = AMD_ADDBITS(tmp, UDC_FS_EPIN0_BUFF_SIZE, UDC_EPIN_BUFF_SIZE);
        else if (dev->gadget.speed == USB_SPEED_HIGH)
                tmp = AMD_ADDBITS(tmp, UDC_EPIN0_BUFF_SIZE, UDC_EPIN_BUFF_SIZE);
        writel(tmp, &dev->ep[UDC_EP0IN_IX].regs->bufin_framenum);
        
        /* set max packet size of EP0_IN */
        tmp = readl(&dev->ep[UDC_EP0IN_IX].regs->bufout_maxpkt);
        if (dev->gadget.speed == USB_SPEED_FULL)
                tmp = AMD_ADDBITS(tmp, UDC_FS_EP0IN_MAX_PKT_SIZE, UDC_EP_MAX_PKT_SIZE);
        else if (dev->gadget.speed == USB_SPEED_HIGH)
                tmp = AMD_ADDBITS(tmp, UDC_EP0IN_MAX_PKT_SIZE, UDC_EP_MAX_PKT_SIZE);
        writel(tmp, &dev->ep[UDC_EP0IN_IX].regs->bufout_maxpkt);

        /* set max packet size of EP0_OUT */
        tmp = readl(&dev->ep[UDC_EP0OUT_IX].regs->bufout_maxpkt);
        if (dev->gadget.speed == USB_SPEED_FULL)
                tmp = AMD_ADDBITS(tmp, UDC_FS_EP0OUT_MAX_PKT_SIZE, UDC_EP_MAX_PKT_SIZE);
        else if (dev->gadget.speed == USB_SPEED_HIGH)
                tmp = AMD_ADDBITS(tmp, UDC_EP0OUT_MAX_PKT_SIZE, UDC_EP_MAX_PKT_SIZE);
        writel(tmp, &dev->ep[UDC_EP0OUT_IX].regs->bufout_maxpkt);

        /* set max packet size of EP0 in UDC CSR  */
        tmp = readl(&dev->csr->ne[0]);
        if (dev->gadget.speed == USB_SPEED_FULL)
                tmp = AMD_ADDBITS(tmp, UDC_FS_EP0OUT_MAX_PKT_SIZE, UDC_CSR_NE_MAX_PKT);
        else if (dev->gadget.speed == USB_SPEED_HIGH)
                tmp = AMD_ADDBITS(tmp, UDC_EP0OUT_MAX_PKT_SIZE, UDC_CSR_NE_MAX_PKT);
        writel(tmp, &dev->csr->ne[0]);

        if (use_dma) {
                dev->ep [UDC_EP0OUT_IX].td->dmastat |= AMD_BIT(UDC_DMA_OUT_STS_L);
                /* write dma desc address */
                writel(dev->ep [UDC_EP0OUT_IX].td_stp_dma, &dev->ep[UDC_EP0OUT_IX].regs->subptr);
                writel(dev->ep [UDC_EP0OUT_IX].td_dma, &dev->ep[UDC_EP0OUT_IX].regs->desptr);
                /* enable DMA */
                tmp = readl(&dev->regs->ctl);
                tmp |= AMD_BIT(UDC_DEVCTL_MODE)
                                | AMD_BIT(UDC_DEVCTL_RDE)
                                | AMD_BIT(UDC_DEVCTL_TDE);
                if (use_dma_bufferfill_mode) {
                        tmp |= AMD_BIT(UDC_DEVCTL_BF);
                } 
                else if (use_dma_ppb_du) {
                        tmp |= AMD_BIT(UDC_DEVCTL_DU);
                }
                writel(tmp, &dev->regs->ctl);                
        }

        /* clear NAK by writing CNAK for EP0IN */
        tmp = readl(&dev->ep[UDC_EP0IN_IX].regs->ctl);
        tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
        writel(tmp, &dev->ep[UDC_EP0IN_IX].regs->ctl);
        dev->ep[UDC_EP0IN_IX].naking = 0;
        UDC_QUEUE_CNAK(&dev->ep[UDC_EP0IN_IX], UDC_EP0IN_IX);
        
        /* clear NAK by writing CNAK for EP0OUT */
        tmp = readl(&dev->ep[UDC_EP0OUT_IX].regs->ctl);
        tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
        writel(tmp, &dev->ep[UDC_EP0OUT_IX].regs->ctl);
        dev->ep[UDC_EP0OUT_IX].naking = 0;
        UDC_QUEUE_CNAK(&dev->ep[UDC_EP0OUT_IX], UDC_EP0OUT_IX);
}

/**
 * \brief
 * Make endpoint 0 ready for control traffic
 *
 * \param dev        pointer to device struc
 * \return 0 if success
 */
static int enable_ep0(struct udc *dev)
{
        /* ... then enable host detection and ep0; and we're ready
         * for set_configuration as well as eventual disconnect.
         */
        ep0_start (dev);
        /* enable ep0 interrupts */
        udc_enable_ep0_interrupts(dev);
        /* enable device setup interrupts */
        udc_enable_dev_setup_interrupts(dev);

        return 0;
}

/**
 * \brief
 * Called by gadget driver to register itself
 * 
 * When a driver is successfully registered, it will receive
 * control requests including set_configuration(), which enables
 * non-control requests.  then usb traffic follows until a
 * disconnect is reported.  then a host may connect again, or
 * the driver might get unbound.
 *
 * \param driver        pointer to gadget driver struct
 * \return 0 if success
 */
int usb_gadget_register_driver (struct usb_gadget_driver *driver)
{
        struct udc              *dev = the_controller;
        int                     retval;

        /* insist on high speed support from the driver */
        if (!driver
                        || driver->speed != USB_SPEED_HIGH
                        || !driver->bind
                        || !driver->unbind
                        || !driver->setup)
                return -EINVAL;
        if (!dev)
                return -ENODEV;
        if (dev->driver)
                return -EBUSY;

        /* hook up the driver ... */
#if     LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
        driver->driver.bus = 0;
#endif
        dev->driver = driver;
#if     LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
        dev->gadget.dev.driver = &driver->driver;
#endif

#if     LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
#ifndef CONFIG_MIPS
        /* FIXME TMP26: temporary disabled device file system usage until
         * working for Au1200 */
        device_create_file (&dev->pdev->dev, &dev_attr_function);
        device_create_file (&dev->pdev->dev, &dev_attr_queues);
#endif
#endif

#ifdef CONFIG_USB_OTG
	dev->gadget.is_otg = 1;
#endif
        retval = driver->bind (&dev->gadget);
        gadget_bind_count++;
        if (retval) {
                DBG( "bind to driver %s --> %d\n",
                                driver->driver.name, retval);
                dev->driver = 0;
#if     LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
                dev->gadget.dev.driver = 0;
#endif
                return retval;
        }

        /* if otg driver already registered */
        /* call otg bind() to mux udc to phy */
        if (dev->otg_transceiver) {
                dev->otg_transceiver->set_peripheral(
                        dev->otg_transceiver, &dev->gadget);
        }

        /* pci writes may still be posted */
        return 0;
}
EXPORT_SYMBOL (usb_gadget_register_driver);

/**
 * Called by OTG driver to register itself
 * 
 *
 * \param get_transceiver  function pointer to get OTG info
 * \return 0 if success
 */
int usb_gadget_register_otg (struct otg_transceiver * (*get_transceiver)(void))
{
        struct udc              *dev = the_controller;
        int                     retval;
        
        if (!get_transceiver)
                return -EINVAL;
        if (!dev)
                return -ENODEV;
        if (dev->otg_transceiver)
                return -EBUSY;

        /* hook up the driver ... */
        dev->otg_transceiver = get_transceiver ();
        if (!dev->otg_transceiver->otg_priv)
                return -EINVAL;
        dev->otg_driver = (struct usb_otg_gadget_extension *)
                                dev->otg_transceiver->otg_priv;

#ifdef UDC_IPBUG_3943_WORKAROUND
        /* init registers here first with suspend bug */
        if (!otg_reg_count) {
                startup_registers(dev);
                otg_reg_count++; 
        }
#endif

#ifdef UDC_IPBUG_3950_WORKAROUND
        /* set notify function */
        dev->otg_driver->notify = otg_notify;
#endif
        /* if gadget driver already registered */
        /* call gadget bind() to switch to mux udc to phy */
        /* TODO use SD bit to disconnect/connect later, in */
        /* AMD5536 RevA0 this does not work ! */
        if (dev->driver) {

               /* otg driver bind() */
                retval = dev->otg_transceiver->set_peripheral(
                                dev->otg_transceiver, &dev->gadget);
               if (retval) {
                       DBG( "error bind to otg driver\n");
                        dev->otg_driver = NULL;
                        dev->otg_transceiver = NULL;
                       return retval;
               }
                /* get ready for ep0 traffic */
                enable_ep0(dev);
        }

        INFO( "registered otg driver\n");

        return 0;
}
EXPORT_SYMBOL (usb_gadget_register_otg);

/**
 * Called by OTG driver to unregister itself
 * 
 *
 * \return 0 if success
 */
int usb_gadget_unregister_otg (void)
{
        struct udc      *dev = the_controller;
        unsigned long   flags;

        if (!dev)
                return -ENODEV;

        spin_lock_irqsave (&dev->lock, flags);
        /* add critical sections here */

        /* mask not needed interrupts */
        udc_mask_unused_interrupts(dev);

        spin_unlock_irqrestore (&dev->lock, flags);

        dev->otg_supported = 0;
        if (dev->otg_transceiver) {
                dev->otg_transceiver->set_peripheral(dev->otg_transceiver, NULL);
                dev->otg_transceiver = NULL;
        }
        if (dev->otg_driver) {
#ifdef UDC_IPBUG_3950_WORKAROUND
                dev->otg_driver->notify = NULL;
#endif
        dev->otg_driver = NULL;
        }

        DBG( "unregistered otg driver\n");
        return 0;
}
EXPORT_SYMBOL (usb_gadget_unregister_otg);

/**
 *  Temporary left in the code for later reusing or removing
 *  dequeue ALL requests
 */
static void
stop_activity (struct udc *dev, struct usb_gadget_driver *driver)
{
        int tmp;

        /* don't disconnect if it's not connected */
        if (dev->gadget.speed == USB_SPEED_UNKNOWN)
                driver = 0;

        /* stop hardware; prevent new request submissions;
         * and kill any outstanding requests.
         */
        usb_reset (dev);
        for (tmp = 0; tmp < UDC_EP_NUM; tmp++) {
                nuke (&dev->ep [tmp]);
        }

        /* report disconnect; the driver is already quiesced */
        if (driver) {
                spin_unlock (&dev->lock);
                driver->disconnect (&dev->gadget);
                spin_lock (&dev->lock);
        }

        usb_reinit (dev);
}

/**
 * Called by gadget driver to unregister itself
 *
 * \param driver        pointer to gadget driver struct
 * \return 0 if success
 */
int usb_gadget_unregister_driver (struct usb_gadget_driver *driver)
{
        struct udc      *dev = the_controller;
        unsigned long   flags;

        if (!dev)
                return -ENODEV;
        if (!driver || driver != dev->driver)
                return -EINVAL;
        if (gadget_bind_count) {
                spin_lock_irqsave (&dev->lock, flags);
                stop_activity (dev, driver);
                spin_unlock_irqrestore (&dev->lock, flags);
        }

        /* unbind from otg driver first */
        if (dev->otg_transceiver) {
                dev->otg_transceiver->set_peripheral(
                        dev->otg_transceiver, NULL);
        }

        if (gadget_bind_count) {
                driver->unbind (&dev->gadget);
        }
        gadget_bind_count = 0;
        dev->driver = 0;

        DBG( "unregistered driver '%s'\n", driver->driver.name);
        return 0;
}
EXPORT_SYMBOL (usb_gadget_unregister_driver);


#ifdef UDC_IPBUG_3958_WORKAROUND
/**
 * Clear pending NAK bits  
 *
 * \param dev           pointer to UDC device object    
 * \return 0 if success
 */
static void udc_process_cnak_queue(struct udc* dev)
{
        u32 tmp;
        u32 reg;
        /* check epin's */
        DBG("CNAK pending queue processing\n");
        for (tmp = 0; tmp < UDC_EPIN_NUM_USED; tmp++) {
                if (cnak_pending & (1 << tmp)) {
                        DBG("CNAK pending for ep%d\n", tmp);
                        /* clear NAK by writing CNAK */
                        reg = readl(&dev->ep[tmp].regs->ctl);
                        reg |= AMD_BIT(UDC_EPCTL_CNAK); 
                        writel(reg, &dev->ep[tmp].regs->ctl);
                        dev->ep[tmp].naking = 0;
                        UDC_QUEUE_CNAK(&dev->ep[tmp], dev->ep[tmp].num);
                } 
        }
        /* ...  and ep0out */
        if (cnak_pending & (1 << UDC_EP0OUT_IX)) {
                DBG("CNAK pending for ep%d\n", UDC_EP0OUT_IX);
                /* clear NAK by writing CNAK */
                reg = readl(&dev->ep[UDC_EP0OUT_IX].regs->ctl);
                reg |= AMD_BIT(UDC_EPCTL_CNAK); 
                writel(reg, &dev->ep[UDC_EP0OUT_IX].regs->ctl);
                dev->ep[UDC_EP0OUT_IX].naking = 0;
                UDC_QUEUE_CNAK(&dev->ep[UDC_EP0OUT_IX], 
                                dev->ep[UDC_EP0OUT_IX].num);
        }
}
#endif

/**
 * Interrupt handler for Control OUT traffic  
 *
 * \param dev           pointer to UDC device object    
 * \return 0 if success
 */
static inline int udc_control_out_isr(struct udc* dev)
{
        int ret_val = 0;
        u32 tmp;
        
        /* clear irq */
        writel(AMD_BIT(UDC_EPINT_OUT_EP0), &dev->regs->ep_irqsts);

        tmp = readl(&dev->ep[UDC_EP0OUT_IX].regs->sts);
        /* check BNA and clear if set */
        if (tmp & AMD_BIT(UDC_EPSTS_BNA)) {
                VDBG("BNA set\n");       
                writel(AMD_BIT(UDC_EPSTS_BNA), &dev->ep[UDC_EP0OUT_IX].regs->sts);
        }
        
        /* type of data: SETUP or DATA 0 bytes */
        tmp = AMD_GETBITS(tmp, UDC_EPSTS_OUT);
        VDBG( "data_typ = %lx\n", (unsigned long) tmp);
        /* setup data */
        if (tmp == UDC_EPSTS_OUT_SETUP) {

                /* set NAK for EP0_IN */
                tmp = readl(&dev->ep[UDC_EP0IN_IX].regs->ctl);
                tmp |= AMD_BIT(UDC_EPCTL_SNAK); 
                writel(tmp, &dev->ep[UDC_EP0IN_IX].regs->ctl);
                dev->ep[UDC_EP0IN_IX].naking = 1;
                /* get setup data */
                if (use_dma) {

                        /* clear OUT bits in ep status */
                        writel(UDC_EPSTS_OUT_SETUP_CLEAR, &dev->ep[UDC_EP0OUT_IX].regs->sts);

                        setup_data.raw[0] = dev->ep[UDC_EP0OUT_IX].td_stp->data12;
                        setup_data.raw[1] = dev->ep[UDC_EP0OUT_IX].td_stp->data34;

                        /* set HOST READY */
                        writel(UDC_DMA_STP_STS_BS_HOST_READY,
                               &dev->ep[UDC_EP0OUT_IX].td_stp->dmastat);
                        /* set RDE */
                        tmp = readl(&dev->regs->ctl);
                        tmp |= AMD_BIT(UDC_DEVCTL_RDE);
                        writel(tmp, &dev->regs->ctl);                
                        tmp = readl(&dev->regs->ctl);
                }
                else {
                        /* read fifo */
                        udc_read_fifo_dwords(dev, setup_data.raw, 2);
                }

                /* call gadget with setup data received */
                spin_unlock (&dev->lock);
                tmp = dev->driver->setup (&dev->gadget, &setup_data.r);
                spin_lock (&dev->lock);

                /* clear NAK by writing CNAK in EP0_IN */
                tmp = readl(&dev->ep[UDC_EP0IN_IX].regs->ctl);
                tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
                writel(tmp, &dev->ep[UDC_EP0IN_IX].regs->ctl);
                dev->ep[UDC_EP0IN_IX].naking = 0;
                UDC_QUEUE_CNAK(&dev->ep[UDC_EP0IN_IX], UDC_EP0IN_IX);
                

                /* FIXME tdahlman - which cnak has to be set ep0 in or out ? */
                /* clear NAK by writing CNAK in EP0_OUT */
                tmp = readl(&dev->ep[UDC_EP0OUT_IX].regs->ctl);
                tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
                writel(tmp, &dev->ep[UDC_EP0OUT_IX].regs->ctl);
                dev->ep[UDC_EP0OUT_IX].naking = 0;
                UDC_QUEUE_CNAK(&dev->ep[UDC_EP0OUT_IX], UDC_EP0OUT_IX);
               
                if (!use_dma) { 
                        /* clear OUT bits in ep status */
                        writel(UDC_EPSTS_OUT_SETUP_CLEAR, &dev->ep[UDC_EP0OUT_IX].regs->sts);
                }

        } /* data packet 0 bytes */
        else if (tmp == UDC_EPSTS_OUT_DATA) {
                /* clear OUT bits in ep status */
                writel(UDC_EPSTS_OUT_DATA_CLEAR, &dev->ep[UDC_EP0OUT_IX].regs->sts);
                
                /* get setup data: only 0 packet */
                if (use_dma) {

                        /* set HOST READY */
                        dev->ep[UDC_EP0OUT_IX].td->dmastat =
                                AMD_ADDBITS(dev->ep[UDC_EP0OUT_IX].td->dmastat,
                                            UDC_DMA_OUT_STS_BS_HOST_READY,
                                            UDC_DMA_OUT_STS_BS);
                        /* set RDE */
                        tmp = readl(&dev->regs->ctl);
                        tmp |= AMD_BIT(UDC_DEVCTL_RDE);
                        writel(tmp, &dev->regs->ctl);                
                        tmp = readl(&dev->regs->ctl);
                }
                else {
                        /* dummy read confirm */
                        readl(&dev->ep[UDC_EP0OUT_IX].regs->confirm);
                }
        }

#ifdef UDC_IPBUG_3958_WORKAROUND
        /* check pending CNAKS */
        if (cnak_pending) {
                /* CNAk processing when rxfifo empty only */
                if (readl(&dev->regs->sts) & AMD_BIT(UDC_DEVSTS_RXFIFO_EMPTY)) {
                        udc_process_cnak_queue(dev);
                }
        }
#endif       
 
        return ret_val;                       
}

/**
 * Interrupt handler for Control IN traffic  
 *
 * \param dev           pointer to UDC device object    
 * \return 0 if success
 */
static inline int udc_control_in_isr(struct udc* dev)
{
        int ret_val = 0;
        u32 tmp;
        struct udc_ep   *ep;
        struct udc_request              *req;
        unsigned        len;

        ep = &dev->ep [0];

        /* clear irq */
        writel(AMD_BIT(UDC_EPINT_IN_EP0), &dev->regs->ep_irqsts);
        
        tmp= readl(&dev->ep[UDC_EP0IN_IX].regs->sts);
        /* DMA completion */
        if (tmp & AMD_BIT(UDC_EPSTS_TDC)) {
                VDBG("isr: TDC clear \n");
#ifdef UDC_IPBUG_2253_WORKAROUND
                /* stop NAKing after small packet DMA */
                if (ep->naking) {
                        /* clear NAK by writing CNAK */
                        tmp = readl(&ep->regs->ctl);
                        tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
                        writel(tmp, &ep->regs->ctl);
                        ep->naking = 0;
                        UDC_QUEUE_CNAK(ep, ep->num);
                }
#endif
                /* clear TDC bit */
                writel(AMD_BIT(UDC_EPSTS_TDC),&dev->ep[UDC_EP0IN_IX].regs->sts);
        } /* status reg has IN bit set ? */
        else if (tmp & AMD_BIT(UDC_EPSTS_IN)) {
                if (ep->dma) {
                        /* clear IN bit */
                        writel(AMD_BIT(UDC_EPSTS_IN),&dev->ep[UDC_EP0IN_IX].regs->sts);
                }

                if (!list_empty (&ep->queue))
                {
                        /* next request */
                        req = list_entry (ep->queue.next,
                                        struct udc_request, queue);
                        
                        if (ep->dma) {
                                /* write desc pointer */
                                writel(req->td_dma, &ep->regs->desptr);
                                /* set HOST READY */
                                req->td_data->dmastat 
                                      = AMD_ADDBITS(req->td_data->dmastat,
                                                    UDC_DMA_STP_STS_BS_HOST_READY,
                                                    UDC_DMA_STP_STS_BS);
                                wb_flush();

#ifdef UDC_IPBUG_2253_WORKAROUND
                                /* NAK if small packet until TDC interrupt */
                                if (req->req.length < UDC_SMALL_PACKET) {
                                        /* set NAK */
                                        tmp = readl(&ep->regs->ctl);
                                        tmp |= AMD_BIT(UDC_EPCTL_SNAK); 
                                        writel(tmp, &ep->regs->ctl);
                                        ep->naking = 1;
                                        wb_flush();
                                }
#endif
                                /* set poll demand bit */
                                tmp = readl(&dev->ep[UDC_EP0IN_IX].regs->ctl);
                                tmp |= AMD_BIT(UDC_EPCTL_P);
                                writel(tmp, &dev->ep[UDC_EP0IN_IX].regs->ctl);

                                /* all bytes will be transferred */
                                req->req.actual = req->req.length;

                                /* complete req */
                                done (ep, req, 0);
                                
                        }
                        else {
                                /* write fifo */
                                write_fifo(ep, &(req->req));

                                /* lengh bytes transfered */
                                len = req->req.length - req->req.actual;
				if (len > ep->ep.maxpacket)
					len = ep->ep.maxpacket;

                                req->req.actual += len;
                                if (req->req.actual == req->req.length
                                    || (len != ep->ep.maxpacket)) {
                                    //&& (!req->req.zero || len != ep->ep.maxpacket)) {
                                        /* complete req */
                                        done (ep, req, 0);
                                }
                        }
                        
                }
                
                ep->stopped = 0;
                dev->protocol_stall = 0;
                if (!ep->dma) {
                        /* clear IN bit */
                        writel(AMD_BIT(UDC_EPSTS_IN),&dev->ep[UDC_EP0IN_IX].regs->sts);
                }
        }

        return ret_val;
}

/**
 * Interrupt handler for data OUT traffic  
 *
 * \param dev           pointer to UDC device object    
 * \param ep_ix         endpoint index    
 * \return 0 if success
 */
static inline int udc_data_out_isr(struct udc* dev, int ep_ix)
{
        int ret_val = 0;
        u32 tmp;
        struct udc_ep   *ep;
        struct udc_request              *req;
        unsigned int count;
        struct udc_data_dma     *td = NULL;
        unsigned dma_done;

        VDBG("ep%d irq\n", ep_ix); 
        ep = &dev->ep [ep_ix];

        tmp = readl(&ep->regs->sts);
        if (use_dma) {
                /* BNA event ? */
                if (tmp & AMD_BIT(UDC_EPSTS_BNA)) {
                        ERR("BNA ep%dout occured - DESPTR = %lx \n", ep->num, (unsigned long) readl(&ep->regs->desptr));
                        
                        /* clear BNA */
                        writel(tmp | AMD_BIT(UDC_EPSTS_BNA), &ep->regs->sts); 
                        /* TODO no early return */
                        return 1;                       
                } 
        }
        /* HE event ? */
        if (tmp & AMD_BIT(UDC_EPSTS_HE)) {
                ERR("HE ep%dout occured\n", ep->num);
                
                /* clear HE */
                writel(tmp | AMD_BIT(UDC_EPSTS_HE), &ep->regs->sts); 
                /* TODO no early return */
                return 1;                       
        } 

        if (!list_empty (&ep->queue)) {

                /* next request */ 
                req = list_entry (ep->queue.next,
                        struct udc_request, queue);
        }
        else
        {
                req = 0;
#ifdef DEBUG
                no_req++;
#endif
                udc_rxfifo_pending = 1;
        }
        VDBG("req = %lx\n", (unsigned long) req);
        /* fifo mode ****************/
        if (!use_dma) {

                /* read fifo */
                if (req && read_fifo(ep, req)) {
                        /* finish */
                        done (ep, req, 0);
                        /* maybe advance queue to next request */
                        if (!list_empty (&ep->queue) && !ep->stopped) {
                                req = list_entry (ep->queue.next,
                                        struct udc_request, queue);
                        } 
                        else
                                req = 0;
                }

        } /* DMA ********************/
        else if (req) {

                /* check for DMA done */
                if (!use_dma_ppb) {
                        dma_done = AMD_GETBITS(req->td_data->dmastat, UDC_DMA_OUT_STS_BS); 
                } /* packet per buffer mode - rx bytes */
                else {
                        td = udc_get_last_dma_desc(req);
                        dma_done = AMD_GETBITS(td->dmastat, UDC_DMA_OUT_STS_BS); 
                }
                if (dma_done == UDC_DMA_OUT_STS_BS_DMA_DONE) {
                        /* buffer fill mode - rx bytes */
                        if (!use_dma_ppb) {
                                /* received number bytes */
                                count = AMD_GETBITS(req->td_data->dmastat, UDC_DMA_OUT_STS_RXBYTES); 
                                VDBG("rx bytes=%lx\n", (unsigned long) count);
                        } /* packet per buffer mode - rx bytes */
                        else {
                                VDBG("req->td_data=%lx\n", (unsigned long) req->td_data);
                                VDBG("last desc = %lx\n", (unsigned long) td);
                                /* received number bytes */
                                if (use_dma_ppb_du) {
                                        /* every desc. counts bytes */
                                        count = udc_get_ppbdu_rxbytes(req);
                                }
                                else {
                                        /* last desc. counts bytes */
                                        count = AMD_GETBITS(td->dmastat, UDC_DMA_OUT_STS_RXBYTES);
                                        if (!count) {
                                                /* on 64k packets the RXBYTES field is zero */
                                                if (req->req.length == UDC_DMA_MAXPACKET)
                                                        count = UDC_DMA_MAXPACKET; 
                                        }
                                }
                                VDBG("last desc rx bytes=%lx\n", (unsigned long) count);
                        }

                        tmp = req->req.length - req->req.actual;
                        if (count > tmp) {
                                /* as with DMA, data overflow gets flushed */
                                if ((tmp % ep->ep.maxpacket) != 0) {
                                        ERR( "%s out fifo %d bytes, expected %d\n",
                                                ep->ep.name, count, tmp);
                                        req->req.status = -EOVERFLOW;
                                }
                                count = tmp;
                        }
                        req->req.actual += count;

                        /* complete request */ 
                        done (ep, req, 0);
                        
                        /* maybe advance queue to next request */
                        if (!list_empty (&ep->queue) && !ep->stopped) {
                                req = list_entry (ep->queue.next,
                                        struct udc_request, queue);

                                /* next dma */
                                ret_val = prep_dma(ep, req);
                                if (ret_val != 0)
                                        goto done;
                                /* write desc pointer */
                                writel(req->td_dma, &ep->regs->desptr);
                                /* set RDE */
                                tmp = readl(&dev->regs->ctl);
                                tmp |= AMD_BIT(UDC_DEVCTL_RDE);
                                writel(tmp, &dev->regs->ctl);                
                                tmp = readl(&dev->regs->ctl);
                        }       
#ifdef DEBUG  
                        else {
                                no_pref_req++;
                                VDBG("OUT queue empty\n");
                        }
#endif

               } 
               else {
                        /* RX DMA must be reenabled for each desc in PPBDU mode */
                        if (use_dma_ppb_du) {
                                /* set RDE */ 
                                tmp = readl(&dev->regs->ctl);
                                tmp |= AMD_BIT(UDC_DEVCTL_RDE);
                                writel(tmp, &dev->regs->ctl);                
                                tmp = readl(&dev->regs->ctl);
                        }
               } 
                
        }

#ifdef UDC_IPBUG_3958_WORKAROUND
        /* check pending CNAKS */
        if (cnak_pending) {
                /* CNAk processing when rxfifo empty only */
                if (readl(&dev->regs->sts) & AMD_BIT(UDC_DEVSTS_RXFIFO_EMPTY)) {
                        udc_process_cnak_queue(dev);
                }
        }
#endif       

        /* clear OUT bits in ep status */
        writel(UDC_EPSTS_OUT_DATA_CLEAR, &ep->regs->sts);
done:
        return ret_val;
}


/**
 * Interrupt handler for data IN traffic  
 *
 * \param dev           pointer to UDC device object    
 * \param ep_ix         endpoint index    
 * \return 0 if success
 */
static inline int udc_data_in_isr(struct udc* dev, int ep_ix)
{
        int ret_val = 0;
        u32 tmp;
        struct udc_ep   *ep;
        struct udc_request              *req;
        struct udc_data_dma     *td;
        unsigned dma_done;
        unsigned        len;

        ep = &dev->ep [ep_ix];
        
        tmp = readl(&ep->regs->sts);
        if (use_dma) {
                /* BNA ? */
                if (tmp & AMD_BIT(UDC_EPSTS_BNA)) {
                        ERR("BNA ep%din occured - DESPTR = %08lx \n",
                            ep->num, (unsigned long) readl(&ep->regs->desptr));
                        
                        /* clear BNA */
                        writel(tmp, &ep->regs->sts); 
                        /* TODO no early return */

                        return 1;                       
                } 
        }
        /* HE event ? */
        if (tmp & AMD_BIT(UDC_EPSTS_HE)) {
                ERR("HE ep%dn occured -  DESPTR = %08lx \n",
                     ep->num, (unsigned long) readl(&ep->regs->desptr));
                
                /* clear HE */
                writel(tmp | AMD_BIT(UDC_EPSTS_HE), &ep->regs->sts); 
                /* TODO no early return */
                return 1;                       
        } 

        tmp= readl(&ep->regs->sts);
        
        /* DMA completion */
        if (tmp & AMD_BIT(UDC_EPSTS_TDC)) {
                VDBG("TDC set- completion\n");
                if (!list_empty (&ep->queue)) {
                       req = list_entry (ep->queue.next,
                                        struct udc_request, queue);
                       if (req) { 
                                /* lengh bytes transfered */
                                /* check dma done of last desc. in PPBDU mode */
                                if (use_dma_ppb_du) {
                                        td = udc_get_last_dma_desc(req);
                                        if (td) {
                                                /* RTL bug (Bugzilla #3672): the TDC */
                                                /* irq is quicker */
                                                /* than the desc. update in memory */
                                                /* This must be fixed in RTL !!! */
                                                 /* TDC is set, but DMA is busy */
                                                dma_done = 
                                                   AMD_GETBITS(td->dmastat, UDC_DMA_IN_STS_BS); 
                                                /* don't care DMA done as long the above mentioned RTL bug is */
                                                /* present, TODO check dma_done if this is fixed */
                                                //if (dma_done == UDC_DMA_IN_STS_BS_DMA_DONE) {
                                                        req->req.actual = req->req.length;
                                                //}
                                        }
                                }
                                else {
                                        /* assume all bytes transferred */
                                        /* TODO check error status */
                                        req->req.actual = req->req.length;
                                }

                                if (req->req.actual == req->req.length) { 
#ifdef UDC_IPBUG_2253_WORKAROUND
                                        /* stop NAKing after small packet DMA */
                                        if (ep->naking) {
                                                /* clear NAK by writing CNAK */
                                                tmp = readl(&ep->regs->ctl);
                                                tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
                                                writel(tmp, &ep->regs->ctl);
                                                ep->naking = 0;
                                                UDC_QUEUE_CNAK(ep, ep->num);
                                        }
#endif
                                        /* complete req */
                                        done (ep, req, 0);
                                        req->dma_pend = 0;
#ifdef UDC_DISABLE_IRQ_IF_EMPTY_IN_QUEUE
                                        /* further request available ? */
                                        if (list_empty (&ep->queue)) {
                                                /* disable interrupt */
                                                tmp = readl(&dev->regs->ep_irqmsk);
                                                tmp |= AMD_BIT(ep->num);
                                                writel(tmp, &dev->regs->ep_irqmsk);        
                                        } 
#endif

                                        /* ep stopped and marked for stall ? */
                                        if (ep->stopped) {
                                                /* flush the txfifo */
                                                tmp = readl(&ep->regs->ctl);
                                                tmp |= AMD_BIT(UDC_EPCTL_F);
                                                writel(tmp, &ep->regs->ctl);
                                                
                                                /* set STALL */
                                                DBG("ep-%d set STALL\n", ep->num);
                                                tmp = readl(&ep->regs->ctl);
                                                tmp |= AMD_BIT(UDC_EPCTL_S);
                                                writel(tmp, &ep->regs->ctl);
                                                //ep->stopped = 0;
                                        }
                                }
                       }
               }
        

                /* clear TDC bit */
                writel(AMD_BIT(UDC_EPSTS_TDC),&ep->regs->sts);

                 
        } /* status reg has IN bit set ? */
        else if (tmp & AMD_BIT(UDC_EPSTS_IN)) {
                if (!list_empty (&ep->queue))
                {
                        /* next request */
                        req = list_entry (ep->queue.next,
                                        struct udc_request, queue);
                        /* FIFO mode ********/ 
                        if (!use_dma) {
                                /* write fifo */
                                write_fifo(ep, &(req->req));
                                len = req->req.length - req->req.actual;
				if (len > ep->ep.maxpacket)
					len = ep->ep.maxpacket;
				req->req.actual += len;
                                if (req->req.actual == req->req.length 
                                    || (len != ep->ep.maxpacket)) {
                                    //&& (!req->req.zero || len != ep->ep.maxpacket)) {
                                        /* complete req */
                                        done (ep, req, 0);
                                }
                        } /* DMA *****************/
                        else if (req && !req->dma_pend) { 
                                VDBG("IN DMA : req=%lx req->td_data=%lx\n",
                                     (unsigned long) req, (unsigned long) req->td_data);
                                if (req->td_data) {

                                        req->dma_pend = 1;

                                        /* unset L bit of first desc. for chain */
                                        if (use_dma_ppb && req->req.length > ep->ep.maxpacket) {
                                                req->td_data->dmastat &= AMD_CLEAR_BIT(UDC_DMA_IN_STS_L);
                                        }

                                        /* write desc pointer */
                                        writel(req->td_dma, &ep->regs->desptr);

                                        wb_flush();                                        
                                        /* set HOST READY */
                                        req->td_data->dmastat 
                                              = AMD_ADDBITS(req->td_data->dmastat,
                                                            UDC_DMA_IN_STS_BS_HOST_READY,
                                                            UDC_DMA_IN_STS_BS);

                                        wb_flush();

#ifdef UDC_IPBUG_2253_WORKAROUND
                                        /* NAK if small packet until TDC interrupt */
                                        if (req->req.length < UDC_SMALL_PACKET) {
                                                /* set NAK */
                                                tmp = readl(&ep->regs->ctl);
                                                tmp |= AMD_BIT(UDC_EPCTL_SNAK); 
                                                writel(tmp, &ep->regs->ctl);
                                                ep->naking = 1;
                                                wb_flush();
                                        }
#endif

        
                                        /* set poll demand bit */
                                        tmp = readl(&ep->regs->ctl);
                                        tmp |= AMD_BIT(UDC_EPCTL_P);
                                        writel(tmp, &ep->regs->ctl);
                                }
                        }

                } 

                /* clear IN bit */
                writel(AMD_BIT(UDC_EPSTS_IN),&ep->regs->sts);
        }
       
        return ret_val;
 
}



/**
 * Interrupt handler for global device events  
 *
 * \param dev           pointer to UDC device object    
 * \param dev_irq       device interrupt bit of DEVINT register    
 * \return 0 if success
 */
static inline int udc_dev_isr(struct udc* dev, u32 dev_irq)
{
        int ret_val = 0;
        u32 tmp;
        u32 cfg;
        struct udc_ep   *ep;
        u16 i;
        u8 udc_csr_epix;

#ifdef UDC_IPCASE_8000009363_WORKAROUND
        /* SOF interrupt for polling STALL */
        if (dev_irq & AMD_BIT(UDC_DEVINT_SOF)) {
                /* check if stopped ep after STALL whether
                   STALL cleared by CLEAR_FEATURE */
                ep = &dev->ep [UDC_EPIN_IX];
                if (ep->stopped) {
                        tmp = readl(&ep->regs->ctl);
                        DBG("check STALL for ep%d\n",ep->num);
                        /* STALL cleared ? */
                        if (!(tmp & AMD_BIT(UDC_EPCTL_S))) {
                                DBG("clear NAK after STALL\n");
                                /* clear NAK by writing CNAK */
                                tmp = readl(&ep->regs->ctl);
                                tmp |= AMD_BIT(UDC_EPCTL_CNAK); 
                                writel(tmp, &ep->regs->ctl);
                                ep->stopped = 0;
                                UDC_QUEUE_CNAK(ep, ep->num);
                        }
                }
                        
        }
#endif
        
        /* SET_CONFIG irq ? */
        if (dev_irq & AMD_BIT(UDC_DEVINT_SC)) {

                /* read config value */
                tmp = readl(&dev->regs->sts);
                cfg = AMD_GETBITS(tmp, UDC_DEVSTS_CFG);
#ifdef DEBUG
                /* this is needed for debug only */
                if (cfg == dev->cur_config) {
                        same_cfg = 1;
                }
                else {
                        same_cfg = 0;
                }
                VDBG("same_cfg=%d\n", same_cfg);
#endif
                DBG("SET_CONFIG interrupt: config=%d\n", cfg);
                dev->cur_config = cfg;

                /* make usb request for gadget driver */
                memset(&setup_data, 0 , sizeof(union udc_setup_data));
                setup_data.r.bRequest = USB_REQ_SET_CONFIGURATION;
                setup_data.r.wValue = dev->cur_config;

                /* programm the NE registers */
                /* TODO - put this to extra function or use usb_reinit() or udc_enable() */
                for (i = 0; i < UDC_EP_NUM; i++) {
                        ep = &dev->ep [i];
                        if (ep->is_in) {
                                
                                /* ep ix in UDC CSR register space */
                                udc_csr_epix = ep->num;

                                
                        } /* OUT ep */
                        else {
                                /* ep ix in UDC CSR register space */
                                udc_csr_epix = ep->num - UDC_CSR_EP_OUT_IX_OFS;
                        }

                        tmp = readl(&dev->csr->ne[udc_csr_epix]);
                        /* ep cfg */
                        tmp = AMD_ADDBITS(tmp, ep->dev->cur_config, UDC_CSR_NE_CFG);
                        /* write reg */

                        writel(tmp, &dev->csr->ne[udc_csr_epix]);
                }
                /* call gadget zero with setup data received */
                spin_unlock (&dev->lock);
                tmp = dev->driver->setup (&dev->gadget, &setup_data.r);
                spin_lock (&dev->lock);

                /* set CSR_DONE to complete */
                tmp = readl(&dev->regs->ctl);
                tmp |= AMD_BIT(UDC_DEVCTL_CSR_DONE);
                writel(tmp, &dev->regs->ctl);
                                      
        } /* SET_INTERFACE ? */
        if (dev_irq & AMD_BIT(UDC_DEVINT_SI)) {
                /* read interface and alt setting values */
                tmp = readl(&dev->regs->sts);
                dev->cur_alt = AMD_GETBITS(tmp, UDC_DEVSTS_ALT);
                dev->cur_intf = AMD_GETBITS(tmp, UDC_DEVSTS_INTF);
                                      
                /* make usb request for gadget driver */
                memset(&setup_data, 0 , sizeof(union udc_setup_data));
                setup_data.r.bRequest = USB_REQ_SET_INTERFACE;
                setup_data.r.bRequestType = USB_RECIP_INTERFACE;
                setup_data.r.wValue = dev->cur_alt;
                setup_data.r.wIndex = dev->cur_intf;
                
                DBG("SET_INTERFACE interrupt: alt=%d intf=%d\n", dev->cur_alt, dev->cur_intf);

                /* programm the NE registers */
                /* TODO - put this to extra function or use usb_reinit() or udc_enable() */
                for (i = 0; i < UDC_EP_NUM; i++) {
                        ep = &dev->ep [i];
                        if (ep->is_in) {
                                
                                /* ep ix in UDC CSR register space */
                                udc_csr_epix = ep->num;

                                
                        } /* OUT ep */
                        else {
                                /* ep ix in UDC CSR register space */
                                udc_csr_epix = ep->num - UDC_CSR_EP_OUT_IX_OFS;
                        }

                        /***** UDC CSR reg ****************************/
                        /* set ep values  */
                        tmp = readl(&dev->csr->ne[udc_csr_epix]);
                        /* ep interface */
                        tmp = AMD_ADDBITS(tmp, ep->dev->cur_intf, UDC_CSR_NE_INTF);
                        //tmp = AMD_ADDBITS(tmp, 2, UDC_CSR_NE_INTF);
                        /* ep alt */
                        tmp = AMD_ADDBITS(tmp, ep->dev->cur_alt, UDC_CSR_NE_ALT);
                        //tmp = AMD_ADDBITS(tmp, 1, UDC_CSR_NE_ALT);
                        /* write reg */
                        writel(tmp, &dev->csr->ne[udc_csr_epix]);
                }

                /* call gadget zero with setup data received */
                spin_unlock (&dev->lock);
                tmp = dev->driver->setup (&dev->gadget, &setup_data.r);
                spin_lock (&dev->lock);

                /* set CSR_DONE to complete */
                tmp = readl(&dev->regs->ctl);
                tmp |= AMD_BIT(UDC_DEVCTL_CSR_DONE);
                writel(tmp, &dev->regs->ctl);
        } /* USB reset */
        if (dev_irq & AMD_BIT(UDC_DEVINT_UR)) {
                DBG("USB Reset interrupt\n");

                /* allow soft reset when suspend occurs */
                soft_reset_occured = 0;

                /* mask not needed interrupts */
                udc_mask_unused_interrupts(dev);

                /* call gadget to reset configs etc. */
                spin_unlock (&dev->lock);
                dev->driver->disconnect (&dev->gadget);
                spin_lock (&dev->lock);

                /* disable ep0 to empty req queue */
                nuke (&dev->ep [UDC_EP0IN_IX]);
                ep_reset (dev->regs,
                          &dev->ep [UDC_EP0IN_IX]);

#ifdef UDC_IPBUG_3958_WORKAROUND_SOFT_RESET_ON_USBRESET
                /* soft reset when rxfifo not empty */
                tmp = readl(&dev->regs->sts);
                if (!(tmp & AMD_BIT(UDC_DEVSTS_RXFIFO_EMPTY)) && 
                    !soft_reset_after_usbreset_occured) {
                        udc_soft_reset(dev);
                        soft_reset_after_usbreset_occured++;
                } 
#endif

                /* DMA reset to kill potential old DMA hw hang, */
                /* POLL bit is already reset by ep_reset() through */
                /* disconnect() */
                UDC_DMARST(tmp,dev);
                
                /* put into initial config */
                usb_reset (dev);

#ifdef UDC_IPBUG_3958_WORKAROUND_RXFIFO_FLUSH_ON_USBRESET
                /* TODO this workaround hang sometimes */
                {
                        u32 reg;
                        u32 val;
                        /* TODO outsource this */
                        /* take care of PCSBUG#1503, empty rxfifo to allow CNAK */
                        /* disable DMA */
                        tmp = readl(&dev->regs->ctl);
                        writel(tmp & AMD_UNMASK_BIT(UDC_DEVCTL_MODE), &dev->regs->ctl);
                        /* empty rxfifo */
                        reg = readl(&dev->regs->sts);
                        while (!(reg & AMD_BIT(UDC_DEVSTS_RXFIFO_EMPTY))) {
                                DBG("RXFIFO not empty\n");
                                /* clear all OUT and BNA bits */
                                for (i = UDC_EPIN_NUM; i < UDC_EP_NUM; i++) {
                                        ep = &dev->ep [i];
                                        reg = readl(&ep->regs->sts);
                                        /* BNA clear */
                                        if (reg & AMD_BIT(UDC_EPSTS_BNA)) {
                                                ERR("BNA ep%d occured - ep sts = %08lx \n", ep->num, (unsigned long) reg);
                                                /* clear BNA */
                                                writel(reg | AMD_BIT(UDC_EPSTS_BNA), &ep->regs->sts); 
                                        }
                                        if (reg & AMD_BIT(UDC_EPSTS_OUT_DATA) || 
                                            reg & AMD_BIT(UDC_EPSTS_OUT_SETUP)) { 
                                                /* received number bytes */
                                                val = readl(&ep->regs->sts);
                                                val = AMD_GETBITS(val, UDC_EPSTS_RX_PKT_SIZE);
                                                /* read rx fifo bytes */
                                                DBG("Clear ep%d rxfifo bytes count=%d\n", ep->num, val);
                                                udc_read_fifo_bytes(dev, udc_rxfifo_trash, val); 
                                                /* clear OUT/SETUP bits in ep status */
                                                writel(reg | UDC_EPSTS_OUT_SETUP_CLEAR | 
                                                                UDC_EPSTS_OUT_DATA_CLEAR, &ep->regs->sts);
                                        }
                                }
                                reg = readl(&dev->regs->sts);
                        }
                        /* re-enable DMA if was active */
                        writel(tmp, &dev->regs->ctl);
                }
#endif

                /* enable device setup interrupts */
                udc_enable_dev_setup_interrupts(dev);

        } /* USB suspend */
#ifndef UDC_IPBUG_3943_WORKAROUND
        if (dev_irq & AMD_BIT(UDC_DEVINT_US)) {
                DBG("USB Suspend interrupt\n");

        } /* new speed ? */
#endif
        if (dev_irq & AMD_BIT(UDC_DEVINT_ENUM)) {
                DBG("ENUM interrupt\n");
#ifdef DEBUG
                num_enums++;
                DBG("%d enumerations !\n", num_enums);
#endif
                soft_reset_after_usbreset_occured = 0;

                /* disable ep0 to empty req queue */
                nuke (&dev->ep [UDC_EP0IN_IX]);
                ep_reset (dev->regs,
                          &dev->ep [UDC_EP0IN_IX]);

                /* link up all endpoints */
                usb_reinit (dev);
                if (dev->gadget.speed == USB_SPEED_HIGH) {
                        INFO("Connect: Speed = HIGH_SPEED\n");
                }
                else if (dev->gadget.speed == USB_SPEED_FULL) {
                        INFO("Connect: Speed = FULL_SPEED\n");
                }

                /* init ep 0 */
                ep0_start(dev);

                /* enable ep0 interrupts */
                udc_enable_ep0_interrupts(dev);
        }
#ifndef UDC_IPBUG_3950_WORKAROUND 
        /* session valid change interrupt */
        if (dev_irq & AMD_BIT(UDC_DEVINT_SVC)) {
                DBG("USB SVC interrupt\n");

                /* check that session is not valid to detect disconnect */
                tmp = readl(&dev->regs->sts);
                if (!(tmp & AMD_BIT(UDC_DEVSTS_SESSVLD))) {
                        DBG("USB Disconnect (session valid low)\n");
                        /* cleanup on disconnect */
                        usb_disconnect(the_controller);
                }

        }
#endif

        return ret_val;
}

/**
 * Interrupt Service Routine, see Linux Kernel Doc for parameters  
 *
 * \param irq           irq number
 * \param _dev          pointer to device object
 * \param r             don't used
 */
static irqreturn_t udc_irq (int irq, void *_dev, struct pt_regs * r)
{
        struct udc              *dev = _dev;
        u32 reg;
        u16 i;
        u32 ep_irq;


#ifdef UDC_IPBUG_3943_WORKAROUND
        /* If UDC is suspended, then don't touch any register, otherwise
           system hangs in endless retry => possibly hang !!! */
        if (dev->otg_driver && dev->otg_driver->query) {
                if (dev->otg_driver->query(0) & OTG_FLAGS_UDC_SUSP) {
                        return IRQ_HANDLED;
                }     
        }
        else
                return IRQ_HANDLED;
#endif 
       
        if (dev->sys_suspended)
                return IRQ_HANDLED;

        spin_lock (&dev->lock);


        /* check for ep irq */
        reg = readl(&dev->regs->ep_irqsts);
        if (reg)
        {
                /* EP0 OUT */
                if (reg & AMD_BIT(UDC_EPINT_OUT_EP0))
                {
                        udc_control_out_isr(dev);
                } /* EP0 IN */
                if (reg & AMD_BIT(UDC_EPINT_IN_EP0)) {
                        udc_control_in_isr(dev);
                        
                } 

                /* data endpoint */
                /* iterate ep's */
                for (i = 1; i < UDC_EP_NUM; i++) {
                        ep_irq = 1 << i;
                        /* irq for out ep ? */
                        if ((reg & ep_irq) && i > UDC_EPIN_NUM) {      
                                /* clear irq */
                                writel(ep_irq, &dev->regs->ep_irqsts);
                                udc_data_out_isr(dev,i);
                        } /* irq for in ep ? */
                        if ((reg & ep_irq) && i < UDC_EPIN_NUM && i > 0) {      
                                /* clear irq */
                                writel(ep_irq, &dev->regs->ep_irqsts);
                                udc_data_in_isr(dev,i);
                        }
                        
                }
                                          
        }


        /* check for dev irq */
        reg = readl(&dev->regs->irqsts);
        if (reg) {
                /* clear irq */
                writel(reg, &dev->regs->irqsts);
                udc_dev_isr(dev, reg);
        }
  

        spin_unlock (&dev->lock);
        return IRQ_HANDLED;
}

/**
 * Tears down device 
 *
 * \param _dev        pointer to device struct
 */
#if     LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
static void gadget_release (struct device *_dev)
{
        struct amd5536udc *dev = dev_get_drvdata (_dev);

        kfree (dev);
}
#endif

/**
 * Tears down the binding between this driver and the pci device
 *
 * \param pdev        pointer to pci device struct
 */
static void udc_remove (struct pci_dev *pdev)
{
        struct udc              *dev;

#ifdef DEBUG
        /* debug */
        print_misc(the_controller);
#endif
        dev = pci_get_drvdata (pdev);
        /* start with the driver above us */
        if (dev->driver) {
                /* should have been done already by driver model core */
                WARN( "pci remove, driver '%s' is still registered\n",
                                dev->driver->driver.name);
                usb_gadget_unregister_driver (dev->driver);
        }
        /* check for otg driver above us */
        if (dev->otg_driver) {
                /* should have been done already by driver model core */
                WARN( "otg driver is still registered\n");
                //usb_gadget_unregister_otg (dev->otg_driver);
        }
        /* then clean up the resources we allocated during probe() */
        if (dev->data_requests) {
                pci_pool_destroy (dev->data_requests);
        }
        if (dev->stp_requests) {
                /* cleanup DMA desc's for ep0in */
                pci_pool_free (dev->stp_requests,
                               dev->ep [UDC_EP0OUT_IX].td_stp,
                               dev->ep [UDC_EP0OUT_IX].td_stp_dma);
                pci_pool_free (dev->stp_requests,
                               dev->ep [UDC_EP0OUT_IX].td,
                               dev->ep [UDC_EP0OUT_IX].td_dma);

                pci_pool_destroy (dev->stp_requests);
        }

        /* init controller by soft reset */
        writel(AMD_BIT(UDC_DEVCFG_SOFTRESET), &dev->regs->cfg);
        if (dev->got_irq)
                free_irq (pdev->irq, dev);
        if (dev->regs)
                iounmap (dev->regs);
        if (dev->region)
                release_mem_region (pci_resource_start (pdev, 0),
                                pci_resource_len (pdev, 0));
#ifndef CONFIG_USB_NON_PCI_OTGDEVICE
        if (dev->enabled)
                pci_disable_device (pdev);
#endif

#if     LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
#ifndef CONFIG_MIPS
        /* FIXME TMP26: temporary disabled device file system usage until
         * working for Au1200 */
        device_unregister (&dev->gadget.dev);
        device_remove_file (&pdev->dev, &dev_attr_registers);
#endif
#endif

        pci_set_drvdata (pdev, 0);

#ifdef CONFIG_USB_NON_PCI_OTGDEVICE
        {
                u32 tmp;
                /* disable UDC memory, DMA and clock */
                tmp = readl((u32*) (USB_MSR_BASE + USB_MSR_MCFG));
                tmp &= AMD_CLEAR_BIT(USBMSRMCFG_DMEMEN)
                       & AMD_CLEAR_BIT(USBMSRMCFG_DBMEN) 
                       & AMD_CLEAR_BIT(USBMSRMCFG_UDCCLKEN);
                writel(tmp, USB_MSR_BASE + USB_MSR_MCFG); 
        }
#endif

#if     LINUX_VERSION_CODE < UDC_NEW_GADGET_KERNEL || defined(CONFIG_MIPS)
        /* kfree is called by gadget.dev.release() for 2.6.x 
         * TMP26: device registration currently disabled for Au1200 */
        kfree (dev);
#endif
        the_controller = 0;
}

/**
 * Called when pci device is recognized and driver is loaded by kernel,
 * init register access and register interrupt, but wait for gadget
 * driver registering before any USB traffic  
 *
 * \param pdev        pointer to pci device struct
 * \param id          pointer to pci device id
 * \return 0 if success
 */
static int udc_probe (struct pci_dev *pdev, const struct pci_device_id *id)
{
        struct udc              *dev;
        unsigned long           resource, len;
        void                    *base = 0;
        char                    buf [8], *bufp;
        struct udc_stp_dma      *td_stp;
        struct udc_data_dma     *td_data;
        int retval;

        VDBG("udc_probe()\n");

#ifdef CONFIG_USB_NON_PCI_OTGDEVICE
        {
                u32     tmp;
                /*
                 * Fill in the dummy pci_dev. Only those members actually
                 * dereferenced in this driver are initialized.
                 */
                memset(pdev, 0, sizeof(struct pci_dev));
#if     LINUX_VERSION_CODE < UDC_NEW_GADGET_KERNEL
                strcpy(pdev->name, "non-PCI UDC");
#endif
                pdev->resource[0].name = "UDC Registers";
                pdev->resource[0].start = USB_UDC_BASE;
                pdev->resource[0].end = USB_UDC_BASE + USB_UDC_LEN - 1;
                pdev->resource[0].flags = 0;
                pdev->irq = AU1200_USB_INT;

                tmp = readl((u32*) (USB_MSR_BASE + USB_MSR_MCFG));
                if (tmp == 0) {
                        /* default value */
                        tmp = USBMSRMCFG_DEFAULT;        
                        writel(tmp, USB_MSR_BASE + USB_MSR_MCFG); 
                        readl((u32*) USB_MSR_BASE + USB_MSR_MCFG); 
                        udelay(1000);
                }
                /* enable UDC memory, DMA, clock, cacheable memory,
                 * read combining and prefetch enable */
                tmp |= AMD_BIT(USBMSRMCFG_DMEMEN) | AMD_BIT(USBMSRMCFG_DBMEN)
                        | AMD_BIT(USBMSRMCFG_UDCCLKEN)
                        | AMD_BIT(USBMSRMCFG_PHYPLLEN)
#ifdef CONFIG_DMA_COHERENT
                        | AMD_BIT(USBMSRMCFG_UCAM)
#endif
                        | AMD_BIT(USBMSRMCFG_RDCOMB)
                        | AMD_BIT(USBMSRMCFG_PFEN);
                writel(tmp, USB_MSR_BASE + USB_MSR_MCFG); 
        }
#endif
        /* if you want to support more than one controller in a system,
         * usb_gadget_driver_{register,unregister}() must change.
         */
        if (the_controller) {
                WARN ("ignoring %04x/%04x\n", UDC_PCI_VENDOR_ID, UDC_PCI_DEVICE_ID);
                return -EBUSY;
        }

        /* alloc, and start init */
        dev = kmalloc (sizeof (struct udc), SLAB_KERNEL);
        if (dev == NULL){
                retval = -ENOMEM;
                goto done;
        }

        memset (dev, 0, sizeof (struct udc));
        spin_lock_init(&dev->lock);
        spin_lock_init(&udc_irq_spinlock);
        dev->pdev = pdev;
        dev->gadget.ops = &udc_ops;

#if     LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
        strcpy(dev->gadget.dev.bus_id, "gadget");
        dev->gadget.dev.parent = &pdev->dev;
        dev->gadget.dev.dma_mask = pdev->dev.dma_mask;
        dev->gadget.dev.release = gadget_release;
        dev->gadget.name = driver_name;
#else
        dev->gadget.dev.bus_id = "gadget";
#endif
        dev->gadget.name = driver_name;
	dev->gadget.is_dualspeed = 1;

#ifndef CONFIG_USB_NON_PCI_OTGDEVICE
        /* now all the pci goodies ... */
        if (pci_enable_device (pdev) < 0) {
                retval = -ENODEV;
                goto done;
        }
#endif
        dev->enabled = 1;

        resource = pci_resource_start (pdev, 0);
        len = pci_resource_len (pdev, 0);

        if (!request_mem_region (resource, len, driver_name)) {
                DBG( "controller already in use\n");
                retval = -EBUSY;
                goto done;
        }

        dev->region = 1;

        base = ioremap_nocache (resource, len);
        if (base == NULL) {
                DBG( "can't map memory\n");
                retval = -EFAULT;
                goto done;
        }

        /* udc csr registers base */
        dev->csr = (struct udc_csrs*) (base + UDC_CSR_ADDR);
        /* dev registers base */
        dev->regs = (struct udc_regs *) (base + UDC_DEVCFG_ADDR);                
        /* ep registers base */
        dev->epregs = (struct udc_epregs *) (base + UDC_EPREGS_ADDR);                
        /* fifo's base */
        dev->rxfifo = (u32*) (base + UDC_RXFIFO_ADDR);                
        dev->txfifo = (u32*) (base + UDC_TXFIFO_ADDR);                

        /* init registers, interrupts, ... */
#ifndef UDC_IPBUG_3943_WORKAROUND
        startup_registers(dev);
#else
        {
                u32 tmp;

                /* TODO put this to extra function,
                 * this all is extracted from usb_init() and
                 * usb_reset() but without register access */
                dev->gadget.ep0 = &dev->ep [UDC_EP0IN_IX].ep;
                dev->ep [UDC_EP0IN_IX].stopped = 0;
                INIT_LIST_HEAD (&dev->gadget.ep0->ep_list);
                dev->gadget.speed = USB_SPEED_HIGH;
                /* standard fifo and endpoint allocations */
                set_fifo_mode (dev, 0);
                /* basic endpoint init */
                for (tmp = 0; tmp < UDC_EP_NUM; tmp++) {
                        struct udc_ep   *ep = &dev->ep [tmp];

                        ep->ep.name = ep_name [tmp];
                        ep->dev = dev;
                        ep->num = tmp;
                        /* txfifo size is calculated at enable time */
                        ep->txfifo = dev->txfifo;

                        /* fifo size */
                        if (tmp < UDC_EPIN_NUM) {
                                ep->fifo_size = UDC_TXFIFO_SIZE;
                                ep->is_in = 1;
                        }
                        else {
                                ep->fifo_size = UDC_RXFIFO_SIZE;
                                ep->is_in = 0;

                        }

                        ep->regs = &dev->epregs [tmp];
                        if (!ep->desc) {
                                ep->desc = 0;
                                INIT_LIST_HEAD (&ep->queue);

                                ep->ep.maxpacket = ~0;
                                ep->ep.ops = &udc_ep_ops;
                        }
                        if (use_dma) {
                                /* TODO ep->dma is not really used, just to indicate that */
                                /* DMA is active: remove this */
                                /* dma regs = dev control regs */        
                                ep->dma = (u32*) &dev->regs->ctl;       
                        }
                }
                dev->ep [UDC_EP0IN_IX].ep.maxpacket = UDC_EP0IN_MAX_PKT_SIZE;
                dev->ep [UDC_EP0OUT_IX].ep.maxpacket = UDC_EP0OUT_MAX_PKT_SIZE;
        }
#endif

        /* irq setup after old hardware is cleaned up */
        if (!pdev->irq) {
                ERR( "No IRQ.  Check PCI setup!\n");
                retval = -ENODEV;
                goto done;
        }
#ifndef __sparc__
        snprintf (buf, sizeof buf, "%d", pdev->irq);
        bufp = buf;
#else
        bufp = __irq_itoa(pdev->irq);
#endif
        if (request_irq (pdev->irq, udc_irq, SA_SHIRQ, driver_name, dev)
                        != 0) {
                ERR( "request interrupt %s failed\n", bufp);
                retval = -EBUSY;
                goto done;
        }
        dev->got_irq = 1;

        if (use_dma) {
                /* consistent DMA mode setting ? */
                if (use_dma_ppb) {
                        use_dma_bufferfill_mode = 0;
                }
                else {
                        use_dma_ppb_du = 0;
                        use_dma_bufferfill_mode = 1;
                }
                /* DMA setup */
#if     LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
                dev->data_requests = pci_pool_create ("data_requests", pdev,
                        sizeof (struct udc_data_dma),
                        UDC_PCIPOOL_ALIGN,
                        UDC_PCIPOOL_CROSS);
#else
                dev->data_requests = pci_pool_create ("data_requests", pdev,
                        sizeof (struct udc_data_dma),
                        UDC_PCIPOOL_ALIGN,
                        UDC_PCIPOOL_CROSS,
                        SLAB_KERNEL /* 2.4 only */ );
#endif
                if (!dev->data_requests) {
                        DBG( "can't get request data pool\n");
                        retval = -ENOMEM;
                        goto done;
                }

                /* EP0 in dma regs = dev control regs */        
                dev->ep[UDC_EP0IN_IX].dma = &dev->regs->ctl;       

                /* dma desc for setup data */
#if     LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
                dev->stp_requests = pci_pool_create ("setup requests", pdev,
                        sizeof (struct udc_stp_dma),
                        UDC_PCIPOOL_ALIGN,
                        UDC_PCIPOOL_CROSS);
#else
                dev->stp_requests = pci_pool_create ("setup requests", pdev,
                        sizeof (struct udc_stp_dma),
                        UDC_PCIPOOL_ALIGN,
                        UDC_PCIPOOL_CROSS,
                        SLAB_KERNEL /* 2.4 only */ );
#endif
                if (!dev->stp_requests) {
                        DBG( "can't get stp request pool\n");
                        retval = -ENOMEM;
                        goto done;
                }
                /* setup */
                td_stp = pci_pool_alloc (dev->stp_requests, UDC_PCIPOOL_GFP_STP,
                                        &dev->ep [UDC_EP0OUT_IX].td_stp_dma);
                if (td_stp == NULL){
                        retval = -ENOMEM;
                        goto done;
                }
                dev->ep [UDC_EP0OUT_IX].td_stp = td_stp;
                /* data: 0 packets !? */
                td_data = pci_pool_alloc (dev->stp_requests, UDC_PCIPOOL_GFP_STP,
                                        &dev->ep [UDC_EP0OUT_IX].td_dma);
                if (td_data == NULL){
                        retval = -ENOMEM;
                        goto done;
                }
                dev->ep [UDC_EP0OUT_IX].td = td_data;
#ifdef UDC_IPDEFECT_9000004946_WORKAROUND
                /* point to itself */
                dev->ep [UDC_EP0OUT_IX].td->next = dev->ep [UDC_EP0OUT_IX].td_dma;
#endif
        }

        dev->chiprev = 0;

        /* normally Linux inits PCI_LATENCY_TIMER */
#ifndef CONFIG_USB_NON_PCI_OTGDEVICE
        pci_set_master (pdev);
#ifdef  HAVE_PCI_SET_MWI
        pci_set_mwi (pdev);
#endif
        /* chip rev for Hiddensee */
        pci_read_config_byte(pdev, PCI_REVISION_ID, (u8*) &dev->chiprev);
#else
        /* chip rev for Au1200 */
        dev->chiprev = (u16) read_c0_prid() & 0xff;
#endif

        /* done */
        pci_set_drvdata (pdev, dev);

        INFO( "%s\n", driver_desc);

        /* TODO make defines for rev. strings */
#ifdef CONFIG_SOC_AU1200
        INFO( "irq %s, mem %08lx, chip rev %02x (Au1200 %s)\n",
                        bufp, resource, dev->chiprev, (dev->chiprev == 0) ? "AB" : "AC");
        bufp = DRIVER_VERSION;
#ifdef CONFIG_DMA_COHERENT
        /* coherent DMA not possible with AB silicon */
        if (dev->chiprev == UDC_AUAB_REV) {
                ERR("Your chip revision is %s, it must be at least %s to use coherent DMA. \nPlease change DMA_COHERENT to DMA_NONCOHERENT in arch/mips/Kconfig and re-compile .\n",
                    "AB", "AC");
                retval = -ENODEV;
                goto done;
        }
#endif
        
#ifdef UDC_AUA1
        if (dev->chiprev < UDC_AUA1) {
                ERR("Your chip revision is %s, it must be at least %s\n",
                    "AB", "AC");
                retval = -ENODEV;
                goto done;
        }
        INFO("driver version: %s (for Au1200 AC)\n", bufp);
#else
        INFO("driver version: %s (for Au1200 AB)\n", bufp);
#endif
#ifdef CONFIG_DMA_COHERENT
        INFO("Compiled for coherent memory.\n");
#endif
#ifdef CONFIG_DMA_NONCOHERENT
        INFO("Compiled for non-coherent memory.\n");
#endif
#else
        INFO("irq %s, pci mem %08lx, chip rev %02x (Geode5536 %s)\n",
              bufp, resource, dev->chiprev, (dev->chiprev == UDC_HSA0_REV) ? "A0" : "B0");
        bufp = DRIVER_VERSION;
#ifdef UDC_HSB0
        if (dev->chiprev == UDC_HSA0_REV) {
                ERR("Your chip revision is %s, it must be at least %s\n",
                    "A0", "B0");
                retval = -ENODEV;
                goto done;
        }
        INFO("driver version: %s (for Geode5536 B0)\n", bufp);
#else
        INFO("driver version: %s (for Geode5536 A0)\n", bufp);
#endif
#endif
        the_controller = dev;

#if     LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL
#ifndef CONFIG_MIPS
        /* FIXME TMP26: temporary disabled device file system usage until
         * working for Au1200 */
        device_register (&dev->gadget.dev);
        device_create_file (&pdev->dev, &dev_attr_registers);
#endif
#endif

        /* print dev register info */
        print_regs(dev);

        return 0;

done:
        if (dev)
                udc_remove (pdev);
        return retval;
}


/**
 *  Initiates a remote wakeup 
 *
 * \return 0 if success
 */
/* initiate remote wakeup */
static int udc_remote_wakeup(struct udc* dev)
{
        u32 tmp;

        INFO("UDC initiates remote wakeup\n");
        
        tmp =  readl(&dev->regs->ctl);
        tmp |= AMD_BIT(UDC_DEVCTL_RES);
        writel(tmp, &dev->regs->ctl);                     
        /* wait 1ms before clear resume bit */
//DISABLEDamd         udelay(1000);
        tmp &= AMD_CLEAR_BIT(UDC_DEVCTL_RES);
        writel(tmp, &dev->regs->ctl);                     

        return 0;
}

/* PM currently under debug */
#ifdef DEBUG
/**
 *  Suspends UDC 
 *
 * \return 0 if success
 */
static int udc_suspend(struct udc* dev)
{
        int retval = 0;

        INFO("UDC suspend\n");
#ifdef CONFIG_MIPS
        u32 tmp;
        /* mask interrupts */
        udc_mask_unused_interrupts(dev);

        if (dev->driver && dev->driver->disconnect) {
                /* call gadget to reset configs etc. */
                if (spin_is_locked(&dev->lock)) {
                        spin_unlock(&dev->lock);
                        dev->driver->disconnect (&dev->gadget);
                        spin_lock(&dev->lock);
                }
                else
                        dev->driver->disconnect (&dev->gadget);

                /* disable ep0 to empty req queue */
                nuke (&dev->ep [UDC_EP0IN_IX]);
                ep_reset (dev->regs,
                          &dev->ep [UDC_EP0IN_IX]);

                /* init controller by soft reset */
                udc_soft_reset(dev);

        }
        if (dev->otg_driver && dev->otg_transceiver 
            && dev->otg_transceiver->set_peripheral) {
                /* if UDC is supended by Host or already disconnected then
                   don't force disconnect by unbind() */
                if (dev->otg_driver->query) {
                        if (!(dev->otg_driver->query(0) & OTG_FLAGS_UDC_SUSP)) {
                                /* unbind from otg driver -> host disconnect */
                                dev->otg_transceiver->set_peripheral(dev->otg_transceiver, NULL);
                                dev->connected = 0;
                        }
                }
                else
                {
                        /* unbind from otg driver -> host disconnect */
                        dev->otg_transceiver->set_peripheral (dev->otg_transceiver, NULL);
                        dev->connected = 0;
                }
        }
                
        dev->sys_suspended = 1;

        /* switch off UDC clock */
        tmp = readl((u32*) (USB_MSR_BASE + USB_MSR_MCFG));
        tmp &= AMD_CLEAR_BIT(USBMSRMCFG_UDCCLKEN);
        writel(tmp, USB_MSR_BASE + USB_MSR_MCFG); 

#endif        

        return retval;
}

/**
 *  Resumes UDC 
 *
 * \return 0 if success
 */
static int udc_resume(struct udc* dev)
{
        int retval = 0;

        INFO("UDC resume\n");
#ifdef CONFIG_MIPS
        u32 tmp;
        /* switch on UDC clock */
        tmp = readl((u32*) (USB_MSR_BASE + USB_MSR_MCFG));
        tmp |= AMD_BIT(USBMSRMCFG_UDCCLKEN);
        writel(tmp, USB_MSR_BASE + USB_MSR_MCFG); 

        dev->sys_suspended = 0;

        usb_connect(dev);
        if (dev->otg_transceiver && dev->otg_transceiver->set_peripheral) {
                /* bind to otg driver */
                dev->otg_transceiver->set_peripheral(dev->otg_transceiver, dev);
        }
        
#endif

        return retval;
}
#endif

#ifdef UDC_USE_DRIVER_REGISTER
/* TODO we provide device_driver functions to attach
 * to device system which allows power management and
 * more. Currently we just call the PCI probe function.
 * So later pci_dev dummy is to be removed */

static int udc_au1xxx_drv_probe(struct device *dev)
{
	int retval;

	DBG("udc_au1xxx_drv_probe()\n");
        retval = udc_probe(pdev, pci_ids);
	if (retval == 0)
		dev_set_drvdata(dev, (struct udc*) pci_get_drvdata(pdev));
	return retval;
}

static int udc_au1xxx_drv_remove(struct device *dev)
{
	DBG("udc_au1xxx_drv_remove()\n");
        udc_remove(pdev);
	dev_set_drvdata(dev, NULL);
	return 0;
}

static int udc_au1xxx_drv_suspend(struct device *dev, u32 state, u32 level)
{
	struct udc *udc_dev = dev_get_drvdata(dev);
        return udc_suspend(udc_dev);
}

static int udc_au1xxx_drv_resume(struct device *dev, u32 level)
{
	struct udc *udc_dev = dev_get_drvdata(dev);
        return udc_resume(udc_dev);
}

static struct device_driver udc_au1xxx_driver = {
	.name		= "au1xxx-udc",
	.bus		= &platform_bus_type,
	.probe		= udc_au1xxx_drv_probe,
	.remove		= udc_au1xxx_drv_remove,
	.suspend	= udc_au1xxx_drv_suspend,
	.resume	        = udc_au1xxx_drv_resume,
};
#endif


/**
 * Inits driver
 *
 * \return 0 if success
 */
static int __init init (void)
{
        int rc;

#ifdef DEBUG
        /* register char device */
        rc = register_chrdev(UDC_MAJOR_NUM, UDC_DEVICE_NAME, &udc_fops);
        if (rc < 0)
        {
                printk("Error registering udc char device");
        }
#endif        

#ifdef CONFIG_USB_NON_PCI_OTGDEVICE
#if   LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL && defined(UDC_USE_DRIVER_REGISTER)

        /* probe by device system */
        rc = driver_register(&udc_au1xxx_driver); 
#else
        /* manual probe */
        rc = udc_probe(pdev, pci_ids);
#endif
#else
        /* probe by PCI bus driver */
        rc = pci_module_init (&udc_pci_driver);
#endif

        return rc;
}
module_init (init);

/**
 * Cleans driver
 */
static void __exit cleanup (void)
{
#ifdef DEBUG
        /* unregister char device */
        unregister_chrdev(UDC_MAJOR_NUM, UDC_DEVICE_NAME);
#endif

#ifdef CONFIG_USB_NON_PCI_OTGDEVICE
#if   LINUX_VERSION_CODE >= UDC_NEW_GADGET_KERNEL && defined(UDC_USE_DRIVER_REGISTER)
        /* unregister at device system */
        driver_unregister(&udc_au1xxx_driver); 
#else
        /* manual remove */
        udc_remove(pdev);
#endif
#else
        /* unregister at PCI bus driver */
        pci_unregister_driver (&udc_pci_driver);
#endif

}
module_exit (cleanup);

