/*
 * AMD 5536 USB OTG controller driver
 */

/*
 * Copyright (C) 2005 AMD (http://www.amd.com)
 * Author: Karsten Boge 
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/*****************************************************************************
 *  Includes
 *****************************************************************************/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/pci.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>

#include <asm/byteorder.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/unaligned.h>

#ifdef CONFIG_USB_OTG_TEST_INTERFACE
#ifndef CONFIG_USB_OTG
#define CONFIG_USB_OTG
#endif
#endif

/* #define DEBUG */
/* #define VERBOSE */

#ifdef CONFIG_SOC_AU1200
#if    LINUX_VERSION_CODE < KERNEL_VERSION(2,5,59)
#include <asm/au1000.h>
#include <asm/au1000_gpio.h>
#else
#include <asm/mach-au1x00/au1000.h>
#include <asm/mach-au1x00/au1000_gpio.h>
#endif
#endif

#include <linux/usb.h>
#include <linux/usb_gadget.h>
#include <linux/usb_otg.h>

#define	DRIVER_DESC		"AMD 5536 USB OTG Controller"
#define	DRIVER_VERSION		"01.00.0200 - $Revision: 1.1 $"
#define DRIVER_NAME_FOR_PRINT   OTG_DRIVER_NAME

#include "amd5536otg.h"
#include "amd5536uoc.h"


/*****************************************************************************
 *  Function Declarations
 *****************************************************************************/

#ifdef CONFIG_USB_OTG_TEST_INTERFACE
static int otg_test_bind(struct otg **);
static int otg_test_unbind(struct otg **);
#endif
#ifdef CONFIG_USB_OTG
static int otg_app_request(u32);
static int otg5536_set_host (struct otg_transceiver *, struct usb_bus *);
static int otg5536_set_power (struct otg_transceiver *, unsigned);
static int otg5536_start_srp (struct otg_transceiver *);
static int otg5536_start_hnp (struct otg_transceiver *);
#endif
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)
static u32 otg_app_query(int);
static int otg5536_set_peripheral (struct otg_transceiver *,
                                   struct usb_gadget *);
#endif


/*****************************************************************************
 *  Data
 *****************************************************************************/

static const char driver_name [] = OTG_DRIVER_NAME;
static const char driver_desc [] = DRIVER_DESC;

static struct otg *the_controller;
#ifdef CONFIG_SOC_AU1200
static const char *transceiver_label = "au1200_otg";
#else
static const char *transceiver_label = "cs5536_otg";
#endif

static u32 init_state = 0;

#ifdef CONFIG_USB_OTG_TEST_INTERFACE
static struct usb_otg_test_extension otg_test_extension = {
	.bind    = otg_test_bind,
	.unbind  = otg_test_unbind,
	.request = otg_app_request,
	.query   = otg_app_query
};
#endif
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

struct usb_otg_gadget_extension otg_gadget_extension = {
#ifdef CONFIG_USB_OTG
	.request = otg_app_request,
#else
	.request = NULL,
#endif
	.query   = otg_app_query,
	.notify  = NULL
};

static u32 state_mask;
u32 otg_tmr_high_count = 0;
#endif

#ifdef CONFIG_SOC_AU1200
#ifdef USE_AU1200_PCI_DUMMY
static struct pci_dev pdev_dummy;
#endif
#if    LINUX_VERSION_CODE < KERNEL_VERSION(2,5,59)
static au1x00_gpio_bit_ctl init_usb_vbus_switch = {
	.direction = 1,  /* out */
	.data      = 1   /* set */
};
#endif
#endif


/*****************************************************************************
 *  Function Definitions
 *****************************************************************************/

/**
 * \brief
 * fill OTG transceiver struct
 *
 * \param  transceiver  OTG transceiver
 *
 * \return              void
 */
static inline void otg_init_transceiver (struct otg_transceiver *transceiver)
{
	transceiver->dev            = NULL;
	transceiver->label          = transceiver_label;
	transceiver->default_a      = 0;
	transceiver->state          = OTG_STATE_UNDEFINED;
	transceiver->prv_state      = OTG_STATE_UNDEFINED;
	transceiver->params         = 0;
	transceiver->hostcount	    = 0;
	transceiver->otg_priv       = (void *) &otg_gadget_extension;
	transceiver->host           = NULL;
#ifdef CONFIG_USB_OTG_HIGHSPEED
	transceiver->companion      = NULL;
#endif
	transceiver->gadget         = NULL;
	transceiver->port_status    = 0;
	transceiver->port_change    = 0;
#ifdef CONFIG_USB_OTG
	transceiver->set_host       = otg5536_set_host;
	transceiver->set_peripheral = otg5536_set_peripheral;
	transceiver->set_power      = otg5536_set_power;
	transceiver->start_srp      = otg5536_start_srp;
	transceiver->start_hnp      = otg5536_start_hnp;
#else
	transceiver->set_host       = NULL;
	transceiver->set_peripheral = otg5536_set_peripheral;
	transceiver->set_power      = NULL;
	transceiver->start_srp      = NULL;
	transceiver->start_hnp      = NULL;
#endif
}

#ifdef CONFIG_USB_OTGMUX_GADGET
/**
 * \brief
 * OTG state change
 *
 * subset of OTG states to support the gadget only or
 * ID pin configuration
 *
 * \param  otg          OTG controller info
 * \param  event_code   event that requested a state change
 * \param  pEvt_mask
 *
 * \return              events that were not handled here
 */
u32 otg_change_state (struct otg *otg, u32 _event, u32 *pEvt_mask)
{
	u32  event_code = _event;
	u32  uoc_status = get_status (otg);

	if (GOT_EVENT (OTG_GADGET_READY, event_code) &&
	    ((otg->transceiver.state & OTG_STATE_MASK) !=
	      OTG_STATE_UNDEFINED)) {

		/* Switch from "NO_B_DEVICE" states to normal operation or   */
		/*       deactivate operations in case gadget was unloaded   */

		CHANGE_STATE (otg, OTG_STATE_UNDEFINED, pEvt_mask);
	}
	if ((OTG_INT_TMX & event_code) && otg_tmr_high_count) {

		/* a long timer is running : decrement the high part         */

		restart_timer (otg);
		otg_tmr_high_count--;
		RES_EVENT (OTG_INT_TMX, event_code);
	}

	do switch (otg->transceiver.state & OTG_STATE_MASK) {

	/* NOT_ASSIGNED (yet): init state, 1st time after loading            */
	/* ======================================================            */

	case OTG_STATE_UNDEFINED :

		CHECK_STATE (otg, OTG_STATE_UNDEFINED, pEvt_mask);

		if (IS_FLAG_RES (otg, OTG_FLAGS_ACTIV)) {

			/* seems to be the first time: let it run !          */

			SET_FLAG (otg, OTG_FLAGS_ACTIV);
		}

		/*      muxer is still neutral                               */

		RES_EVENT ((OTG_INT_IDC | OTG_INT_TMX), event_code);

		if (IS_FLAG_RES (otg, OTG_GADGET_READY)) {

			/* NOT_READY (substate): gadget drivers not ready    */
			/* ==============================================    */
			/*                                                   */
			/*     check ID pin and enable host as needed        */

			if (IS_BIT_RES (OTG_STS_ID, uoc_status)) {

				/* ID pin connected: A-device (host)         */

				if (OTG_STATE_NO_B_DEVICE_A !=
				    otg->transceiver.state) {
					CHANGE_STATE (otg,
					              OTG_STATE_NO_B_DEVICE_A,
					              pEvt_mask);
				}
			}
			else if (IS_BIT_SET (OTG_STS_ID, uoc_status)) {

				/* ID pin not connected: disable (neutral)   */

				if (OTG_STATE_NO_B_DEVICE_B !=
				    otg->transceiver.state) {
					CHANGE_STATE (otg,
					              OTG_STATE_NO_B_DEVICE_B,
					              pEvt_mask);
				}
			}
		}
		else if ((OTG_STATE_NO_B_DEVICE_A ==
		          otg->transceiver.state) ||
		         (OTG_STATE_NO_B_DEVICE_B ==
		          otg->transceiver.state)) {

			/* Exit "not ready" state                            */
			/* ----------------------                            */

			RES_EVENT (OTG_GADGET_READY, event_code);
			CHANGE_STATE (otg, OTG_STATE_UNDEFINED, pEvt_mask);
		}

		/* ==================================================        */

#ifdef CONFIG_USB_OTGMUX_IDPIN

		else if (IS_BIT_RES (OTG_STS_ID, uoc_status)) {

			/* ID pin connected: A-device (host)                 */

			CHANGE_STATE (otg, OTG_STATE_A_HOST, pEvt_mask);
		}
#endif
		else {

			/* ID pin is not connected: B-device (peripheral)    */

			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);
		}
		break;

#ifdef CONFIG_USB_OTGMUX_IDPIN

	/* A: ID pin is connected, A-device drives VBus                      */
	/* ============================================                      */

	case OTG_STATE_A_HOST :

		CHECK_STATE (otg, OTG_STATE_A_HOST, pEvt_mask);

		if (IS_BIT_SET (OTG_STS_ID, uoc_status)) {

			/* ID change => B_IDLE                               */

			RES_EVENT ((OTG_INT_IDC | OTG_INT_PCC), event_code);
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);
		}
		break;

	/* B: ID pin is not connected                                        */
	/* ==========================                                        */
#endif

	case OTG_STATE_B_IDLE :

		/* B_IDLE: init state for B-devices                          */
		/*         monitor VBus, no connection, no activity          */

		CHECK_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);

#ifdef CONFIG_USB_OTGMUX_IDPIN

		if (IS_BIT_RES (OTG_STS_ID, uoc_status)) {

			/* ID pin changed => A_HOST                          */

			RES_EVENT (OTG_INT_IDC, event_code);
			CHANGE_STATE (otg, OTG_STATE_A_HOST, pEvt_mask);
		}
		else if (IS_BIT_SET (OTG_STS_SESSVLD, uoc_status))
#else
		if (IS_BIT_SET (OTG_STS_SESSVLD, uoc_status))
#endif
		{
			/* Session valid => B_PERIPHERAL                     */

			RES_EVENT (OTG_INT_SVC, event_code);
			CHANGE_STATE (otg, OTG_STATE_B_PERIPHERAL, pEvt_mask);
#ifndef CONFIG_TEST_MODE_ONLY
			if (otg_gadget_extension.notify) {
				otg_gadget_extension.notify (
					OTG_GADGET_EVT_SVALID);
			}
#endif
		}
		break;

	case OTG_STATE_B_PERIPHERAL :

		/* B_PERIPHERAL: connected to A-host, responding             */
		/*               VBus driven by A, remote activity           */

		CHECK_STATE (otg, OTG_STATE_B_PERIPHERAL, pEvt_mask);

#ifdef CONFIG_USB_OTGMUX_IDPIN

		if (IS_BIT_RES (OTG_STS_ID, uoc_status) ||
		    IS_BIT_RES (OTG_STS_SESSVLD, uoc_status))
#else
		if (IS_BIT_RES (OTG_STS_SESSVLD, uoc_status))
#endif
		{
			/* ID pin changed | ~Session valid => B_IDLE         */

			RES_EVENT ((OTG_INT_IDC | OTG_INT_SVC), event_code);
			CHANGE_STATE (otg, OTG_STATE_B_IDLE, pEvt_mask);
#ifndef CONFIG_TEST_MODE_ONLY
			if (otg_gadget_extension.notify) {
				otg_gadget_extension.notify (
					OTG_GADGET_EVT_SVDROP);
			}
#endif
		}
		break;

	/* unlikely */ default :

		/* something went wrong */

		BUG ();
		break;
	}
	while ((otg->transceiver.state ^ otg->transceiver.prv_state) &
	       OTG_STATE_MASK);

	DBG ("OTG-state change done\n");

	return event_code;
}
#endif /* CONFIG_USB_OTGMUX_GADGET */
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

/**
 * \brief
 * OTG state change request
 *
 * \param  dev    OTG device info
 * \param  params
 *
 * \return void
 */
static inline void otg_req_state_chg (struct otg *otg, u32 params)
{
	u32 temp, tmp2, flags;

	local_irq_save (flags);

	/* disable global OTG interrupt, clear int status:                   */
	temp = ~((u32) OTG_INT_GLOBAL) & readl (&otg->regs->inten);
	writel (temp, &otg->regs->inten);
	tmp2 = readl (&otg->regs->intr);
	writel (tmp2, &otg->regs->intr);
	temp &= tmp2;

	/* update OTG state:                                                 */
	otg_change_state (otg, (params | temp), &temp);

	/* enable global OTG interrupt:                                      */
/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3950: gadget needs disconnect notification              ***/
#ifdef UDC_HSA0_WORKAROUNDS
	state_mask = ~temp & OTG_INT_ADDS;
	writel ((OTG_INT_ADDS | temp | OTG_INT_GLOBAL), &otg->regs->inten);
#else
	writel ((temp | OTG_INT_GLOBAL), &otg->regs->inten);
#endif
	local_irq_restore (flags);
}
#endif /* CONFIG_USB_OTG || CONFIG_USB_OTGMUX_GADGET */

#ifdef CONFIG_USB_OTG_TEST_INTERFACE

/**
 * \brief
 * Bind the OTG controller to the test driver
 *
 * \param  dev   otg controller info
 *
 * \return error code
 */
static int otg_test_bind (struct otg **otg)
{
	*otg = the_controller;
	if (!*otg)
		return -EINVAL;
#ifdef CONFIG_TEST_MODE_ONLY
	INFO ("bind OTG driver to the test driver, enable device\n");
	SET_FLAG (*otg, OTG_HOST_READY | OTG_GADGET_READY);
	if (IS_FLAG_SET (*otg, OTG_FLAGS_ACTIV)) {
		otg_req_state_chg (*otg, OTG_HOST_READY | OTG_GADGET_READY);
	}
	(*otg)->transceiver.gadget = (struct usb_gadget *) 1;
#else
	INFO ("bind OTG driver to the test interface\n");
#endif
	return 0;
}

/**
 * \brief
 * Unbind the OTG controller from the test driver
 *
 * \param  dev   otg controller info
 *
 * \return error code
 */
static int otg_test_unbind (struct otg **otg)
{
#ifdef CONFIG_TEST_MODE_ONLY
	INFO ("unbind OTG driver from the test driver, disable device\n");
	(*otg)->transceiver.gadget = NULL;
	RES_FLAG (*otg, OTG_HOST_READY | OTG_GADGET_READY);
	if (IS_FLAG_SET (*otg, OTG_FLAGS_ACTIV)) {
		otg_req_state_chg (*otg, OTG_HOST_READY | OTG_GADGET_READY);
	}
	*otg = NULL;
	return (otg_unbind_gadget ());
#else
	INFO ("unbind OTG driver from the test interface\n");
	*otg = NULL;
	return 0;
#endif
}
#endif /* CONFIG_USB_OTG_TEST_INTERFACE */

#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

/**
 * \brief
 * OTG set transceiver:
 *
 * \param  pointer to transceiver struct
 *
 * \return void
 */
int otg_set_transceiver (struct otg_transceiver *transceiver)
{
	struct otg *otg = the_controller;

	if (unlikely (transceiver != otg_to_transceiver (otg))) {
		ERR ("USB OTG: unknown transceiver\n");
		return -EINVAL;
	}
	else {
		return 0;
	}
}

/**
 * \brief
 * OTG get transceiver: provide info to others
 *
 * \param  void
 *
 * \return pointer to transceiver struct
 */
struct otg_transceiver * otg_get_transceiver (void)
{
	return otg_to_transceiver (the_controller);
}
#endif

#ifdef CONFIG_USB_OTG

/**
 * \brief
 * Bind/unbind the OTG controller to/from usbcore
 *
 * \param  transceiver  this transceiver
 * \param  bus          usb bus the host is assigned to
 *
 * \return error code
 */
static int otg5536_set_host (struct otg_transceiver *transceiver,
                             struct usb_bus *bus)
{
	struct otg *otg = the_controller;
	int flag, count = 0;

	if (unlikely (transceiver != otg_to_transceiver (otg))) {
		ERR ("USB OTG: unknown transceiver\n");
		return -EINVAL;
	}
	if (bus) {

		/* This bus supports OTG */

		if (transceiver->host) {
			count++;
		}
#ifdef CONFIG_USB_OTG_HIGHSPEED
		if (transceiver->companion) {
			count++;
		}
		/* Now checking consistence ...                              */
		/* Depending on the driver loading sequence is possible      */
		/* that the "Load state defaults" function was already       */
		/*  called so the state could be inconsistent.               */
		if (transceiver->host && transceiver->companion) {
			flag = (transceiver->host->is_b_host |
			        transceiver->companion->is_b_host);
			if (transceiver->gadget &&
			    !transceiver->gadget->is_a_peripheral) {
				flag |= 1;
			}
			transceiver->host->is_b_host = flag;
			transceiver->companion->is_b_host = flag;
		}
#endif
		DBG ("USB OTG: bind host: bus %d, port %d (OTG host #%d)\n",
		     bus->busnum, bus->otg_port, count);
		if (IS_FLAG_RES (otg, OTG_HOST_READY)) {
			RES_FLAG (otg, OTG_A_BUS_DROP);
			SET_FLAG (otg, OTG_HOST_READY | OTG_A_BUS_REQ);
			if (IS_FLAG_SET (otg, OTG_FLAGS_ACTIV)) {
				otg_req_state_chg (otg, OTG_HOST_READY);
			}
		}
	}
	else {
#ifdef CONFIG_USB_OTG_HIGHSPEED
		if (transceiver->companion) {
			count++;
		}
#endif
		if (transceiver->host) {
			count++;
		}
		DBG ("USB OTG: unbind host (actual OTG hosts: %d)\n", count);
		if (!count) {
			RES_FLAG (otg, OTG_HOST_READY | OTG_A_BUS_REQ);
			SET_FLAG (otg, OTG_A_BUS_DROP);
			if (IS_FLAG_SET (otg, OTG_FLAGS_ACTIV)) {
				otg_req_state_chg (otg, OTG_HOST_READY);
			}
		}
	}
	return 0;
}
#endif /* CONFIG_USB_OTG */

#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

/**
 * \brief
 * Bind/unbind the OTG controller to/from usb gadget
 *
 * \param  transceiver  this transceiver
 * \param  gadget       usb gadget info
 *
 * \return error code
 */
static int otg5536_set_peripheral (struct otg_transceiver *transceiver,
                                   struct usb_gadget *gadget)
{
	struct otg *otg = the_controller;
	int flag = 0;

	if (unlikely (transceiver != otg_to_transceiver (otg))) {
		ERR ("USB OTG: unknown transceiver\n");
		return -EINVAL;
	}
	if (gadget) {
		if (transceiver->gadget) {
			ERR ("USB gadget: OTG driver already registered\n");
			return -EBUSY;
		}
		DBG ("bind OTG driver to USB gadget\n");
		transceiver->gadget = gadget;
		SET_FLAG (otg, OTG_GADGET_READY | OTG_B_BUS_REQ);

		/* Now checking consistence ...                              */
		/* Depending on the driver loading sequence is possible      */
		/* that the "Load state defaults" function was already       */
		/*  called so the state could be inconsistent.               */
#ifdef CONFIG_USB_OTG_HIGHSPEED
		if (transceiver->companion &&
		    !transceiver->companion->is_b_host) {
			flag |= 1;
		}
#endif
		if (transceiver->host && !transceiver->host->is_b_host) {
			flag |= 1;
		}
		transceiver->gadget->is_a_peripheral = flag;

		if (IS_FLAG_SET (otg, OTG_FLAGS_ACTIV)) {
			otg_req_state_chg (otg, OTG_GADGET_READY);
		}
		if (IS_BIT_SET (OTG_STS_SESSVLD, readl (&otg->regs->sts)) &&
		    (otg->transceiver.state & OTG_STATE_MASK)
		     == OTG_STATE_B_PERIPHERAL) {
			VDBG ("calling gadget: connect\n");
			otg_gadget_extension.notify (OTG_GADGET_EVT_SVALID);
		}
		return 0;
	}
	else {
		DBG ("unbind OTG driver from USB gadget\n");
		RES_FLAG (otg, OTG_GADGET_READY | OTG_B_BUS_REQ);
		if (IS_FLAG_SET (otg, OTG_FLAGS_ACTIV)) {
			otg_req_state_chg (otg, OTG_GADGET_READY);
		}
		transceiver->gadget = NULL;
		return 0;
	}
}
#endif /* CONFIG_USB_OTGMUX || CONFIG_USB_OTGMUX_GADGET */

#ifdef CONFIG_USB_OTG

/**
 * \brief
 * Set host controllers power budget
 *
 * \param  transceiver  this transceiver
 * \param  mA           value to fill in
 *
 * \return error code
 */
static int otg5536_set_power (struct otg_transceiver *transceiver,
                              unsigned mA)
{
	struct otg *otg = the_controller;

	if (unlikely (transceiver != otg_to_transceiver (otg))) {
		ERR ("USB OTG: unknown transceiver\n");
		return -EINVAL;
	}
	/* TODO kboge: do we need to support this ?                          */

	DBG ("USB OTG: power budget change not supported\n");
	return 0;
}

/**
 * \brief
 * Start a HNP request
 *
 * \param  transceiver  this transceiver
 *
 * \return error code
 */
static int otg5536_start_hnp (struct otg_transceiver *transceiver)
{
	struct otg *otg = the_controller;

	if (unlikely (transceiver != otg_to_transceiver (otg))) {
		ERR ("USB OTG: unknown transceiver\n");
		return -ENODEV;
	}
	if (transceiver->default_a
	    && (!transceiver->host || !transceiver->host->b_hnp_enable)
#ifdef CONFIG_USB_OTG_HIGHSPEED
	    && (!transceiver->companion || !transceiver->companion->b_hnp_enable)
#endif
	   )
		return -ENOTCONN;
	if (!transceiver->default_a &&
	    (!transceiver->gadget || !transceiver->gadget->b_hnp_enable))
		return -ENOTCONN;

	switch (transceiver->state & OTG_STATE_MASK) {
	case OTG_STATE_B_HOST:
	case OTG_STATE_A_HOST:
	case OTG_STATE_A_PERIPHERAL:
		break;
	default:
		ERR ("USB OTG: wrong OTG state for role swap\n");
		return -EILSEQ;
	}
	return 0;
}

/**
 * \brief
 * Start a SRP request
 *
 * \param  transceiver  this transceiver
 *
 * \return error code
 */
static int otg5536_start_srp (struct otg_transceiver *transceiver)
{
	struct otg *otg = the_controller;

	if (unlikely (transceiver != otg_to_transceiver (otg))) {
		ERR ("USB OTG: unknown transceiver\n");
		return -EINVAL;
	}
	if ((transceiver->state & OTG_STATE_MASK) != OTG_STATE_B_IDLE) {
		RES_FLAG (otg, OTG_B_aSSN_REQ);
		ERR ("USB OTG: wrong OTG state for SRP\n");
		return -EBUSY;
	}
	SET_FLAG (otg, OTG_B_aSSN_REQ);
	otg_req_state_chg (otg, OTG_B_aSSN_REQ);
	return 0;
}

/**
 * \brief
 * Start a USB host enumeration
 *
 * \param  transceiver  this transceiver
 *
 * \return error code
 */
int otg_hnp_start_enum (struct otg_transceiver *transceiver)
{
	struct otg *otg = the_controller;
	struct usb_bus *bus;
	int status = 0;

	if (unlikely (transceiver != otg_to_transceiver (otg))) {
		ERR ("USB OTG: unknown transceiver\n");
		return -EINVAL;
	}
	if (otg->transceiver.state == OTG_STATE_B_HOST) {

		bus = otg->transceiver.host;

#ifdef CONFIG_USB_OTG_HIGHSPEED

		if (!bus || !bus->is_b_host) {
			bus = otg->transceiver.companion;
		}
#endif
		if (bus && bus->is_b_host) {
			DBG ("USB OTG: in HNP, start enumeration ...\n");
			status = usb_bus_start_enum (bus, bus->otg_port);
		}
		else {
			DBG ("USB OTG: no device to enumerate\n");
		}
	}
	else {
		DBG ("USB OTG: in HNP: wrong state %02x\n",
		      otg->transceiver.state);
	}
	return status;
}

/**
 * \brief
 * OTG application request
 *
 * \param  params
 *
 * \return OTG_APP_REQ_ACK
 */
static int otg_app_request (u32 params)
{
	struct otg *otg = the_controller;
	u32  flags = params;
	struct usb_bus *bus;

	if (OTG_UDC_RWK_REQ & flags) {

		if (otg_gadget_extension.notify) {
			otg_gadget_extension.notify (OTG_GADGET_REQ_WAKE);
		}
		flags &= ~((u32) OTG_UDC_RWK_REQ);
	}

	if (OTG_AB_HNP_REQ & flags) {

#if    LINUX_VERSION_CODE < KERNEL_VERSION(2,5,59)

		INFO ("Force HNP not implemented in this kernel version\n");
#else
		/* request HNP */

		bus = otg->transceiver.host;
		if (!bus->b_hnp_enable) {

#ifdef CONFIG_USB_OTG_HIGHSPEED
			bus = otg->transceiver.companion;
		}
		if (!bus->b_hnp_enable) {
#endif
			INFO ("HNP not enabled, ignoring request\n");
		}
		else {
			if (!otg5536_start_hnp (&otg->transceiver) &&
			    bus->root_hub->children [bus->otg_port - 1]) {
				usb_suspend_device (bus->root_hub->children [bus->otg_port - 1],
				                    PM_SUSPEND_MEM);
			}
			else {
				ERR ("no device found for HNP\n");
			}
		}
#endif
		flags &= ~((u32) OTG_AB_HNP_REQ);
	}

	if (OTG_A_CLR_ERR & flags) {

		/* request to clear error flags                              */

		flags |= OTG_A_VBUS_FAILED;
	}

	/* create mask of flags to clear                                     */

	flags &= (OTG_B_SRP_ERROR | OTG_A_VBUS_FAILED);

	if (flags) {

		/* reset error flags if requested                            */

		RES_FLAG (otg, (flags));
	}

	/* prepare for pending requests (no more actions on OTG_A_bHNP_EN)   */

	flags = params & SW_REQUEST_MASK;
	SET_FLAG (otg, (flags));

	if (OTG_A_BUS_DROP & flags) {

		/* request to drop A-VBus: reset "rise" request if pending   */

		flags &= ~OTG_A_BUS_REQ;
		RES_FLAG (otg, OTG_A_BUS_REQ);
	}
	if (OTG_A_BUS_REQ & flags) {

		/* request to rise A-VBus: reset "drop" request if pending   */

		RES_FLAG (otg, OTG_A_BUS_DROP);
	}
	if (OTG_B_BUS_DIS & flags) {

		/* request to disable B-dev: reset "ena" request if pending  */

		flags &= ~OTG_B_BUS_REQ;
		RES_FLAG (otg, OTG_B_BUS_REQ);
	}
	if (OTG_B_BUS_REQ & flags) {

		/* request to enable B-dev: reset "disa" request if pending  */

		RES_FLAG (otg, OTG_B_BUS_DIS);
	}
	if (flags) {

		/* request still pending may force a state change            */

		otg_req_state_chg (otg, flags);
	}
	return OTG_APP_REQ_ACK;
}
#endif /* CONFIG_USB_OTG */

#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

/**
 * \brief
 * OTG application query
 *
 * \param  index   select status info data
 *
 * \return status
 */
static u32 otg_app_query (int index)
{
	struct otg *otg = the_controller;
	u32  temp = 0;

	if (index == 0) {
		temp = otg->transceiver.params |
		       readl (&otg->regs->sts);

/*** HS-A0 WA: BUG-3943: gadget phy suspend issue                          ***/
#ifdef UDC_HSA0_WORKAROUNDS
		if (((readl (&otg->regs->ctl) & OTG_CTL_MUX_MASK) ==
		     OTG_CTL_ENABLE_UDC) &&
		    ((temp & OTG_STS_PSUS) || (~temp & OTG_STS_VBUSVLD))) {
			temp |= OTG_FLAGS_UDC_SUSP;
		}
#endif
	}
	else if (index == 1) {
		temp = otg->transceiver.state;
	}
	return (temp);
}

/**
 * \brief
 * OTG ISR calling the main state machine
 *
 * \param  irq    IRQ number
 * \param  _dev  
 * \param  r     
 *
 * \return IRQ_HANDLED (system code)
 */
static irqreturn_t otg_isr (int irq, void *dev, struct pt_regs *r)
{
	struct otg *otg = (struct otg *) dev;
	u32         interrupts, int_mask, temp;

	int_mask = readl (&otg->regs->inten);
	if ((OTG_INT_GLOBAL & int_mask) &&
	    (int_mask &= ~((u32) OTG_INT_GLOBAL)) &&
	    (interrupts = int_mask &
		          (temp = readl (&otg->regs->intr)))) {

		writel (int_mask, &otg->regs->inten);
		/* clear interrupt status */
		writel (temp, &otg->regs->intr);
/*
#ifdef CONFIG_USB_OTG

		if (GOT_EVENT (OTG_INT_PCC, interrupts) &&
		    IS_BIT_SET (OTG_STS_PCON, readl (&otg->regs->sts))) {
			otg_hnp_start_enum (&otg->transceiver);
		}
#endif */
		/* filter out additional WA interrupts, they're done         */
		/*       don't want to see them in the state machine         */

		if (interrupts & ~state_mask) {

			/* events pending for the state machine              */

			otg_change_state (otg, (interrupts & ~state_mask),
			                  &int_mask);
		}

		/* enable interrupts and keep information about WA ints:     */

		state_mask = OTG_INT_ADDS & ~int_mask;
		writel ((OTG_INT_ADDS | int_mask | OTG_INT_GLOBAL),
		        &otg->regs->inten);
	}
	return IRQ_HANDLED;
}
#endif /* CONFIG_USB_OTG || CONFIG_USB_OTGMUX_GADGET */

/**
 * \brief
 * OTG probe: init hardware, register the driver
 *
 * \param  otg   otg controller info
 *
 * \return  success
 */
static inline int __init otg_probe (struct otg *otg)
{
	int         retval;
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)
	u32         temp;
	int         i;
#endif

	/* initialize the OTG controller */

	VDBG ("OTG init ...\n");

#ifdef VERBOSE
	/* print regs */
	print_regs (otg);
#endif
	/* Make sure we'll remember the initial state                        */
	init_state = readl (&otg->regs->ctl);
	VDBG ("  OTG init state was %08x\n", init_state);

#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	/* turn on the OTG controller                                        */

	writel ((init_state | OTG_CTL_PADEN), &otg->regs->ctl);

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
#ifndef UDC_HSA0_WORKAROUNDS

	/* Enable OTG hardware (incl. auto pull-up enable)                   */

	writel ((readl (&dev->regs->cap) | OTG_CAP_APU), &otg->regs->cap);
#endif

	/* initialize flags                                                  */
	otg->transceiver.params = 0;
#endif

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget disconnect notification                    ***/
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	/* make sure all interrupts are disabled                             */

	writel (OTG_INT_DISALL, &otg->regs->inten);
	writel (OTG_INT_ENALL, &otg->regs->intr);

	/* Set multiplexer to neutral, get power control, drop VBus          */

	if (((init_state & OTG_CTL_MUX_MASK) == OTG_CTL_ENABLE_UHC) &&
	    ((((init_state & OTG_CTL_PPO) && (init_state & OTG_CTL_PPWR))) ||
	     ((~init_state & OTG_CTL_PPO) &&
	      (readl (&otg->regs->sts) & OTG_STS_SESSVLD)))) {

		/* VBus still powered try to discharge VBus and set timer    */

		DBG ("Setting init state, trying to discharge VBus ...\n");

		for (i = 0; i < 4; i++) {
			writel (TIMER_PERIOD, &otg->regs->tmr);
			writel ((OTG_CTL_PADEN | OTG_CTL_IDSNSEN |
			         OTG_CTL_PPO | OTG_CTL_DISCHRG |
			         OTG_CTL_TMR_UNCOND), &otg->regs->ctl);
			do { }
			while (!(readl (&otg->regs->sts) & OTG_STS_TMH));
		}
		writel ((OTG_CTL_PADEN | OTG_CTL_IDSNSEN | OTG_CTL_PPO),
		        &otg->regs->ctl);
		writel (OTG_INT_ENALL, &otg->regs->intr);
#ifdef DEBUG
		if (readl (&otg->regs->sts) & OTG_STS_SESSVLD) {

			DBG ("  VBus still high, external host connected\n");
		}
		else {

			DBG ("  VBus discharged\n");
		}
#endif
	}
	else {
		DBG ("Setting init state\n");

		writel ((OTG_CTL_PADEN | OTG_CTL_IDSNSEN | OTG_CTL_PPO),
		        &otg->regs->ctl);
	}
#endif /* CONFIG_USB_OTG || CONFIG_USB_OTGMUX_GADGET */

#ifdef CONFIG_USB_OTGMUX_HOST

	/* Set multiplexer to host                                           */
	writel ((readl (&otg->regs->cap) & ~((u32) OTG_CAP_APU)),
	        &otg->regs->cap);

#ifdef CONFIG_SOC_AU1200

	writel ((OTG_CTL_ENABLE_UHC | OTG_CTL_PPO | OTG_CTL_PPWR),
	        &otg->regs->ctl);

	/* Turn on VBus                                                      */
	AU1000GPIO_CLR_BIT (USB_VBUS_GPIO);
#else
	writel (OTG_CTL_ENABLE_UHC, &otg->regs->ctl);
#endif

	INFO ("OTG functionality is disabled, port is assigned to host\n");
#endif
	VDBG ("OTG init done\n");

#ifdef CONFIG_USB_OTG_TEST_INTERFACE

	/* registering to the test driver */
	usb_otg_test_register_driver (&otg_test_extension);
#endif
#ifndef CONFIG_TEST_MODE_ONLY
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	/* registering to the device driver */
	if (usb_gadget_register_otg (otg_get_transceiver)) {
		ERR ("gadget driver registration failed\n");
		retval = -ENODEV;
		goto err1;
	}
#endif
#ifdef CONFIG_USB_OTG

	/* registering to the usbcore */
	if (usb_host_register_otg (otg_get_transceiver)) {
		ERR ("usbcore registration failed\n");
		retval = -ENODEV;
		goto err2;
	}
#endif
#endif
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	/* finally activate OTG functionality */
	/* Enable timer interrupt, start timer, set state                    */

	SET_OTG_TIMER (otg, IDSNS_WAIT);
	CHANGE_STATE (otg, OTG_STATE_UNDEFINED, &temp);
	CHECK_STATE (otg, OTG_STATE_UNDEFINED, &temp);

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget needs disconnect notification              ***/

	/* clear all interrupts before enable */
	writel (readl (&otg->regs->intr), &otg->regs->intr);

	state_mask = ~temp & OTG_INT_ADDS;
	writel ((OTG_INT_ADDS | temp | OTG_INT_GLOBAL), &otg->regs->inten);

	DBG ("OTG-HW initialized, now checking ID ...\n");
#endif

	return 0;

#ifndef CONFIG_TEST_MODE_ONLY
#ifdef CONFIG_USB_OTG

err2:

#endif
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	usb_gadget_unregister_otg ();
err1:

#endif
#endif
#ifdef CONFIG_USB_OTG_TEST_INTERFACE

	usb_otg_test_unregister_driver ();
#endif
	return retval;
}

/**
 * \brief
 * OTG remove: deregister the driver, clean-up hardware
 *
 * \param  otg   otg controller info
 *
 * \return void
 */
static inline void __exit otg_remove (struct otg *otg)
{
	int muxer;

#ifndef CONFIG_TEST_MODE_ONLY
#ifdef CONFIG_USB_OTG

	/* unregistering from the usb core */
	usb_host_deregister_otg ();
#endif
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	/* unregistering from the usb gadget */
	usb_gadget_unregister_otg ();

#endif
#endif
#ifdef CONFIG_USB_OTG_TEST_INTERFACE

	/* unregistering from the test driver */
	usb_otg_test_unregister_driver ();
#endif

	/* clean up the OTG controller */

#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	/* Disable all interrupts                                            */
	writel (OTG_INT_DISALL, &otg->regs->inten);
	writel (OTG_INT_ENALL, &otg->regs->intr);

	/* reset state, terminate all connections                            */
	CHANGE_STATE (otg, OTG_STATE_UNDEFINED, &state_mask);
	CHECK_STATE (otg, OTG_STATE_UNDEFINED, &state_mask);
	otg->transceiver.params = 0;
#endif

	muxer = init_state & (OTG_CTL_ENABLE_UHC | OTG_CTL_ENABLE_UDC);

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
#ifdef UDC_HSA0_WORKAROUNDS

	/* Don't assign the port to the device controller                    */

	if (!(muxer ^ OTG_CTL_ENABLE_UDC)) {

		init_state &= ~((u32) (OTG_CTL_MUX_MASK | OTG_CTL_PUEN));
		muxer = OTG_CTL_DISABLE_ALL;
	}
	VDBG ("OTG writing back corrected init state: %08x\n", init_state);
#else
	/* Reset APU                                                         */

	writel ((readl (&otg->regs->cap) & ~((u32) OTG_CAP_APU)),
	        &otg->regs->cap);

	VDBG ("OTG writing back init state: %08x\n", init_state);
#endif
	/* Now, that's the moment to remember                                */
	/* Set dev muxer and pull up bits, turn off the OTG controller       */

#ifdef CONFIG_SOC_AU1200

	/* Turn off VBus                                                     */
	AU1000GPIO_SET_BIT (USB_VBUS_GPIO);
#endif
	writel (init_state, &otg->regs->ctl);

	if (!(muxer ^ OTG_CTL_ENABLE_UHC)) {
		INFO ("disabling OTG-HW, port is assigned to host\n");
	}
	else if (!(muxer ^ OTG_CTL_ENABLE_UDC)) {
		INFO ("disabling OTG-HW, port is assigned to device\n");
	}
	else {
		INFO ("disabling OTG-HW, port is not assigned\n");
	}

	VDBG ("OTG exit: OTG-HW disabled\n");

	if (!muxer) {
		INFO ("OTG HW disabled, port is not assigned\n");
	}
}

#ifdef CONFIG_SOC_AU1200
#ifdef USE_AU1200_PCI_DUMMY

/**
 * \brief
 * OTG dev probe: enable, init controller hardware
 *
 * \param pdev   dummy pci device pointer
 *
 * \return  success
 */
static int __init otg_pci_probe (struct pci_dev *pdev)
{
	struct otg *otg;
	void       *base;
	int         retval;
	char        buf [8] = {0,0,0,0,0,0,0,0}, *bufp;

	/* alloc, and start init */
	otg = (struct otg *) kmalloc (sizeof(struct otg), SLAB_KERNEL);
	if (!otg) {
		ERR ("couldn't allocate memory for OTG driver\n");
		retval = -ENOMEM;
		goto err1;
	}
	DBG ( "kmalloc: OTG driver: %p\n", otg);

	/* hold global device pointer */
	the_controller = otg;

	memset (otg, 0, sizeof(struct otg));
	spin_lock_init (&otg->lock);

	/*
	* Fill in the dummy pci_dev. Only those members actually
	* dereferenced in this driver are initialized.
	*/
	memset(pdev, 0, sizeof(struct pci_dev));
#if    LINUX_VERSION_CODE < KERNEL_VERSION(2,5,59)
	strcpy(pdev->name, "non-PCI UOC");
#endif
	pdev->resource[0].name = "UOC Registers";
	pdev->resource[0].start = USB_UOC_BASE;
	pdev->resource[0].end = USB_UOC_BASE + USB_UOC_LEN - 1;
	pdev->resource[0].flags = 0;
	pdev->irq = AU1200_USB_INT;

	otg->pdev = pdev;

	au_writel ((au_readl (USB_MSR_BASE + USB_MSR_MCFG) |
	            (1 << USBMSRMCFG_GMEMEN)), (USB_MSR_BASE + USB_MSR_MCFG));
	au_readl (USB_MSR_BASE + USB_MSR_MCFG);
	/* au_sync (); */
	udelay (1000);
	if (AU1000GPIO_INIT_BIT (USB_VBUS_GPIO))
		WARN ("Couldn't initialize power switch GPIO\n");
	otg->enabled = 1;

	if (!request_mem_region (pdev->resource[0].start,
	                         pdev->resource[0].end + 1
	                         - pdev->resource[0].start,
	                         driver_name)) {
		ERR ("controller already in use\n");
		retval = -EBUSY;
		goto err3;
	}
	otg->region = 1;

	base = ioremap_nocache (pdev->resource[0].start,
	                        pdev->resource[0].end + 1
	                        - pdev->resource[0].start);
	if (!base) {
		ERR ("couldn't map memory\n");
		retval = -EFAULT;
		goto err4;
	}
	otg->regs = (struct otg_regs *) base;
	bufp = buf;

	otg->chiprev = (u16) read_c0_prid() & 0xff;

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget disconnect notification                    ***/
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	/* OTG transceiver info */
	otg->transceiver.dev = &pdev->dev;
	otg_init_transceiver (otg_to_transceiver (otg));

	/* make sure all interrupts are disabled */
	writel (OTG_INT_DISALL, &otg->regs->inten);
	writel (OTG_INT_ENALL, &otg->regs->intr);
	readl (&otg->regs->inten);

	/* irq setup after old hardware is cleaned up */
	if (!pdev->irq) {
		ERR ("No IRQ. Check system setup!\n");
		retval = -ENODEV;
		goto err5;
	}
	snprintf (buf, sizeof buf, "%d", pdev->irq);
	bufp = buf;
	if (request_irq (pdev->irq, otg_isr, SA_INTERRUPT | SA_SHIRQ,
	                 driver_name, otg) != 0) {
		ERR ("request interrupt %s failed\n", bufp);
		retval = -EBUSY;
		goto err5;
	}
	otg->got_irq = 1;

	/* done */
	INFO ("%s\n", driver_desc);
	INFO ("irq %s, mem %08lx, chip rev %02x (Au1200 %s)\n",
	      bufp, pdev->resource[0].start, otg->chiprev,
	      (otg->chiprev ? "AC" : "AB"));

#else /* CONFIG_USB_OTG || CONFIG_USB_OTGMUX_GADGET */

	INFO ("%s\n", driver_desc);
	INFO ("mem %08lx, chip rev %02x (Au1200 %s)\n",
	      pdev->resource[0].start, otg->chiprev,
	      (otg->chiprev ? "AC" : "AB"));
#endif
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_IDPIN) || defined(CONFIG_USB_OTGMUX_HOST)

	INFO ("using GPIO_%d for VBus control\n", USB_VBUS_GPIO);
#endif
	bufp = DRIVER_VERSION;
	INFO ("driver version: %s\n", bufp);

	if ((retval = otg_probe (otg)) == 0) {
		pci_set_drvdata (pdev, otg);
		return 0;
	}

	/* something went wrong */

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget disconnect notification                    ***/
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	otg->got_irq = 0;
	free_irq (pdev->irq, otg);
err5:

#endif
	otg->regs = NULL;
	iounmap (base);
err4:
	otg->region = 0;
	release_mem_region (pdev->resource[0].start,
	                    pdev->resource[0].end + 1
	                    - pdev->resource[0].start);
err3:
	otg->enabled = 0;
	AU1000GPIO_TERM_BIT (USB_VBUS_GPIO);
	au_writel ((au_readl (USB_MSR_BASE + USB_MSR_MCFG) &
	            ~((u32) (1 << USBMSRMCFG_GMEMEN))),
	           (USB_MSR_BASE + USB_MSR_MCFG));
	au_readl (USB_MSR_BASE + USB_MSR_MCFG);
	/* au_sync (); */
	udelay (1000);

	otg->pdev = NULL;
	the_controller = NULL;
	DBG ( "kfree: OTG driver: %p\n", otg);
	kfree (otg);
err1:
	otg = NULL;
	return retval;
}

/**
 * \brief
 * OTG dev remove: clean-up, disable controller hardware
 *
 * \param  pdev   dummy pci device info
 *
 * \return void
 */
static void __exit otg_pci_remove (struct pci_dev *pdev)
{
	struct otg *otg = pci_get_drvdata (pdev);

	otg_remove (otg);

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget disconnect notification                    ***/
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	otg->got_irq = 0;
	free_irq (pdev->irq, otg);
#endif
	iounmap (otg->regs);
	otg->regs = NULL;
	otg->region = 0;
	release_mem_region (pdev->resource[0].start,
	                    pdev->resource[0].end + 1
	                    - pdev->resource[0].start);
	otg->enabled = 0;
	AU1000GPIO_TERM_BIT (USB_VBUS_GPIO);
/*	au_writel ((au_readl (USB_MSR_BASE + USB_MSR_MCFG) &
	            ~((u32) (1 << USBMSRMCFG_GMEMEN))),
	           (USB_MSR_BASE + USB_MSR_MCFG)); */
	au_readl (USB_MSR_BASE + USB_MSR_MCFG);
	/* au_sync (); */
	udelay (1000);
	otg->pdev = NULL;
	the_controller = NULL;
	DBG ( "kfree: OTG driver: %p\n", otg);
	kfree (otg);
	otg = NULL;
	pci_set_drvdata(pdev, NULL);
}
#else

/**
 * \brief
 * OTG dev probe: enable, init controller hardware
 *
 * \param  dev   platform device info
 *
 * \return success
 */
static int __init otg_drv_probe (struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct otg *otg;
	u32         resource, len, irq;
	void       *base;
	int         retval;
	char        buf [8] = {0,0,0,0,0,0,0,0}, *bufp;

	/* alloc, and start init */
	otg = (struct otg *) kmalloc (sizeof(struct otg), SLAB_KERNEL);
	if (!otg) {
		ERR ("couldn't allocate memory for OTG driver\n");
		retval = -ENOMEM;
		goto err1;
	}
	DBG ( "kmalloc: OTG driver: %p\n", otg);

	/* hold global device pointer */
	the_controller = otg;

	memset (otg, 0, sizeof(struct otg));
	spin_lock_init (&otg->lock);

	if (pdev->resource[0].flags != IORESOURCE_MEM) {
		ERR ("resource is not IORESOURCE_MEM\n");
		retval = -ENOMEM;
		goto err2;
	}
	resource = pdev->resource[0].start;
	len = pdev->resource[0].end + 1 - pdev->resource[0].start;
	if (pdev->resource[1].flags != IORESOURCE_IRQ) {
		ERR ("resource is not IORESOURCE_IRQ\n");
		retval = -ENOMEM;
		goto err2;
	}
	irq = pdev->resource[1].start;

	otg->pdev = pdev;

	au_writel ((au_readl (USB_MSR_BASE + USB_MSR_MCFG) |
	            (1 << USBMSRMCFG_GMEMEN)), (USB_MSR_BASE + USB_MSR_MCFG));
	au_readl (USB_MSR_BASE + USB_MSR_MCFG);
	/* au_sync (); */
	udelay (1000);
	if (AU1000GPIO_INIT_BIT (USB_VBUS_GPIO))
		WARN ("Couldn't initialize power switch GPIO\n");
	otg->enabled = 1;

	if (!request_mem_region (resource, len, driver_name)) {
		ERR ("controller already in use\n");
		retval = -EBUSY;
		goto err3;
	}
	otg->region = 1;

	base = ioremap_nocache (resource, len);
	if (!base) {
		ERR ("couldn't map memory\n");
		retval = -EFAULT;
		goto err4;
	}
	otg->regs = (struct otg_regs *) base;
	bufp = buf;

	otg->chiprev = (u16) read_c0_prid() & 0xff;

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget disconnect notification                    ***/
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	/* OTG transceiver info */
	otg->transceiver.dev = dev;
	otg_init_transceiver (otg_to_transceiver (otg));

	/* make sure all interrupts are disabled */
	writel (OTG_INT_DISALL, &otg->regs->inten);
	writel (OTG_INT_ENALL, &otg->regs->intr);
	readl (&otg->regs->inten);

	/* irq setup after old hardware is cleaned up */
	if (!irq) {
		ERR ("No IRQ. Check system setup!\n");
		retval = -ENODEV;
		goto err5;
	}
	snprintf (buf, sizeof buf, "%d", irq);
	bufp = buf;
	if (request_irq (irq, otg_isr, SA_INTERRUPT | SA_SHIRQ,
	                 driver_name, otg) != 0) {
		ERR ("request interrupt %s failed\n", bufp);
		retval = -EBUSY;
		goto err5;
	}
	otg->got_irq = 1;

	/* done */
	INFO ("%s\n", driver_desc);
	INFO ("irq %s, mem %08x, chip rev %02x (Au1200 %s)\n",
	      bufp, resource, otg->chiprev, (otg->chiprev ? "AC" : "AB"));

#else /* CONFIG_USB_OTG || CONFIG_USB_OTGMUX_GADGET */

	INFO ("%s\n", driver_desc);
	INFO ("mem %08x, chip rev %02x (Au1200 %s)\n",
	      resource, otg->chiprev, (otg->chiprev ? "AC" : "AB"));
#endif
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_IDPIN) || defined(CONFIG_USB_OTGMUX_HOST)

	INFO ("using GPIO_%d for VBus control\n", USB_VBUS_GPIO);
#endif
	bufp = DRIVER_VERSION;
	INFO ("driver version: %s\n", bufp);

	if ((retval = otg_probe (otg)) == 0) {
		dev_set_drvdata (dev, otg);
		return 0;
	}

	/* something went wrong */

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget disconnect notification                    ***/
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	otg->got_irq = 0;
	free_irq (irq, otg);
err5:

#endif
	otg->regs = NULL;
	iounmap (base);
err4:
	otg->region = 0;
	release_mem_region (resource, len);
err3:
	otg->enabled = 0;
	AU1000GPIO_TERM_BIT (USB_VBUS_GPIO);
	au_writel ((au_readl (USB_MSR_BASE + USB_MSR_MCFG) &
	            ~((u32) (1 << USBMSRMCFG_GMEMEN))),
	           (USB_MSR_BASE + USB_MSR_MCFG));
	au_readl (USB_MSR_BASE + USB_MSR_MCFG);
	/* au_sync (); */
	udelay (1000);
err2:
	otg->pdev = NULL;
	the_controller = NULL;
	DBG ( "kfree: OTG driver: %p\n", otg);
	kfree (otg);
err1:
	otg = NULL;

	return retval;
}

/**
 * \brief
 * OTG dev remove: clean-up, disable controller hardware
 *
 * \param  dev   platform device info
 *
 * \return void
 */
static int __exit otg_drv_remove (struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct otg *otg = dev_get_drvdata (dev);

	otg_remove (otg);

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget disconnect notification                    ***/
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	otg->got_irq = 0;
	free_irq (pdev->resource[1].start, otg);
#endif
	iounmap (otg->regs);
	otg->regs = NULL;
	otg->region = 0;
	release_mem_region (pdev->resource[0].start,
	                    pdev->resource[0].end + 1
	                    - pdev->resource[0].start);
	otg->enabled = 0;
	AU1000GPIO_TERM_BIT (USB_VBUS_GPIO);
/*	au_writel ((au_readl (USB_MSR_BASE + USB_MSR_MCFG) &
	            ~((u32) (1 << USBMSRMCFG_GMEMEN))),
	           (USB_MSR_BASE + USB_MSR_MCFG)); */
	au_readl (USB_MSR_BASE + USB_MSR_MCFG);
	/* au_sync (); */
	udelay (1000);
	otg->pdev = NULL;
	the_controller = NULL;
	DBG ( "kfree: OTG driver: %p\n", otg);
	kfree (otg);
	otg = NULL;
	dev_set_drvdata(dev, NULL);
	return 0;
}
#endif
#else
#ifdef CONFIG_PCI

/**
 * \brief
 * OTG PCI probe: enable, init controller hardware
 *
 * \param  pdev   pci device info
 * \param  id     pci ids
 *
 * \return  success
 */
static int __init otg_pci_probe (struct pci_dev *pdev,
                                 const struct pci_device_id *id)
{
	struct otg *otg;
	u32         resource, len;
	void       *base;
	int         retval;
	char        buf [8] = {0,0,0,0,0,0,0,0}, *bufp;

	/* alloc, and start init */
	otg = (struct otg *) kmalloc (sizeof(struct otg), SLAB_KERNEL);
	if (!otg) {
		ERR ("couldn't allocate memory for OTG driver\n");
		retval = -ENOMEM;
		goto err1;
	}
	DBG ( "kmalloc: OTG driver: %p\n", otg);

	/* hold global device pointer */
	the_controller = otg;

	memset (otg, 0, sizeof(struct otg));
	spin_lock_init (&otg->lock);

	otg->pdev = pdev;

	/* now all the pci goodies ...                                       */
	if (pci_enable_device (pdev) < 0) {
		ERR ("couldn't enable PCI device\n");
		retval = -ENODEV;
		goto err2;
	}
	otg->enabled = 1;

	resource = pci_resource_start (pdev, 0);
	len = pci_resource_len (pdev, 0);

	if (!request_mem_region (resource, len, driver_name)) {
		ERR ("controller already in use\n");
		retval = -EBUSY;
		goto err3;
	}
	otg->region = 1;

	base = ioremap_nocache (resource, len);
	if (!base) {
		ERR ("couldn't map memory\n");
		retval = -EFAULT;
		goto err4;
	}
	otg->regs = (struct otg_regs *) base;
	bufp = buf;

	pci_read_config_byte (pdev, PCI_REVISION_ID, (u8*) &otg->chiprev);

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget disconnect notification                    ***/
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	/* OTG transceiver info */

#if    LINUX_VERSION_CODE < KERNEL_VERSION(2,5,59)
	otg->transceiver.dev = NULL;
#else
	otg->transceiver.dev = &pdev->dev;
#endif
	otg_init_transceiver (otg_to_transceiver (otg));

	/* make sure all interrupts are disabled */
	writel (OTG_INT_DISALL, &otg->regs->inten);
	writel (OTG_INT_ENALL, &otg->regs->intr);
	readl (&otg->regs->inten);

	/* irq setup after old hardware is cleaned up */
	if (!pdev->irq) {
		ERR ("No IRQ. Check PCI setup!\n");
		retval = -ENODEV;
		goto err5;
	}
#ifndef __sparc__
	snprintf (buf, sizeof buf, "%d", pdev->irq);
	bufp = buf;
#else
	bufp = __irq_itoa(pdev->irq);
#endif
	if (request_irq (pdev->irq, otg_isr, SA_INTERRUPT | SA_SHIRQ,
	                 driver_name, otg) != 0) {
		ERR ("request interrupt %s failed\n", bufp);
		retval = -EBUSY;
		goto err5;
	}
	otg->got_irq = 1;

	/* done */
	INFO ("%s\n", driver_desc);
	INFO ("irq %s, pci mem %08x, chip rev %02x (Geode5536 %s)\n",
	      bufp, resource, otg->chiprev, (otg->chiprev - 1 ? "B0" : "A0"));

#else /* CONFIG_USB_OTG || CONFIG_USB_OTGMUX_GADGET */

	INFO ("%s\n", driver_desc);
	INFO ("pci mem %08x, chip rev %02x (Geode5536 %s)\n",
	      resource, otg->chiprev, (otg->chiprev - 1 ? "B0" : "A0"));
#endif
	bufp = DRIVER_VERSION;
	INFO ("driver version: %s\n", bufp);

	if ((retval = otg_probe (otg)) == 0) {
		pci_set_drvdata (pdev, otg);
		return 0;
	}

	/* something went wrong */

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget disconnect notification                    ***/
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	otg->got_irq = 0;
	free_irq (pdev->irq, otg);
err5:

#endif
	otg->regs = NULL;
	iounmap (base);
err4:
	otg->region = 0;
	release_mem_region (resource, len);
err3:
	otg->enabled = 0;
	pci_disable_device (pdev);
err2:
	otg->pdev = NULL;
	the_controller = NULL;
	DBG ( "kfree: OTG driver: %p\n", otg);
	kfree (otg);
err1:
	otg = NULL;
	return retval;
}

/**
 * \brief
 * OTG PCI remove: clean-up, disable controller hardware
 *
 * \param  pdev   pci device info
 *
 * \return void
 */
static void __exit otg_pci_remove (struct pci_dev *pdev)
{
	struct otg *otg = pci_get_drvdata (pdev);

	otg_remove (otg);

/*** HS-A0 WA: BUG-3885: VB_SESS_VLD value too high                        ***/
/*** HS-A0 WA: BUG-3943: gadget suspend issue                              ***/
/*** HS-A0 WA: BUG-3950: gadget disconnect notification                    ***/
#if defined(CONFIG_USB_OTG) || defined(CONFIG_USB_OTGMUX_GADGET)

	otg->got_irq = 0;
	free_irq (pdev->irq, otg);
#endif
	iounmap (otg->regs);
	otg->regs = NULL;
	otg->region = 0;
	release_mem_region (pci_resource_start(pdev, 0),
	                    pci_resource_len(pdev, 0));
	otg->enabled = 0;
	pci_disable_device (pdev);
	otg->pdev = NULL;
	the_controller = NULL;
	DBG ( "kfree: OTG driver: %p\n", otg);
	kfree (otg);
	otg = NULL;
	pci_set_drvdata(pdev, NULL);
}
#endif
#endif


/*****************************************************************************
 *  More data
 *****************************************************************************/

#ifdef CONFIG_SOC_AU1200
#ifndef USE_AU1200_PCI_DUMMY
/**
 * \brief
 * driver struct to be used for driver registration
 *
 */
static struct device_driver otg_device_driver = {
	.name =		"au1xxx-uoc",
	.bus =		&platform_bus_type,
	.probe =	otg_drv_probe,
	.remove =	otg_drv_remove,
/* 	.suspend =	otg_drv_suspend, */
/* 	.resume =	otg_drv_resume, */
};
#endif
#else
#ifdef CONFIG_PCI
/**
 * \brief
 * PCI ID table
 *
 */
static struct pci_device_id pci_ids [] = { {
	.vendor =	PCI_VENDOR_ID_AMD,
	.device =	PCI_DEVICE_ID_AMD_CS5536_UOC,
	.subvendor =	PCI_ANY_ID,
	.subdevice =	PCI_ANY_ID,
	.class = 	0,
	.class_mask = 	0,

}, { /* end: all zeroes */ }
};

/**
 * \brief
 * PCI driver struct to be used for driver registration
 * ( this is a "new style" PCI driver module )
 *
 */
static struct pci_driver otg_pci_driver = {
	.name =		(char *) driver_name,
	.id_table =	pci_ids,
	.probe =	otg_pci_probe,
	.remove =	otg_pci_remove,
/* 	.suspend =	otg_pci_suspend, */
/* 	.resume =	otg_pci_resume, */
};
MODULE_DEVICE_TABLE (pci, pci_ids);
#endif
#endif

/* This comment closes the module definition from above. There can be multiple 
   definitions of this kind in a file. See the doxygen documentation for more
   information. */
/** \}*/

MODULE_DESCRIPTION (DRIVER_DESC);
MODULE_AUTHOR ("Karsten Boge");
MODULE_LICENSE ("GPL");

#ifdef CONFIG_SOC_AU1200
#ifdef USE_AU1200_PCI_DUMMY

static int __init init (void)
{
	return (otg_pci_probe (&pdev_dummy));
}
static void __exit cleanup (void)
{
	otg_pci_remove (&pdev_dummy);
}
#else

static int __init init (void)
{
        return (driver_register (&otg_device_driver));
}
static void __exit cleanup (void)
{
	driver_unregister (&otg_device_driver);
}
#endif
#else
#ifdef CONFIG_PCI

static int __init init (void)
{
        return (pci_module_init (&otg_pci_driver));
}
static void __exit cleanup (void)
{
	pci_unregister_driver (&otg_pci_driver);
}
#endif
#endif

module_init (init);
module_exit (cleanup);
