/*
 * EHCI HCD (Host Controller Driver) for USB.
 *
 * (C) Copyright 2000-2004 David Brownell <dbrownell@users.sourceforge.net>
 *
 * Bus Glue for AMD Alchemy Au1xxx
 *
 * Based on "ohci-au1xxx.c" by Matt Porter <mporter@kernel.crashing.org>
 *
 * Modified for AMD Alchemy Au1200 EHC
 *  by K.Boge <karsten.boge@amd.com>
 *
 * This file is licenced under the GPL.
 */

#include <asm/mach-au1x00/au1000.h>

#ifndef CONFIG_SOC_AU1200
#error "Alchemy chip doesn't have EHC"
#else   /* Au1200 */

#define USB_HOST_CONFIG   (USB_MSR_BASE + USB_MSR_MCFG)
#define USB_MCFG_PFEN     (1<<31)
#define USB_MCFG_RDCOMB   (1<<30)
#define USB_MCFG_SSDEN    (1<<23)
#define USB_MCFG_PHYPLLEN (1<<19)
#define USB_MCFG_EHCCLKEN (1<<17)
#define USB_MCFG_UCAM     (1<<7)
#define USB_MCFG_EBMEN    (1<<3)
#define USB_MCFG_EMEMEN   (1<<2)

#define USBH_ENABLE_CE    (USB_MCFG_PHYPLLEN | USB_MCFG_EHCCLKEN)
#ifdef CONFIG_DMA_COHERENT
#define USBH_ENABLE_INIT  (USBH_ENABLE_CE \
                         | USB_MCFG_PFEN | USB_MCFG_RDCOMB \
                         | USB_MCFG_SSDEN | USB_MCFG_UCAM \
                         | USB_MCFG_EBMEN | USB_MCFG_EMEMEN)
#else
#define USBH_ENABLE_INIT  (USBH_ENABLE_CE \
                         | USB_MCFG_PFEN | USB_MCFG_RDCOMB \
                         | USB_MCFG_SSDEN \
                         | USB_MCFG_EBMEN | USB_MCFG_EMEMEN)
#endif
#define USBH_DISABLE      (USB_MCFG_EBMEN | USB_MCFG_EMEMEN)

#endif  /* Au1200 */

extern int usb_disabled(void);

/*-------------------------------------------------------------------------*/

static void au1xxx_start_ehc(struct platform_device *dev)
{
	pr_debug(__FILE__ ": starting Au1xxx EHCI USB Controller\n");

	/* write HW defaults again in case Yamon cleared them */
	if (au_readl(USB_HOST_CONFIG) == 0) {
		au_writel(0x00d02000, USB_HOST_CONFIG);
		au_readl(USB_HOST_CONFIG);
		udelay(1000);
	}
	/* enable host controller */
	au_writel(USBH_ENABLE_CE | au_readl(USB_HOST_CONFIG), USB_HOST_CONFIG);
	au_readl(USB_HOST_CONFIG);
	udelay(1000);
	au_writel(USBH_ENABLE_INIT | au_readl(USB_HOST_CONFIG), USB_HOST_CONFIG);
	au_readl(USB_HOST_CONFIG);
	udelay(1000);

	pr_debug(__FILE__ ": Clock to USB host has been enabled\n");
}

static void au1xxx_stop_ehc(struct platform_device *dev)
{
	pr_debug(__FILE__ ": stopping Au1xxx EHCI USB Controller\n");

	/* Disable mem */
	au_writel(~USBH_DISABLE & au_readl(USB_HOST_CONFIG), USB_HOST_CONFIG);
	udelay(1000);
	/* Disable clock */
	au_writel(~USB_MCFG_EHCCLKEN & au_readl(USB_HOST_CONFIG), USB_HOST_CONFIG);
	au_readl(USB_HOST_CONFIG);
}


/*-------------------------------------------------------------------------*/


static irqreturn_t usb_ehci_au1xxx_hcim_irq (int irq, void *__hcd,
					    struct pt_regs * r)
{
	struct usb_hcd *hcd = __hcd;

	return usb_hcd_irq(irq, hcd, r);
}

/*-------------------------------------------------------------------------*/

void usb_ehci_au1xxx_remove (struct usb_hcd *, struct platform_device *);

/* configure so an HC device and id are always provided */
/* always called with process context; sleeping is OK */


/**
 * usb_ehci_au1xxx_probe - initialize Au1xxx-based HCDs
 * Context: !in_interrupt()
 *
 * Allocates basic resources for this USB host controller, and
 * then invokes the start() method for the HCD associated with it
 * through the hotplug entry's driver_data.
 *
 */
int usb_ehci_au1xxx_probe (const struct hc_driver *driver,
			   struct usb_hcd **hcd_out,
			   struct platform_device *dev)
{
	int retval;
	struct usb_hcd *hcd = 0;

	unsigned int *addr = NULL;

#if defined(CONFIG_SOC_AU1200) && ( \
     defined(CONFIG_DMA_COHERENT) || defined(CONFIG_CPU_HAS_LLSC))

	/* Au1200 AB USB does not support coherent memory */
	if (!(read_c0_prid() & 0xff)) {
		pr_info("Au1200 ehci: !!! This is chip revision AB                     !!!\n");
		pr_info("             !!! update your board or re-configure the kernel !!!\n");
		return -ENODEV;
	}
#endif
	if (!request_mem_region(dev->resource[0].start,
				dev->resource[0].end + 1
				- dev->resource[0].start, hcd_name)) {
		pr_debug("request_mem_region failed\n");
		return -EBUSY;
	}
	
	au1xxx_start_ehc(dev);
	
	addr = ioremap(dev->resource[0].start,
		       dev->resource[0].end + 1
		       - dev->resource[0].start);
	if (!addr) {
		pr_debug("ioremap failed\n");
		retval = -ENOMEM;
		goto err1;
	}

	if(dev->resource[1].flags != IORESOURCE_IRQ) {
		pr_debug ("resource[1] is not IORESOURCE_IRQ\n");
		retval = -ENOMEM;
		goto err1;
	}

	hcd = usb_create_hcd(driver);
	if (hcd == NULL) {
		pr_debug ("usb_create_hcd failed\n");
		retval = -ENOMEM;
		goto err1;
	}

	hcd->irq = dev->resource[1].start;
	hcd->regs = addr;
	hcd->self.controller = &dev->dev;

	retval = hcd_buffer_create (hcd);
	if (retval != 0) {
		pr_debug ("pool alloc fail\n");
		goto err2;
	}

	if ((retval = driver->reset (hcd)) < 0) {
		pr_debug ("can't reset hc\n");
		goto err2;
	}

	retval = request_irq (hcd->irq, usb_ehci_au1xxx_hcim_irq,
			      SA_SHIRQ | SA_INTERRUPT,
			      hcd->driver->description, hcd);
	if (retval != 0) {
		pr_debug("request_irq failed\n");
		retval = -EBUSY;
		goto err3;
	}

	pr_info ("%s (Au1xxx) at 0x%p, irq %d\n",
	     hcd->driver->description, hcd->regs, hcd->irq);

	hcd->self.bus_name = "au1xxx";

	usb_register_bus (&hcd->self);

#ifdef CONFIG_USB_OTG_HIGHSPEED
	if ((retval = driver->start_otg (hcd)) < 0)
		pr_info ("OTG init failed\n");
#endif
	if ((retval = driver->start (hcd)) < 0)
	{
		usb_ehci_au1xxx_remove(hcd, dev);
		pr_info ("bad driver->start\n");
		return retval;
	}

	*hcd_out = hcd;
	return 0;

 err3:
	hcd_buffer_destroy (hcd);
 err2:
	usb_put_hcd(hcd);
 err1:
	au1xxx_stop_ehc(dev);
	release_mem_region(dev->resource[0].start,
			   dev->resource[0].end + 1
			   - dev->resource[0].start);
	return retval;
}


/* may be called without controller electrically present */
/* may be called with controller, bus, and devices active */

/**
 * usb_ehci_hcd_au1xxx_remove - shutdown processing for Au1xxx-based HCDs
 * @dev: USB Host Controller being removed
 * Context: !in_interrupt()
 *
 * Reverses the effect of usb_ehci_hcd_au1xxx_probe(), first invoking
 * the HCD's stop() method.  It is always called from a thread
 * context, normally "rmmod", "apmd", or something similar.
 *
 */
void usb_ehci_au1xxx_remove (struct usb_hcd *hcd, struct platform_device *dev)
{
	pr_info ("remove: %s, state %x\n", hcd->self.bus_name, hcd->state);

	if (in_interrupt ())
		BUG ();

	if (HCD_IS_RUNNING (hcd->state))
		hcd->state = USB_STATE_QUIESCING;

	pr_debug ("%s: roothub graceful disconnect\n", hcd->self.bus_name);
	usb_disconnect (&hcd->self.root_hub);

	hcd->driver->stop (hcd);
#ifdef CONFIG_USB_OTG_HIGHSPEED
	hcd->driver->stop_otg (hcd);
#endif
	hcd_buffer_destroy (hcd);
	hcd->state = USB_STATE_HALT;

	free_irq (hcd->irq, hcd);
	iounmap (hcd->regs);
	release_mem_region(dev->resource[0].start,
			   dev->resource[0].end + 1
			   - dev->resource[0].start);

	usb_deregister_bus (&hcd->self);

	au1xxx_stop_ehc(dev);
}

/*-------------------------------------------------------------------------*/

static const struct hc_driver ehci_au1xxx_hc_driver = {
	.description =		hcd_name,
	.product_desc =		"Au1xxx EHCI",
	.hcd_priv_size =	sizeof(struct ehci_hcd),

	/*
	 * generic hardware linkage
	 */
	.irq =			ehci_irq,
	.flags =		HCD_MEMORY | HCD_USB2,

	/*
	 * basic lifecycle operations
	 */
	.reset =		ehci_hc_reset,
	.start =		ehci_start,
#ifdef	CONFIG_PM
	/* suspend:		ehci_au1xxx_suspend,  -- tbd */
	/* resume:		ehci_au1xxx_resume,   -- tbd */
#endif /*CONFIG_PM*/
	.stop =			ehci_stop,

	/*
	 * managing i/o requests and associated device resources
	 */
	.urb_enqueue =		ehci_urb_enqueue,
	.urb_dequeue =		ehci_urb_dequeue,
	.endpoint_disable =	ehci_endpoint_disable,

	/*
	 * scheduling support
	 */
	.get_frame_number =	ehci_get_frame,

	/*
	 * root hub support
	 */
	.hub_status_data =	ehci_hub_status_data,
	.hub_control =		ehci_hub_control,
#ifdef	CONFIG_USB_SUSPEND
	.hub_suspend =		ehci_hub_suspend,
	.hub_resume =		ehci_hub_resume,
#endif
	.start_port_reset =	ehci_start_port_reset,
#ifdef	CONFIG_USB_OTG_HIGHSPEED
	.start_otg =		ehci_start_otg,
	.stop_otg =		ehci_stop_otg,
#endif
};

/*-------------------------------------------------------------------------*/

static int ehci_hcd_au1xxx_drv_probe(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = NULL;
	int ret;

	pr_debug ("In ehci_hcd_au1xxx_drv_probe\n");

	if (usb_disabled())
		return -ENODEV;

	ret = usb_ehci_au1xxx_probe(&ehci_au1xxx_hc_driver, &hcd, pdev);

	if (ret == 0)
		dev_set_drvdata(dev, hcd);

	return ret;
}

static int ehci_hcd_au1xxx_drv_remove(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = dev_get_drvdata(dev);

	usb_ehci_au1xxx_remove(hcd, pdev);
	dev_set_drvdata(dev, NULL);
	return 0;
}
	/*TBD*/
/*static int ehci_hcd_au1xxx_drv_suspend(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = dev_get_drvdata(dev);

	return 0;
}
static int ehci_hcd_au1xxx_drv_resume(struct device *dev)
{
	struct platform_device *pdev = to_platform_device(dev);
	struct usb_hcd *hcd = dev_get_drvdata(dev);

	return 0;
}
*/

static struct device_driver ehci_hcd_au1xxx_driver = {
	.name		= "au1xxx-ehci",
	.bus		= &platform_bus_type,
	.probe		= ehci_hcd_au1xxx_drv_probe,
	.remove		= ehci_hcd_au1xxx_drv_remove,
	/*.suspend	= ehci_hcd_au1xxx_drv_suspend, */
	/*.resume	= ehci_hcd_au1xxx_drv_resume, */
};

static int __init ehci_hcd_au1xxx_init (void)
{
	pr_debug (DRIVER_INFO " (Au1xxx)\n");

	return driver_register(&ehci_hcd_au1xxx_driver);
}

static void __exit ehci_hcd_au1xxx_cleanup (void)
{
	driver_unregister(&ehci_hcd_au1xxx_driver);
}

module_init (ehci_hcd_au1xxx_init);
module_exit (ehci_hcd_au1xxx_cleanup);
