#include "os.h"
#include "focus.h"
#include "sio.h"
#include "sioll.h"

static uint32 focus_read(FOCUS_REGISTER_ID id)
{
	unsigned long data;
	SIO_read(FS453_ADDRESS, focus_registers[id].address, &data, focus_registers[id].size);
	return (uint32) data;
}

static void focus_write(FOCUS_REGISTER_ID id, uint32 value)
{
	SIO_write(FS453_ADDRESS, focus_registers[id].address, (unsigned long) value, focus_registers[id].size);
}

static void focus_set_bits(FOCUS_REGISTER_ID id, uint32 bits)
{
	focus_write(id, focus_read(id) | bits);
}

static void focus_clear_bits(FOCUS_REGISTER_ID id, uint32 bits)
{
	focus_write(id, focus_read(id) & ~bits);
}

void focus_soft_reset(void)
{
	focus_set_bits(CR, CR_SRESET);
	focus_clear_bits(CR, CR_SRESET);
}

void focus_latch_nco_pll(void)
{
	focus_set_bits(CR, CR_NCO_EN);
	focus_clear_bits(CR, CR_NCO_EN);
	focus_set_bits(CR, CR_GCC_CK_LVL);	//I don't think this applies to this operation
}

void focus_bridge_reset(void)
{
	focus_set_bits(MISC, MISC_BRDG_RST);
	OS_mdelay(100);
	focus_clear_bits(MISC, MISC_BRDG_RST);
}

void focus_set_qpr(FOCUS_VIDEO_OUTPUT_FORMAT format)
{
	uint32 qpr = QK_PROGRAM | QK_FF | QK_UIM_NATIONAL;
	
	switch(format)
	{
		case FORMAT_HDTV:		qpr |= QK_OS_480P | QK_YC_IN | QK_OM_COMPONENT | QK_GMODE_720X480;
								break;
		case FORMAT_COMPONENT:	qpr |= QK_OS_SDTV | QK_OM_COMPONENT | QK_GMODE_640X480;
								break;
		case FORMAT_CVSV:		qpr |= QK_OS_SDTV | QK_OM_CVSV | QK_GMODE_640X480;
								break;
	}
	
	focus_write(QPR, qpr);
}

void focus_set_video_mode(FOCUS_VIDEO_OUTPUT_FORMAT format)
{
	focus_clear_bits(VID_CNTRL0, 0x3);	//zero out current video mode
				
	switch(format)
	{
		case FORMAT_HDTV:		focus_set_bits(MISC_47, MISC_47_COMP_YUV);
								focus_set_bits(VID_CNTRL0, VID_MODE_HDTV);
								break;
		case FORMAT_COMPONENT:	focus_set_bits(MISC_47, MISC_47_COMP_YUV);
								focus_set_bits(VID_CNTRL0, VID_MODE_SDTV);
								break;
		case FORMAT_CVSV:		focus_set_bits(VID_CNTRL0, VID_MODE_CVSV);
								break;
	}
}

void focus_reg_dump(void)
{
	int i;
	for(i = 0; focus_registers[i].id != UNDEFINED_REGISTER; ++i)
	{
		OS_printf("%12s (0x%02X) = 0x%X\n", focus_registers[i].description, focus_registers[i].address, focus_read(focus_registers[i].id));		
	}	
}

int board_au1200fb_focus_shutdown(void)
{
	SIOLL_shutdown_focus();
	return 1;
}

int focus_init(FOCUS_VIDEO_OUTPUT_FORMAT format, const FOCUS_REGISTER_VALUE* regs)
{	
	int i;
	if(regs == NULL)
		return 0;
		
	SIOLL_init_focus();
	SIO_init();

	focus_clear_bits(CR, CR_NCO_EN | CR_SRESET);
	
	focus_set_qpr(format);
	
	focus_soft_reset();

	//Program configuration
	for(i = 0; regs[i].id != UNDEFINED_REGISTER; ++i)
		focus_write(regs[i].id, regs[i].value);
	
	focus_latch_nco_pll();	
	
	focus_bridge_reset();
	
	//Clear the CACQ and FIFO flags	
	focus_set_bits(CR, CR_FIFO_CLR | CR_CACQ_CLR);
	OS_mdelay(100);
	focus_clear_bits(CR, CR_FIFO_CLR | CR_CACQ_CLR);
	
	focus_set_video_mode(format);

//	focus_reg_dump();
	SIO_shutdown();
	
	return 1;
}

int board_au1200fb_focus_init_component(void)
{
	return focus_init(FORMAT_COMPONENT, register_set_sdtv);
}

int board_au1200fb_focus_init_cvsv(void)
{
	return focus_init(FORMAT_CVSV, register_set_sdtv);
}

int board_au1200fb_focus_init_hdtv(void)
{
	return focus_init(FORMAT_HDTV, register_set_hdtv);	
}

