//	SIO_Focus.c

// Copyright (c) 1999-2001, FOCUS Enhancements, Inc., All Rights Reserved.

#include "os.h"
#include "sioll.h"
#include "sio.h"


#define	SIOWRITE		0x00		// Write address
#define	SIOREAD			0x01		// Read address

#define	SIOACK			0x00		// Ack value
#define	SIONACK			0x01		// Not - ack value

// #define TRACE(x) printf(x);

//----------------------------------------------------------------------
//
//	void send_start(void)
//	
//	Sends an SIO start signal on the bus.
//	
//----------------------------------------------------------------------
static void send_start(void)
{
	SIOLL_output_data(1);
	SIOLL_output_clock(1);
	SIOLL_output_data(0);
	SIOLL_output_clock(0);
}

//----------------------------------------------------------------------
//
//	void send_stop(void)
//	
//	Sends an SIO stop signal on the bus.
//	
//----------------------------------------------------------------------
static void send_stop(void)
{
	SIOLL_output_data(0);
	SIOLL_output_clock(1);
	SIOLL_output_data(1);
}

//----------------------------------------------------------------------
//
//	void send_ack(void)
//	
//	Sends the Ack signal on the SIO bus.
//	
//----------------------------------------------------------------------
static void send_ack(void)
{
	SIOLL_output_data(0);
	SIOLL_output_clock(1);
	SIOLL_output_clock(0);
}

//----------------------------------------------------------------------
//
//	void send_nack(void)
//	
//	Sends the Nt-Ack signal on the SIO bus.
//	
//----------------------------------------------------------------------
static void send_nack(void)
{
	SIOLL_output_data(1);
	SIOLL_output_clock(1);
	SIOLL_output_clock(0);
}

//----------------------------------------------------------------------
//
//	UInt8 send_data( UInt8 inData )
//	
//	Sends a byte of data on the SIO bus and returns the TRUE if the slave ACK'd
//	the data.
//	
//	Input:	inData			- the byte of data to send          
//	Output:	(return)		- TRUE (1) if ACK was received, FALSE (0) if not
//	
//----------------------------------------------------------------------
static void send_data(unsigned char data)
{
	unsigned char bit;
	  
//printf("Sending Data: %X\n", data);
	// Send all 8 bits of data byte, MSB to LSB
	for (bit = 0x80; bit != 0; bit >>= 1)
	{       
		if (data & bit)
			SIOLL_output_data(1);
		else
			SIOLL_output_data(0);
			
		SIOLL_output_clock(1);
		SIOLL_output_clock(0);		
	}
}

//----------------------------------------------------------------------
//
//	UInt8 receive_ack(  )
//	
//	Receives the Ack (or Nack) from the slave. 
//	
//	Output:	(return) - TRUE (1) if ACK was received, FALSE (0) if not
//	
//----------------------------------------------------------------------
static unsigned char receive_ack(void)
{
	unsigned char bit;

	//printf("Receive ACK\n"); getc();
	// Test for Ack/Nack
	SIOLL_set_data_for_input();
	//SIOLL_output_data(1);
	SIOLL_output_clock(1);
	bit = SIOLL_input_data();
	SIOLL_output_clock(0);
	SIOLL_set_data_for_output();
	//printf("Ack: %d\n", bit);
	return !bit;
}
	
//----------------------------------------------------------------------
//
//	unsigned char receive_data(void)
//	
//	Receives a byte of data from the SIO bus.
//	
//	Output:	(return) - The data byte recehved from the bus
//	
//----------------------------------------------------------------------
static unsigned char receive_data(void)
{
	unsigned char data = 0;
	unsigned char x;

	// make sure the data line is released
	SIOLL_set_data_for_input();
	SIOLL_output_data(1);
	
	// shift in the data
	for (x=0; x<8; x++)
    {   
    	// shift the data left  
    	SIOLL_output_clock(1);
    	data <<= 1;
	   	data |= SIOLL_input_data();
		SIOLL_output_clock(0);
    }

	SIOLL_set_data_for_output();
	SIOLL_output_data(1);
	return data;                        
}


//----------------------------------------------------------------------
//
//	void SIO_init(void)
//	
//	This routine initializes the SIO interface. Clients of the SIO module
//	will call this routine before calling any other routine in the module.
//	
//----------------------------------------------------------------------
static int g_initialized = 0;

int SIO_init(void)
{
	int err;

	DPRINTF("\n");

	// init SIO low-level abstraction layer
	err = SIOLL_init_smbus();
	if (err) return err;

	// set the clock and data lines to the proper states
	SIOLL_output_clock(1);
	SIOLL_output_data(1);
	SIOLL_set_data_for_output();

	send_start();
	send_stop();
	send_stop();

	g_initialized = 1;

	return 0;
}

//----------------------------------------------------------------------
//	
//	void SIO_cleanup(void)
//	
//	This routine disables the SIO interface. Clients of the SIO module will not
//	call any other SIO routine after calling this routine.
//	
//----------------------------------------------------------------------
void SIO_shutdown(void)
{
	if (g_initialized)
	{
		DPRINTF("\n");

		// set the clock and data lines to a harmless state
		SIOLL_output_clock(1);
		SIOLL_output_data(1);

		g_initialized = 0;
	}

	SIOLL_shutdown_smbus();
}


int SIO_read(unsigned char address, unsigned int reg, unsigned long *p_value, unsigned int bytes)
{
	int restart_count = 0;
	unsigned int data;
		  
	//printf("SIO_READ, addr: %X, reg: %X\n", address, reg);
	if (!p_value)
		return ERR_INVALID_PARAMETER;

	while (restart_count++ < 5)
	{
		*p_value = 0x0D0A;

		// set the access pointer register.
		// The address is shifted left by one to make room for Read/Write bit 
		send_start();
		send_data((char)((address << 1) | SIOWRITE));
		if (!receive_ack())
		{
			DPRINTF("NACK received after component address.\n");
			send_stop();
			OS_mdelay(10);
			continue;
		}
		send_data((unsigned char)(reg & 0xFF));
		send_nack();
		
		if (bytes >= 1)
		{
			// read the first data byte.
			send_start();
			send_data((char)((address << 1) | SIOREAD));
			if (!receive_ack())
			{
				DPRINTF("NACK received after register.\n");
				send_stop();
				OS_mdelay(10);
				continue;
			}
			*p_value = receive_data();
		}

		if (bytes >= 2)
		{
			// read the second byte.
			send_ack();
			data = receive_data();
			*p_value |= (data << 8);
		}	

		if (bytes >= 3)
		{
			// read the third byte.
			send_ack();
			data = receive_data();
			*p_value |= (data << 16);
		}	

		if (bytes >= 4)
		{
			// read the fourth byte.
			send_ack();
			data = receive_data();
			*p_value |= (data << 24);
		}	

		if (bytes > 4)
		{
			DPRINTF("requested more than 4 bytes!\n");
		}

		// done.
		send_nack();
		send_stop();

//		TRACE(("SIO_read(0x%x,0x%x,0x%lx,%u)\n",address,reg,*p_value,bytes))
		
		return 0;
	}

	DPRINTF("(0x%x) failed, too many NACKs.\n",reg);

	return ERR_SIO_READ_FAILED;
}

int SIO_read_current_offset(unsigned char address, unsigned long *p_value, unsigned int bytes)
{
	int restart_count = 0;
	unsigned int data;
	
	if (!p_value)
		return ERR_INVALID_PARAMETER;

	while (restart_count++ < 5)
	{
		*p_value = 0x0D0A;

		// The address is shifted left by one to make room for Read/Write bit 
		
		if (bytes >= 1)
		{
			// read the first data byte.
			send_start();
			send_data((char)((address << 1) | SIOREAD));
			if (!receive_ack())
			{
				DPRINTF("NACK received after register.\n");
				send_stop();
				OS_mdelay(10);
				continue;
			}
			*p_value = receive_data();
		}

		if (bytes >= 2)
		{
			// read the second byte.
			send_ack();
			data = receive_data();
			*p_value |= (data << 8);
		}	

		if (bytes >= 3)
		{
			// read the third byte.
			send_ack();
			data = receive_data();
			*p_value |= (data << 16);
		}	

		if (bytes >= 4)
		{
			// read the fourth byte.
			send_ack();
			data = receive_data();
			*p_value |= (data << 24);
		}	

		if (bytes > 4)
		{
			DPRINTF("requested more than 4 bytes!\n");
		}

		// done.
		send_nack();
		send_stop();

		return 0;
	}

	DPRINTF("failed, too many NACKs.\n");

	return ERR_SIO_READ_FAILED;
}

int SIO_write(unsigned char address, unsigned int reg, unsigned long value, unsigned int bytes)
{
	int restart_count = 0;

	while (restart_count++ < 5)
	{
		// set the access pointer register.
		// The address is shifted left by one to make room for Read/Write bit 
		send_start();
		send_data((char)((address << 1) | SIOWRITE));
		if (!receive_ack())
		{
			DPRINTF("NACK received after component address.\n");
			send_stop();
			OS_mdelay(10);
			continue;
		}
		send_data((unsigned char)reg);
		if (!receive_ack())
		{
			DPRINTF("NACK received after register.\n");
			send_stop();
			OS_mdelay(10);
			continue;
		}

		if (bytes >= 1)
		{
			// write the first byte.
			send_data((unsigned char)value);
			if (!receive_ack())
			{
				DPRINTF("NACK received after first data byte.\n");
				send_stop();
				OS_mdelay(10);
				continue;
			}
		}

		if (bytes >= 2)
		{
			// write the second byte.
			send_data((unsigned char)(value >> 8));
			if (!receive_ack())
			{
				DPRINTF("NACK received after second data byte.\n");
				send_stop();
				OS_mdelay(10);
				continue;
			}
		}

		if (bytes >= 3)
		{
			// write the third byte.
			send_data((unsigned char)(value >> 16));
			if (!receive_ack())
			{
				DPRINTF("NACK received after second data byte.\n");
				send_stop();
				OS_mdelay(10);
				continue;
			}
		}

		if (bytes >= 4)
		{
			// write the fourth byte.
			send_data((unsigned char)(value >> 24));
			if (!receive_ack())
			{
				DPRINTF("NACK received after second data byte.\n");
				send_stop();
				OS_mdelay(10);
				continue;
			}
		}

		if (bytes > 4)
		{
			DPRINTF("requested more than 4 bytes!\n");
		}

		// done.
		send_stop();

//		TRACE(("SIO_write(0x%x,0x%x,0x%lx,%u)\n",address,reg,value,bytes))

		unsigned long data = 0;
		SIO_read(address, reg, &data, bytes);
		
		if(data != value)
		{
			OS_printf("FOCUS Data integrity check failed!\n");	
			OS_printf("FOCUS Reg 0x%02X = 0x%08X, Wrote: 0x%08X\n", reg, (uint32) data, (uint32) value);					
			return ERR_SIO_WRITE_FAILED;
		}
		else
		{
			#ifdef _DEBUG
				//OS_printf("FOCUS Register 0x%02X = 0x%08X\n", reg, (uint32) data);
			#endif			
		}
		
		//OS_mdelay(100);
		return 0;
	}

	DPRINTF("(0x%x) failed, too many NACKs.\n",reg);

	return ERR_SIO_WRITE_FAILED;
}

