/*
 *  au1xxx_pm.h - AMD Au1200 power management interface
 *
 *  Based heavily on include/linux/pm.h by Andrew Henroid
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef _AU1XXX_PM_H
#define _AU1XXX_PM_H

#ifdef __KERNEL__

#include <linux/config.h>
#include <linux/list.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>

#ifdef CONFIG_PM

/* Device Definitions */
#define ERROR_INVALID_DEVICE 		(-1)
#define MAX_SUPPORTED_DEVICES   (20)
#define MAX_SUPPORTED_ERRORS    (20)
#define MAX_SUPPORTED_TIMERS    (20)
#define MAX_NAME_WIDTH          (30)
#define SLEEP_STATE 						(0)
#define AWAKE_STATE 						(100)
#define SYSTEM_TIMEOUT          (0)

/* Settings Defintions */
#define GPIO_WAKEUP 						(0)
#define TIMER_WAKEUP 						(1)
#define DEFAULT_WAKEUP 					GPIO_WAKEUP
#define ASCII_FEEDBACK 					(0)
#define BINARY_FEEDBACK 				(1)
#define DEFAULT_FEEDBACK 				ASCII_FEEDBACK

enum au1xxx_request {
	AU1XXX_PM_SLEEP,
	AU1XXX_PM_WAKEUP,
	AU1XXX_PM_GETSTATUS,
	AU1XXX_PM_ACCESS,
	AU1XXX_PM_IDLE,
	AU1XXX_PM_CLEANUP
};
typedef enum au1xxx_request au1xxx_request_t;

enum au1xxx_pm_report {
	AU1XXX_PM_ERROR,
	AU1XXX_PM_PENDING,
	AU1XXX_PM_DEVICE,
	AU1XXX_PM_SETTING,
	AU1XXX_PM_TIMEOUT
};
typedef enum au1xxx_pm_report au1xxx_pm_report_t;

enum au1xxx_pm_error_code {
	AU1XXX_PM_INVALID_SETTING_ERROR,
	AU1XXX_PM_NO_MEMORY_ERROR,
	AU1XXX_PM_FILESYSTEM_ERROR,
	AU1XXX_PM_OUT_OF_RESOURCES_ERROR
};
typedef enum au1xxx_pm_error_code au1xxx_pm_error_code_t;

struct au1xxx_pm_public_settings {
	int wakeup_mode;
	int sleep_time;
	int feedback_mode;
};
typedef struct au1xxx_pm_public_settings au1xxx_pm_public_settings_t;

struct au1xxx_pm_public_dev {
	int state_change_pending;
	int last_error;
	int cur_state;
	int prev_state;
	char name[MAX_NAME_WIDTH];
	int valid;
};
typedef struct au1xxx_pm_public_dev au1xxx_pm_public_dev_t;

typedef struct au1xxx_power_dev au1xxx_power_dev_t;

typedef int (*au1xxx_power_callback)(au1xxx_power_dev_t *device,
	au1xxx_request_t request, void *data);

struct au1xxx_power_dev
{
	/* The first elements of this structure are the same as
	 * au1xxx_pm_public_dev_t because we want to be able to do some
	 * pseudo-polymorphic stuff with them. */
	int state_change_pending;
	int last_error;
	int cur_state;
	int prev_state;
	char name[MAX_NAME_WIDTH];
	int valid;
	
	au1xxx_power_callback callback;
	struct proc_dir_entry *proc_entry;

	//int access_mode;  /* 0 for device controlled, non-zero for system controlled */
	struct timeval cur_state_timestamp;
	struct timeval prev_state_timestamp;
	unsigned long timeout;  /* in system ticks */
	struct timer_list timer;
	struct proc_dir_entry *timer_proc_entry;

	void *data;
};

struct au1xxx_pm_error {
	au1xxx_power_dev_t *dev;
	int error_code;
	int error_count;
	struct timeval first_timestamp;
	struct timeval last_timestamp;
};
typedef struct au1xxx_pm_error au1xxx_pm_error_t;

struct au1xxx_pm_bin_header {
	au1xxx_pm_report_t report_type;
	int num_reports;
};
typedef struct au1xxx_pm_bin_header au1xxx_pm_bin_header_t;

au1xxx_power_dev_t *new_au1xxx_power_device(const char *name, 
		au1xxx_power_callback cb_func, void *data);
int remove_au1xxx_power_device(au1xxx_power_dev_t *dev);
int au1xxx_pm_access(au1xxx_power_dev_t *dev);
int au1xxx_pm_idle(au1xxx_power_dev_t *dev);
int au1xxx_pm_struct_init(void);
int au1xxx_pm_set_timer(au1xxx_power_dev_t *dev, unsigned long timeout_val);
int au1xxx_pm_del_timer(au1xxx_power_dev_t *dev);

extern int pm_active;

#define PM_IS_ACTIVE() (pm_active != 0)

#else /* CONFIG_PM */

#define PM_IS_ACTIVE() 0

#endif /* CONFIG_PM */
#endif /* __KERNEL__ */
#endif /* _AU1XXX_PM_H */
